#ifndef __CHANNEL_H
#define __CHANNEL_H

#include <qtextbrowser.h>
#include <qaction.h>

class QPopupMenu;
class ChatTab;

class Channel: public QTextBrowser
{
    Q_OBJECT
private:
    bool prv;
    bool board;
    bool activity;

    QAction *selectAllAction;
    QAction *clearChatAction;
    QAction *saveAction;
    QAction *leaveAction;

    QString chname;
    QString topic;
    QString filename;
    QString chatlinetext;
    
    void mouseMoveEvent ( QMouseEvent * e );

protected:
    bool doSave();
    QPopupMenu *createPopupMenu(const QPoint &pos);
    void keyPressEvent ( QKeyEvent * e );
    void setupActions();

signals:
    void sig_topic_clear(const QString &);		
    void leave(const QString &name);
    void keypressed(const QString &);
    void activityDetected();

private slots:
    void topic_clear();
    void leave();

public slots:
    void save();
    void saveAs();		
public:
    Channel(QWidget *parent, const QString &name, QColor back, QFont chatfont, bool prv=false, bool brd=false);

    ~Channel();

    bool isPrivate() { return prv; }
    bool isBoard() { return board; }
    bool getActivity() { return activity; }
    bool setActivity(bool a) { return activity = a; }
    bool setChatLineText(const QString &text) { chatlinetext = text; return true; }

    QString getChatLineText() { return chatlinetext; }
    QString name() { return chname; }
    QString setTopic(const QString &t) { return topic = t; }
    QString getTopic() { return topic; }
		
    friend class ChatTab;
};

#endif

