#include "crypto.h"
//#include <iostream>

PROV_ENUMALGS_EX aProvEnumAlgsEx[4][RSAENH_MAX_ENUMALGS+1] =
{
    {
	{CALG_RC4, 40, 40, 56, 0, 4, "RC4", 24, "RSA Data Security's RC4"},
	{CALG_SHA, 160,160, 160, CRYPT_FLAG_SIGNING, 6, "SHA-1", 30, "Secure Hash Algorithm (SHA-1)"},
	{CALG_MD5, 128, 128, 128, CRYPT_FLAG_SIGNING, 4, "MD5", 23,"Message Digest 5 (MD5)"},
	{CALG_RSA_SIGN, 512, 384, 16384, CRYPT_FLAG_SIGNING|CRYPT_FLAG_IPSEC, 9, "RSA_SIGN", 14, "RSA Signature"},
	{CALG_RSA_KEYX, 512, 384, 1024, CRYPT_FLAG_SIGNING|CRYPT_FLAG_IPSEC, 9, "RSA_KEYX", 17, "RSA Key Exchange"},
	{0,               0,  0,    0,0,                    1,"",         1,""}
    },
{
    {CALG_RC4,      128, 40,  128,0,                    4,"RC4",     24,"RSA Data Security's RC4"},
    {CALG_SHA,      160,160,  160,CRYPT_FLAG_SIGNING,   6,"SHA-1",   30,"Secure Hash Algorithm (SHA-1)"},
    {CALG_MD5,      128,128,  128,CRYPT_FLAG_SIGNING,   4,"MD5",     23,"Message Digest 5 (MD5)"},
    {CALG_RSA_SIGN,1024,384,16384,CRYPT_FLAG_SIGNING|CRYPT_FLAG_IPSEC,9,"RSA_SIGN",14,"RSA Signature"},
    {CALG_RSA_KEYX,1024,384,16384,CRYPT_FLAG_SIGNING|CRYPT_FLAG_IPSEC,9,"RSA_KEYX",17,"RSA Key Exchange"},
    {0,               0,  0,    0,0,                    1,"",         1,""}
},
{
{CALG_RC4,      128, 40,  128,0,                    4,"RC4",     24,"RSA Data Security's RC4"},
{CALG_SHA,      160,160,  160,CRYPT_FLAG_SIGNING,   6,"SHA-1",   30,"Secure Hash Algorithm (SHA-1)"},
{CALG_MD5,      128,128,  128,CRYPT_FLAG_SIGNING,   4,"MD5",     23,"Message Digest 5 (MD5)"},
{CALG_RSA_SIGN,1024,384,16384,CRYPT_FLAG_SIGNING|CRYPT_FLAG_IPSEC,9,"RSA_SIGN",14,"RSA Signature"},
{CALG_RSA_KEYX,1024,384,16384,CRYPT_FLAG_SIGNING|CRYPT_FLAG_IPSEC,9,"RSA_KEYX",17,"RSA Key Exchange"},
{0,               0,  0,    0,0,                    1,"",         1,""}
 },
{
 {CALG_RC4,      128, 40,  128,RSAENH_PCT1_SSL2_SSL3_TLS1, 4,"RC4",        24,"RSA Data Security's RC4"},
 {CALG_SHA,160,160,160,CRYPT_FLAG_SIGNING|RSAENH_PCT1_SSL2_SSL3_TLS1,6,"SHA-1",30,"Secure Hash Algorithm (SHA-1)"},
 {CALG_MD5,128,128,128,CRYPT_FLAG_SIGNING|RSAENH_PCT1_SSL2_SSL3_TLS1,4,"MD5",23,"Message Digest 5 (MD5)"},
 {CALG_RSA_SIGN,1024,384,16384,CRYPT_FLAG_SIGNING|RSAENH_PCT1_SSL2_SSL3_TLS1,9,"RSA_SIGN",14,"RSA Signature"},
  {CALG_RSA_KEYX,1024,384,16384,CRYPT_FLAG_SIGNING|RSAENH_PCT1_SSL2_SSL3_TLS1,9,"RSA_KEYX",17,"RSA Key Exchange"},
 {0,               0,  0,    0,0,                          1,"",            1,""}
 }
};

static inline void reverse_bytes(unsigned char *pbData, DWORD dwLen)
{
    unsigned char swap;
    uint i;

    for (i = 0; i < dwLen / 2; i++) {
        swap = pbData[i];
        pbData[i] = pbData[dwLen - i - 1];
        pbData[dwLen - i - 1] = swap;
    }
}

static inline const PROV_ENUMALGS_EX* get_algid_info(ALG_ID algid) 
{
    PROV_ENUMALGS_EX *iterator;

    for (iterator = aProvEnumAlgsEx[RSAENH_PERSONALITY_BASE]; iterator->aiAlgid; iterator++) 
    {
	if (iterator->aiAlgid == algid) return iterator;
    }
    
    return NULL;
}

bool pkcs_1_mgf1(ALG_ID Algid, const unsigned char *seed, unsigned int seedlen, unsigned char *mask, unsigned int masklen)
{
    CRYPTHASH *pHash;
    unsigned int hLen, x;
    ulong32       counter;
    unsigned char *buf;
 
    if (!CryptCreateHash(Algid, 0, &pHash)) return false;

    /* get hash output size */
    hLen = pHash->dwHashSize;

    /* allocate memory */
    buf = (unsigned char*)malloc(hLen);
    if (buf == NULL) {
	CryptDestroyHash(pHash);
	return false;
    }

    /* start counter */
    counter = 0;

    while (masklen > 0) {
	/* handle counter */
	STORE32H(counter, buf);
	++counter;

	/* get hash of seed || counter */
	init_hash_impl(Algid, &pHash->context);

	update_hash_impl(Algid, &pHash->context, seed, seedlen);

	update_hash_impl(Algid, &pHash->context, buf, 4);

	finalize_hash_impl(Algid, &pHash->context, buf);

	/* store it */
	for (x = 0; x < hLen && masklen > 0; x++, masklen--) {
	    *mask++ = buf[x];
	}
    }
    free(buf);
    CryptDestroyHash(pHash);

    return true;
}

int PKCS1_MGF1(unsigned char *mask, long len, const unsigned char *seed, long seedlen, ALG_ID aiAlgid)
{
    long i, outlen = 0;
    unsigned char cnt[4];
    unsigned char md[64];
    CRYPTHASH *pHash;

    if (!CryptCreateHash(aiAlgid, 0, &pHash)) return -1;
	
    DWORD dwHashLen = pHash->dwHashSize;
    for (i = 0; outlen < len; i++)
    {
	cnt[0] = (unsigned char)((i >> 24) & 255);
	cnt[1] = (unsigned char)((i >> 16) & 255);
	cnt[2] = (unsigned char)((i >> 8)) & 255;
	cnt[3] = (unsigned char)(i & 255);
		
	init_hash_impl(aiAlgid, &pHash->context);
	update_hash_impl(aiAlgid, &pHash->context, seed, seedlen);
	update_hash_impl(aiAlgid, &pHash->context, cnt, 4);
	if (outlen + dwHashLen <= len)
	{
	    finalize_hash_impl(aiAlgid, &pHash->context, mask + outlen);
	    outlen += dwHashLen;
	}
	else
	{
	    finalize_hash_impl(aiAlgid, &pHash->context, md);
	    memcpy(mask + outlen, md, len - outlen);
	    outlen = len;
	}
    }
    CryptDestroyHash(pHash);
    return 0;
}

int MGF1(unsigned char *mask, long len, const unsigned char *seed, long seedlen)
{
    return PKCS1_MGF1(mask, len, seed, seedlen, CALG_SHA);
}

bool pkcs_1_oaep_encode(const unsigned char *msg, unsigned int msglen, const unsigned char *lparam, unsigned int lparamlen, unsigned int modulus_bitlen, ALG_ID Algid, unsigned char *out, unsigned int *outlen)
{
    CRYPTHASH *pCryptHash;
    const PROV_ENUMALGS_EX *peaAlgidInfo;
    unsigned char *DB, *seed, *mask;
    unsigned int hLen, x, y, modulus_len;
    bool ret = false;

    peaAlgidInfo = get_algid_info(Algid);
    if (!peaAlgidInfo) 
	return false;

    hLen = peaAlgidInfo->dwDefaultLen >> 3;
    modulus_len = (modulus_bitlen >> 3) + (modulus_bitlen & 7 ? 1 : 0);

    /* test message size */
    if ((2*hLen >= (modulus_len - 2)) || (msglen > (modulus_len - 2*hLen - 2)))
	return false;

    /* allocate ram for DB/mask/salt of size modulus_len */
    DB   = (unsigned char*)malloc(modulus_len);
    mask = (unsigned char*)malloc(modulus_len);
    seed = (unsigned char*)malloc(hLen);
    if (DB == NULL || mask == NULL || seed == NULL) {
	if (DB != NULL) {
	    free(DB);
	}
	if (mask != NULL) {
	    free(mask);
	}
	if (seed != NULL) {
	    free(seed);
	}
	return false;
    }

    /* get lhash */
    /* DB == lhash || PS || 0x01 || M, PS == k - mlen - 2hlen - 2 zeroes */
    x = modulus_len;

    CryptCreateHash(Algid, 0, &pCryptHash);

    CryptHashData(pCryptHash, (const unsigned char *)lparam, lparamlen, 0);

    CryptGetHashParam(pCryptHash, HP_HASHVAL, (unsigned char*)DB, &x, 0);

    CryptDestroyHash(pCryptHash);

    /* append PS then 0x01 (to lhash)  */
    x = hLen;
    y = modulus_len - msglen - 2*hLen - 2;
    memset(DB+x, 0, y);
    x += y;

    /* 0x01 byte */
    DB[x++] = 0x01;

    /* message (length = msglen) */
    memcpy(DB+x, msg, msglen);
    x += msglen;

    /* now choose a random seed */
    if (!gen_rand(seed, hLen)) goto LBL_ERR;

    /* compute MGF1 of seed (k - hlen - 1) */
    if (!pkcs_1_mgf1(Algid, seed, hLen, mask, modulus_len - hLen - 1)) goto LBL_ERR;

    /* xor against DB */
    for (y = 0; y < (modulus_len - hLen - 1); y++) {
	DB[y] ^= mask[y]; 
    }

    /* compute MGF1 of maskedDB (hLen) */ 
    if (!pkcs_1_mgf1(Algid, DB, modulus_len - hLen - 1, mask, hLen)) goto LBL_ERR;

    /* XOR against seed */
    for (y = 0; y < hLen; y++) {
	seed[y] ^= mask[y];
    }

    /* create string of length modulus_len */
    if (*outlen < modulus_len) {
	*outlen = modulus_len;
	goto LBL_ERR;
    }

    /* start output which is 0x00 || maskedSeed || maskedDB */
    x = 0;
    out[x++] = 0x00;
    memcpy(out+x, seed, hLen);
    x += hLen;
    memcpy(out+x, DB, modulus_len - hLen - 1);
    x += modulus_len - hLen - 1;

    *outlen = x;
    
    ret = true;

LBL_ERR:

    free(seed);
    free(mask);
    free(DB);

    return ret;
}

int RSA_padding_add_PKCS1_OAEP(unsigned char *to, int tlen, const unsigned char *from, int flen, const unsigned char *param, int plen)
{
    DWORD dwHashLen = SHA_DIGEST_LENGTH;
    CRYPTHASH *pCryptHash;
    int i, emlen = tlen - 1;
    unsigned char *db, *seed;
    unsigned char *dbmask, seedmask[SHA_DIGEST_LENGTH];
    
    if (flen > emlen - 2 * SHA_DIGEST_LENGTH - 1)
	return 0;

    if (emlen < 2 * SHA_DIGEST_LENGTH + 1)
	return 0;

    dbmask = (unsigned char*)malloc(emlen - SHA_DIGEST_LENGTH);
    if (dbmask == NULL)
	return 0;

    to[0] = 0;
    seed = to + 1;
    db = to + SHA_DIGEST_LENGTH + 1;
    
    CryptCreateHash(CALG_SHA, 0, &pCryptHash);

    CryptHashData(pCryptHash, (const unsigned char *)param, plen, 0);

    CryptGetHashParam(pCryptHash, HP_HASHVAL, (unsigned char*)db, &dwHashLen, 0);
    CryptDestroyHash(pCryptHash);

    memset(db + SHA_DIGEST_LENGTH, 0, emlen - flen - 2 * SHA_DIGEST_LENGTH - 1);
    
    db[emlen - flen - SHA_DIGEST_LENGTH - 1] = 0x01;
    
    memcpy(db + emlen - flen - SHA_DIGEST_LENGTH, from, (unsigned int) flen);
    
    if (!gen_rand(seed, SHA_DIGEST_LENGTH))
	return 0;
    
    MGF1(dbmask, emlen - SHA_DIGEST_LENGTH, seed, SHA_DIGEST_LENGTH);
    for (i = 0; i < emlen - SHA_DIGEST_LENGTH; i++)
	db[i] ^= dbmask[i];
    
    MGF1(seedmask, SHA_DIGEST_LENGTH, db, emlen - SHA_DIGEST_LENGTH);
    for (i = 0; i < SHA_DIGEST_LENGTH; i++)
	seed[i] ^= seedmask[i];
    
    free(dbmask);
    return 1;
}

int RSA_padding_check_PKCS1_OAEP(unsigned char *to, int tlen, const unsigned char *from, int flen, int num, const unsigned char *param, int plen)
{
    DWORD dwHashLen = SHA_DIGEST_LENGTH;
    CRYPTHASH *pCryptHash;
    int i, dblen, mlen = -1;
    const unsigned char *maskeddb;
    int lzero;
    unsigned char *db = NULL, seed[SHA_DIGEST_LENGTH], phash[SHA_DIGEST_LENGTH];
    int bad = 0;

    if (--num < 2 * SHA_DIGEST_LENGTH + 1)
	goto decoding_err;

    lzero = num - flen;

    if (lzero < 0)
    {
	/* lzero == -1 */
	/* signalling this error immediately after detection might allow
	 * for side-channel attacks (e.g. timing if 'plen' is huge
	 * -- cf. James H. Manger, "A Chosen Ciphertext Attack on RSA Optimal
	 * Asymmetric Encryption Padding (OAEP) [...]", CRYPTO 2001),
	 * so we use a 'bad' flag */
	bad = 1;
	lzero = 0;
    }
    maskeddb = from - lzero + SHA_DIGEST_LENGTH;

    dblen = num - SHA_DIGEST_LENGTH;
    db = (unsigned char*)malloc(dblen);
    if (!db) return -1;

    MGF1(seed, SHA_DIGEST_LENGTH, maskeddb, dblen);
    
    for (i = lzero; i < SHA_DIGEST_LENGTH; i++)
	seed[i] ^= from[i - lzero];
    
    MGF1(db, dblen, seed, SHA_DIGEST_LENGTH);
    for (i = 0; i < dblen; i++)
	db[i] ^= maskeddb[i];
    
    CryptCreateHash(CALG_SHA, 0, &pCryptHash);

    CryptHashData(pCryptHash, (const unsigned char *)param, plen, 0);

    CryptGetHashParam(pCryptHash, HP_HASHVAL, (unsigned char*)phash, &dwHashLen, 0);

    CryptDestroyHash(pCryptHash);
    
    if (memcmp(db, phash, SHA_DIGEST_LENGTH) != 0 || bad)
	goto decoding_err;
    else
    {
	for (i = SHA_DIGEST_LENGTH; i < dblen; i++)
	    if (db[i] != 0x00) break;
		
	if (db[i] != 0x01 || i++ >= dblen)
	    goto decoding_err;
	else
	{
	    /* everything looks OK */
	    mlen = dblen - i;
	    if (tlen < mlen) 
		mlen = -1;
	    else
		memmove(to, db + i, mlen);
	}
    }
    free(db);
    return mlen;

    decoding_err:
	/* to avoid chosen ciphertext attacks, the error message should not reveal
	which kind of decoding error happened */

	if (db) free(db);
	return -1;
}

static bool pad_data(const unsigned char *abData, DWORD dwDataLen, unsigned char *abBuffer, DWORD dwBufferLen, DWORD dwFlags)
{
    DWORD i;
    
    /* Ensure there is enough space for PKCS1 #2 padding */
    if (dwDataLen > dwBufferLen - 11) 
	return false;

    switch(dwFlags)
    {
    case CRYPT_SSL2_FALLBACK:
	memmove(abBuffer + dwBufferLen - dwDataLen, abData, dwDataLen);           		abBuffer[0] = 0x00;
	abBuffer[1] = RSAENH_PKC_BLOCKTYPE; 
	for (i = 2; i < dwBufferLen - dwDataLen - 1; i++) 
	    do gen_rand(&abBuffer[i], 1); while (!abBuffer[i]);
	if (dwFlags & CRYPT_SSL2_FALLBACK) 
	    for (i -= 8; i < dwBufferLen - dwDataLen - 1; i++) 
		abBuffer[i] = 0x03;
	abBuffer[i] = 0x00;
	break;
    case CRYPT_OAEP:
	//	if (RSA_padding_add_PKCS1_OAEP(abBuffer, dwBufferLen, abData, dwDataLen, NULL, 0) < 1) return false;
	if (!pkcs_1_oaep_encode(abData, dwDataLen, NULL, 0, dwBufferLen << 3, CALG_SHA, abBuffer, &dwBufferLen)) return false;
	break;
    case CRYPT_DESTROYKEY:
	break;
    default: 
	return false;
    }
    return true; 
}

static bool unpad_data(const unsigned char *abData, DWORD dwDataLen, unsigned char *abBuffer, DWORD *dwBufferLen, DWORD dwFlags)
{
    DWORD i;
    int mlen = -1;
    
    switch (dwFlags)
    {
    case CRYPT_SSL2_FALLBACK:
	for (i=2; i<dwDataLen; i++)
	    if (!abData[i])
		break;

	if ((i == dwDataLen) || (*dwBufferLen < dwDataLen - i - 1) ||
	    (abData[0] != 0x00) || (abData[1] != RSAENH_PKC_BLOCKTYPE))
	    return false;

	*dwBufferLen = dwDataLen - i - 1;
	memmove(abBuffer, abData + i + 1, *dwBufferLen);
	break;
    case CRYPT_OAEP:
	if (abData[0] != 0x00) return false;	
	mlen = RSA_padding_check_PKCS1_OAEP(abBuffer, *dwBufferLen, abData + 1, dwDataLen - 1, dwDataLen, NULL, 0);
	*dwBufferLen = mlen;
	if (mlen < 0) return false;
	break;
    case CRYPT_DESTROYKEY:

	break;
    default:
	return false;
    }
    return true;
}

bool init_hash_impl(ALG_ID aiAlgid, HASH_CONTEXT *pHashContext)
{
    switch (aiAlgid) 
    {
    case CALG_MD5:
	MD5Init(&pHashContext->md5);
	break;
        
    case CALG_SHA:
	A_SHAInit(&pHashContext->sha);
	break;
    default:
	return false;
    }

    return true;
}

bool update_hash_impl(ALG_ID aiAlgid, HASH_CONTEXT *pHashContext, const unsigned char *pbData, DWORD dwDataLen) 
{
    switch (aiAlgid)
    {
    case CALG_MD5:
	MD5Update(&pHashContext->md5, pbData, dwDataLen);
	break;
        
    case CALG_SHA:
	A_SHAUpdate(&pHashContext->sha, (unsigned char *)pbData, dwDataLen);
	break;
        
    default:
	return false;
    }

    return true;
}

bool finalize_hash_impl(ALG_ID aiAlgid, HASH_CONTEXT *pHashContext, unsigned char *pbHashValue) 
{
    switch (aiAlgid)
    {
    case CALG_MD5:
	MD5Final(&pHashContext->md5);
	memcpy(pbHashValue, pHashContext->md5.digest, 16);
	break;
        
    case CALG_SHA:
	A_SHAFinal(&pHashContext->sha, (unsigned int *)pbHashValue);
	break;
    default:
	return false;
    }

    return true;
}

bool encrypt_block_impl(ALG_ID aiAlgid, DWORD dwKeySpec, KEY_CONTEXT *pKeyContext, const unsigned char *in, unsigned char *out, DWORD enc) 
{
    unsigned long inlen, outlen;
    unsigned char *in_reversed = NULL;
    
    switch (aiAlgid) {

    case CALG_RSA_KEYX:
    case CALG_RSA_SIGN:
	outlen = inlen = (mp_count_bits(&pKeyContext->rsa.N)+7)/8;
	    
	if (enc) {
	    if (rsa_exptmod(in, inlen, out, &outlen, dwKeySpec, &pKeyContext->rsa) != CRYPT_OK)
		return false;
	    reverse_bytes(out, outlen);
	} else {
	    in_reversed = (unsigned char*)malloc(inlen);
	    if (!in_reversed)
		return false;
    
	    memcpy(in_reversed, in, inlen);
	    reverse_bytes(in_reversed, inlen);
	    if (rsa_exptmod(in_reversed, inlen, out, &outlen, dwKeySpec, &pKeyContext->rsa) != CRYPT_OK) 
	    {
		free(in_reversed);
		return false;
	    }
	    free(in_reversed);
	}
	break;

    default:
	return false;
    }
    return true;
}

bool encrypt_stream_impl(ALG_ID aiAlgid, KEY_CONTEXT *pKeyContext, unsigned char *stream, DWORD dwLen)
{
    switch (aiAlgid) {
        case CALG_RC4:
            rc4_read(stream, dwLen, &pKeyContext->rc4);
            break;

        default:
            return false;
    }

    return true;
}

bool  build_hash_signature(unsigned char *pbSignature, uint dwLen, ALG_ID aiAlgid, const unsigned char *abHashValue, uint dwHashLen, uint dwFlags) 
{
    /* These prefixes are meant to be concatenated with hash values of the
     * respective kind to form a PKCS #7 DigestInfo. */
    static const struct tagOIDDescriptor {
	ALG_ID aiAlgid;
	DWORD dwLen;
	const unsigned char abOID[18];
    } aOIDDescriptor[5] = {
        { CALG_MD2, 18, { 0x30, 0x20, 0x30, 0x0c, 0x06, 0x08, 0x2a, 0x86, 0x48,
                          0x86, 0xf7, 0x0d, 0x02, 0x02, 0x05, 0x00, 0x04, 0x10 } },
        { CALG_MD4, 18, { 0x30, 0x20, 0x30, 0x0c, 0x06, 0x08, 0x2a, 0x86, 0x48, 
                          0x86, 0xf7, 0x0d, 0x02, 0x04, 0x05, 0x00, 0x04, 0x10 } },
        { CALG_MD5, 18, { 0x30, 0x20, 0x30, 0x0c, 0x06, 0x08, 0x2a, 0x86, 0x48,
                          0x86, 0xf7, 0x0d, 0x02, 0x05, 0x05, 0x00, 0x04, 0x10 } },
        { CALG_SHA, 15, { 0x30, 0x21, 0x30, 0x09, 0x06, 0x05, 0x2b, 0x0e, 0x03, 
                          0x02, 0x1a, 0x05, 0x00, 0x04, 0x14 } },
        { 0,        0,  {} }
    };
    uint dwIdxOID, i, j;

    for (dwIdxOID = 0; aOIDDescriptor[dwIdxOID].aiAlgid; dwIdxOID++) {
        if (aOIDDescriptor[dwIdxOID].aiAlgid == aiAlgid) break;
    }

    if (!aOIDDescriptor[dwIdxOID].aiAlgid) 
	return false;

    /* Build the padded signature */
    if (dwFlags & CRYPT_X931_FORMAT) {
        pbSignature[0] = 0x6b;
        for (i=1; i < dwLen - dwHashLen - 3; i++) {
            pbSignature[i] = 0xbb;
        }
        pbSignature[i++] = 0xba;
        for (j=0; j < dwHashLen; j++, i++) {
            pbSignature[i] = abHashValue[j];
        }
        pbSignature[i++] = 0x33;
        pbSignature[i++] = 0xcc;
    } else {
        pbSignature[0] = 0x00;
        pbSignature[1] = 0x01;
        if (dwFlags & CRYPT_NOHASHOID) {
            for (i=2; i < dwLen - 1 - dwHashLen; i++) {
                pbSignature[i] = 0xff;
            }
            pbSignature[i++] = 0x00;
        } else {
            for (i=2; i < dwLen - 1 - aOIDDescriptor[dwIdxOID].dwLen - dwHashLen; i++) {
                pbSignature[i] = 0xff;
            }
            pbSignature[i++] = 0x00;
            for (j=0; j < aOIDDescriptor[dwIdxOID].dwLen; j++) {
                pbSignature[i++] = aOIDDescriptor[dwIdxOID].abOID[j];
            }
        }
        for (j=0; j < dwHashLen; j++) {
            pbSignature[i++] = abHashValue[j];
        }
    }

    return true;
}

bool new_key_impl(ALG_ID aiAlgid, KEY_CONTEXT *pKeyContext, DWORD dwKeyLen) 
{
    switch (aiAlgid)
    {
        case CALG_RSA_KEYX:
        case CALG_RSA_SIGN:
            if (rsa_make_key((int)dwKeyLen, 65537, &pKeyContext->rsa) != CRYPT_OK)
                return false;
            return true;
    }

    return true;
}

bool setup_key_impl(ALG_ID aiAlgid, KEY_CONTEXT *pKeyContext, DWORD dwKeyLen, DWORD dwSaltLen, unsigned char *abKeyValue) 
{
    switch (aiAlgid) 
    {
    case CALG_RC4:
	rc4_start(&pKeyContext->rc4);
	rc4_add_entropy(abKeyValue, dwKeyLen + dwSaltLen, &pKeyContext->rc4);
	rc4_ready(&pKeyContext->rc4);
	break;
    default: break;
    }
    return true;
}

bool import_private_key_impl(const unsigned char *pbSrc, KEY_CONTEXT *pKeyContext, DWORD dwKeyLen, DWORD dwPubExp)
{
    unsigned char *pbTemp, *pbBigNum;

    if (mp_init_multi(&pKeyContext->rsa.e, &pKeyContext->rsa.d, &pKeyContext->rsa.N, 
                      &pKeyContext->rsa.dQ,&pKeyContext->rsa.dP,&pKeyContext->rsa.qP, 
                      &pKeyContext->rsa.p, &pKeyContext->rsa.q, NULL) != MP_OKAY)
    {
        return false;
    }

    pbTemp = (unsigned char*)malloc(2*dwKeyLen+5*((dwKeyLen+1)>>1));
    if (!pbTemp) return false;
    memcpy(pbTemp, pbSrc, 2*dwKeyLen+5*((dwKeyLen+1)>>1));
    pbBigNum = pbTemp;

    pKeyContext->rsa.type = PK_PRIVATE;
    reverse_bytes(pbBigNum, dwKeyLen);
    mp_read_unsigned_bin(&pKeyContext->rsa.N, pbBigNum, dwKeyLen);
    pbBigNum += dwKeyLen;
    reverse_bytes(pbBigNum, (dwKeyLen+1)>>1);
    mp_read_unsigned_bin(&pKeyContext->rsa.p, pbBigNum, (dwKeyLen+1)>>1);
    pbBigNum += (dwKeyLen+1)>>1;
    reverse_bytes(pbBigNum, (dwKeyLen+1)>>1);
    mp_read_unsigned_bin(&pKeyContext->rsa.q, pbBigNum, (dwKeyLen+1)>>1);
    pbBigNum += (dwKeyLen+1)>>1;
    reverse_bytes(pbBigNum, (dwKeyLen+1)>>1);
    mp_read_unsigned_bin(&pKeyContext->rsa.dP, pbBigNum, (dwKeyLen+1)>>1);
    pbBigNum += (dwKeyLen+1)>>1;
    reverse_bytes(pbBigNum, (dwKeyLen+1)>>1);
    mp_read_unsigned_bin(&pKeyContext->rsa.dQ, pbBigNum, (dwKeyLen+1)>>1);
    pbBigNum += (dwKeyLen+1)>>1;
    reverse_bytes(pbBigNum, (dwKeyLen+1)>>1);
    mp_read_unsigned_bin(&pKeyContext->rsa.qP, pbBigNum, (dwKeyLen+1)>>1);
    pbBigNum += (dwKeyLen+1)>>1;
    reverse_bytes(pbBigNum, dwKeyLen);
    mp_read_unsigned_bin(&pKeyContext->rsa.d, pbBigNum, dwKeyLen);
    mp_set_int(&pKeyContext->rsa.e, dwPubExp);
    
    free(pbTemp);
    return true;
}

bool export_private_key_impl(unsigned char *pbDest, KEY_CONTEXT *pKeyContext, DWORD dwKeyLen, DWORD *pdwPubExp)
{
    mp_to_unsigned_bin(&pKeyContext->rsa.N, pbDest);
    reverse_bytes(pbDest, dwKeyLen);
    pbDest += dwKeyLen;
    mp_to_unsigned_bin(&pKeyContext->rsa.p, pbDest);
    reverse_bytes(pbDest, (dwKeyLen+1)>>1);
    pbDest += (dwKeyLen+1)>>1;
    mp_to_unsigned_bin(&pKeyContext->rsa.q, pbDest);
    reverse_bytes(pbDest, (dwKeyLen+1)>>1);
    pbDest += (dwKeyLen+1)>>1;
    mp_to_unsigned_bin(&pKeyContext->rsa.dP, pbDest);
    reverse_bytes(pbDest, (dwKeyLen+1)>>1);
    pbDest += (dwKeyLen+1)>>1;
    mp_to_unsigned_bin(&pKeyContext->rsa.dQ, pbDest);
    reverse_bytes(pbDest, (dwKeyLen+1)>>1);
    pbDest += (dwKeyLen+1)>>1;
    mp_to_unsigned_bin(&pKeyContext->rsa.qP, pbDest);
    reverse_bytes(pbDest, (dwKeyLen+1)>>1);
    pbDest += (dwKeyLen+1)>>1;
    mp_to_unsigned_bin(&pKeyContext->rsa.d, pbDest);
    reverse_bytes(pbDest, dwKeyLen);
    *pdwPubExp = (DWORD)mp_get_int(&pKeyContext->rsa.e);

    return true;
}

bool import_public_key_impl(const unsigned char *pbSrc, KEY_CONTEXT *pKeyContext, uint dwKeyLen, uint dwPubExp)
{
    unsigned char *pbTemp;

    if (mp_init_multi(&pKeyContext->rsa.e, &pKeyContext->rsa.d, &pKeyContext->rsa.N, 
                      &pKeyContext->rsa.dQ,&pKeyContext->rsa.dP,&pKeyContext->rsa.qP, 
                      &pKeyContext->rsa.p, &pKeyContext->rsa.q, NULL) != MP_OKAY)
    {
        return false;
    }

    pbTemp = (unsigned char*)malloc(dwKeyLen);
    if (!pbTemp) return false;
    memcpy(pbTemp, pbSrc, dwKeyLen);
    
    pKeyContext->rsa.type = PK_PUBLIC;
    reverse_bytes(pbTemp, dwKeyLen);
    mp_read_unsigned_bin(&pKeyContext->rsa.N, pbTemp, dwKeyLen);
    free(pbTemp);
    mp_set_int(&pKeyContext->rsa.e, dwPubExp);

    return true;
}

bool export_public_key_impl(unsigned char *pbDest, KEY_CONTEXT *pKeyContext, DWORD dwKeyLen, DWORD *pdwPubExp)
{
    mp_to_unsigned_bin(&pKeyContext->rsa.N, pbDest);
    reverse_bytes(pbDest, dwKeyLen);
    *pdwPubExp = (DWORD)mp_get_int(&pKeyContext->rsa.e);
    return true;
}


static inline void setup_key(CRYPTKEY *pCryptKey) 
{
    pCryptKey->dwState = RSAENH_KEYSTATE_IDLE;
    memcpy(pCryptKey->abChainVector, pCryptKey->abInitVector, sizeof(pCryptKey->abChainVector));
    setup_key_impl(pCryptKey->aiAlgid, &pCryptKey->context, pCryptKey->dwKeyLen, 
                   pCryptKey->dwSaltLen, pCryptKey->abKeyValue);
}

static inline bool init_hash(CRYPTHASH *pCryptHash) 
{
    return init_hash_impl(pCryptHash->aiAlgid, &pCryptHash->context);
}

static inline void update_hash(CRYPTHASH *pCryptHash, const unsigned char *pbData, DWORD dwDataLen) 
{
    update_hash_impl(pCryptHash->aiAlgid, &pCryptHash->context, pbData, dwDataLen);
}

static inline void finalize_hash(CRYPTHASH *pCryptHash) 
{
    finalize_hash_impl(pCryptHash->aiAlgid, &pCryptHash->context, pCryptHash->abHashValue);
}

static inline bool copy_param(unsigned char *pbBuffer, DWORD *pdwBufferSize, const unsigned char *pbParam, DWORD dwParamSize) 
{
    if (pbBuffer) 
    {
        if (dwParamSize > *pdwBufferSize) 
        {
            *pdwBufferSize = dwParamSize;
            return false;
        }
        memcpy(pbBuffer, pbParam, dwParamSize);
    }
    *pdwBufferSize = dwParamSize;
    return true;
}

void CryptDestroyKey(CRYPTKEY *pCryptKey)
{
    KEY_CONTEXT *pKeyContext;
    
    if (!pCryptKey) return;
	
    switch(GET_ALG_TYPE(pCryptKey->aiAlgid))
    {
    case ALG_TYPE_RSA:
	pKeyContext = &pCryptKey->context;
	rsa_free(&pKeyContext->rsa);
	break;
    case ALG_TYPE_BLOCK:
	break;
    case ALG_TYPE_STREAM:
	break;
    default: break;
    }
    free(pCryptKey);
}

bool new_key(ALG_ID aiAlgid, DWORD dwFlags, CRYPTKEY **ppCryptKey)
{
    CRYPTKEY *pCryptKey;
    DWORD dwKeyLen = HIWORD(dwFlags);

    const PROV_ENUMALGS_EX *peaAlgidInfo;

    *ppCryptKey = NULL; 

    /* 
     * Retrieve the CSP's capabilities for the given ALG_ID value
     */
    peaAlgidInfo = get_algid_info(aiAlgid);
    if (!peaAlgidInfo) return false;

    /*
     * Assume the default key length, if none is specified explicitly
     */
    if (dwKeyLen == 0) dwKeyLen = peaAlgidInfo->dwDefaultLen;
    /*
     * Check if the requested key length is supported by the current CSP.
     */
    if (dwKeyLen % 8 || dwKeyLen > peaAlgidInfo->dwMaxLen || dwKeyLen < peaAlgidInfo->dwMinLen) 
	return false;

    pCryptKey = (CRYPTKEY*)malloc(sizeof(CRYPTKEY));

    if (pCryptKey)
    {
	pCryptKey->aiAlgid = aiAlgid;
	pCryptKey->dwModeBits = 0;
	pCryptKey->dwPermissions = CRYPT_ENCRYPT | CRYPT_DECRYPT | CRYPT_READ | CRYPT_WRITE | CRYPT_MAC;
	pCryptKey->dwKeyLen = dwKeyLen >> 3;
	if ((dwFlags & CRYPT_CREATE_SALT) || (dwKeyLen == 40 && !(dwFlags & CRYPT_NO_SALT))) 
	    pCryptKey->dwSaltLen = 16 /*FIXME*/ - pCryptKey->dwKeyLen;
	else
	    pCryptKey->dwSaltLen = 0;
	memset(pCryptKey->abKeyValue, 0, sizeof(pCryptKey->abKeyValue));
	memset(pCryptKey->abInitVector, 0, sizeof(pCryptKey->abInitVector));
            
	switch(aiAlgid)
	{
	case CALG_SSL2_MASTER:
	case CALG_RC4:
	    pCryptKey->dwBlockLen = 0;
	    pCryptKey->dwMode = 0;
	    break;

	case CALG_RSA_KEYX:
	case CALG_RSA_SIGN:
	    pCryptKey->dwBlockLen = dwKeyLen >> 3;
	    pCryptKey->dwMode = 0;
	    break;
	}

	*ppCryptKey = pCryptKey;
    }

    return true;
}

bool CryptExportKey(CRYPTKEY *pCryptKey, CRYPTKEY *pPubKey, DWORD dwBlobType, DWORD dwFlags, unsigned char *pbData, DWORD *pdwDataLen)
{
    BLOBHEADER *pBlobHeader = (BLOBHEADER*)pbData;
    RSAPUBKEY *pRSAPubKey = (RSAPUBKEY*)(pBlobHeader+1);
    ALG_ID *pAlgid = (ALG_ID*)(pBlobHeader+1);
    DWORD dwDataLen;
    
    if (!pCryptKey) return false; 
	
    if (dwFlags & CRYPT_SSL2_FALLBACK)
	if (pCryptKey->aiAlgid != CALG_SSL2_MASTER) return false;
    
    switch ((unsigned char)dwBlobType)
    {
    case SIMPLEBLOB:
	if(!pPubKey) return false;

	if (!(GET_ALG_CLASS(pCryptKey->aiAlgid)&(ALG_CLASS_DATA_ENCRYPT|ALG_CLASS_MSG_ENCRYPT)))
	    return false;

	dwDataLen = sizeof(BLOBHEADER) + sizeof(ALG_ID) + pPubKey->dwBlockLen;
	if (pbData) {
	    if (*pdwDataLen < dwDataLen) {
		*pdwDataLen = dwDataLen;
		return false;
	    }

	    pBlobHeader->bType = SIMPLEBLOB;
	    pBlobHeader->bVersion = CUR_BLOB_VERSION;
	    pBlobHeader->reserved = 0;
	    pBlobHeader->aiKeyAlg = pCryptKey->aiAlgid;

	    *pAlgid = pPubKey->aiAlgid;

	    if (!pad_data(pCryptKey->abKeyValue, pCryptKey->dwKeyLen, (unsigned char*)(pAlgid+1), pPubKey->dwBlockLen, dwFlags))
	    {
		return false;
	    }

	    encrypt_block_impl(pPubKey->aiAlgid, PK_PUBLIC, &pPubKey->context, (unsigned char*)(pAlgid+1), (unsigned char*)(pAlgid+1), RSAENH_ENCRYPT);
	}
	*pdwDataLen = dwDataLen;
	return true;
            
        case PUBLICKEYBLOB:

            if ((pCryptKey->aiAlgid != CALG_RSA_KEYX) && (pCryptKey->aiAlgid != CALG_RSA_SIGN))
                return false;

            dwDataLen = sizeof(BLOBHEADER) + sizeof(RSAPUBKEY) + pCryptKey->dwKeyLen;
            if (pbData) {
                if (*pdwDataLen < dwDataLen) {
                    *pdwDataLen = dwDataLen;
                    return false;
                }

                pBlobHeader->bType = PUBLICKEYBLOB;
                pBlobHeader->bVersion = CUR_BLOB_VERSION;
                pBlobHeader->reserved = 0;
                pBlobHeader->aiKeyAlg = pCryptKey->aiAlgid;

                pRSAPubKey->magic = RSAENH_MAGIC_RSA1; 
                pRSAPubKey->bitlen = pCryptKey->dwKeyLen << 3;
        
                export_public_key_impl((unsigned char*)(pRSAPubKey+1), &pCryptKey->context, pCryptKey->dwKeyLen, &pRSAPubKey->pubexp);
            }
            *pdwDataLen = dwDataLen;
            return true;

        case PRIVATEKEYBLOB:
            if ((pCryptKey->aiAlgid != CALG_RSA_KEYX) && (pCryptKey->aiAlgid != CALG_RSA_SIGN))
                return false;
    
            dwDataLen = sizeof(BLOBHEADER) + sizeof(RSAPUBKEY) + 2 * pCryptKey->dwKeyLen + 5 * ((pCryptKey->dwKeyLen + 1) >> 1);
            if (pbData) {
                if (*pdwDataLen < dwDataLen) {
                    *pdwDataLen = dwDataLen;
                    return false;
                }
                
                pBlobHeader->bType = PRIVATEKEYBLOB;
                pBlobHeader->bVersion = CUR_BLOB_VERSION;
                pBlobHeader->reserved = 0;
                pBlobHeader->aiKeyAlg = pCryptKey->aiAlgid;

                pRSAPubKey->magic = RSAENH_MAGIC_RSA2;
                pRSAPubKey->bitlen = pCryptKey->dwKeyLen << 3;
                
                export_private_key_impl((unsigned char*)(pRSAPubKey+1), &pCryptKey->context, pCryptKey->dwKeyLen, &pRSAPubKey->pubexp);
            }
            *pdwDataLen = dwDataLen;
            return true;
            
        default:
            return false;
    }
}

bool CryptImportKey(const unsigned char *pbData, DWORD dwDataLen,  CRYPTKEY *pPubKey, DWORD dwFlags, CRYPTKEY **ppKey)
{
    CRYPTKEY *pCryptKey;
    const BLOBHEADER *pBlobHeader = (const BLOBHEADER*)pbData;
    const RSAPUBKEY *pRSAPubKey = (const RSAPUBKEY*)(pBlobHeader+1);
    const ALG_ID *pAlgid = (const ALG_ID*)(pBlobHeader+1);
    const unsigned char *pbKeyStream = (const unsigned char*)(pAlgid + 1);
    ALG_ID algID;
    unsigned char *pbDecrypted;
    DWORD dwKeyLen;

    *ppKey = NULL;

    if (dwDataLen < sizeof(BLOBHEADER) || pBlobHeader->bVersion != CUR_BLOB_VERSION ||pBlobHeader->reserved != 0) 
	return false;
    switch (pBlobHeader->bType)
    {
    case PRIVATEKEYBLOB:    
	if ((dwDataLen < sizeof(BLOBHEADER) + sizeof(RSAPUBKEY)) || 
	    (pRSAPubKey->magic != RSAENH_MAGIC_RSA2) ||
	    (dwDataLen < sizeof(BLOBHEADER) + sizeof(RSAPUBKEY) + 
	     (2 * pRSAPubKey->bitlen >> 3) + (5 * ((pRSAPubKey->bitlen+8)>>4)))) 
	{
	    return false;
	}
    
	new_key(pBlobHeader->aiKeyAlg, MAKELONG(0,pRSAPubKey->bitlen), &pCryptKey);
	if (!pCryptKey) return false;
	setup_key(pCryptKey);
	if (import_private_key_impl((const unsigned char*)(pRSAPubKey+1), &pCryptKey->context, pRSAPubKey->bitlen/8, pRSAPubKey->pubexp))
	{
	    *ppKey = pCryptKey;
	    return true;
	} else return false;
                
    case PUBLICKEYBLOB:
	if ((dwDataLen < sizeof(BLOBHEADER) + sizeof(RSAPUBKEY)) || (pRSAPubKey->magic != RSAENH_MAGIC_RSA1) ||(dwDataLen < sizeof(BLOBHEADER) + sizeof(RSAPUBKEY) + (pRSAPubKey->bitlen >> 3))) 
	    return false;
    
            /* Since this is a public key blob, only the public key is
             * available, so only signature verification is possible.
             */
	algID = pBlobHeader->aiKeyAlg;
	if (algID == CALG_RSA_KEYX)
	    algID = CALG_RSA_SIGN;
	new_key(algID, MAKELONG(0,pRSAPubKey->bitlen), &pCryptKey); 
	if (!pCryptKey) return false;
	setup_key(pCryptKey);
	if (import_public_key_impl((const unsigned char*)(pRSAPubKey+1), &pCryptKey->context, pRSAPubKey->bitlen >> 3, pRSAPubKey->pubexp))
	{
	    *ppKey = pCryptKey;
	    return true;
	} else return false;
                
    case SIMPLEBLOB:
	if (!pPubKey || pPubKey->aiAlgid != CALG_RSA_KEYX)
	    return false;
	if (dwDataLen < sizeof(BLOBHEADER)+sizeof(ALG_ID)+pPubKey->dwBlockLen) 
	    return false;

	pbDecrypted = (unsigned char*)malloc(pPubKey->dwBlockLen);
	if (!pbDecrypted) 
	    return false;
	encrypt_block_impl(pPubKey->aiAlgid, PK_PRIVATE, &pPubKey->context, pbKeyStream, pbDecrypted, RSAENH_DECRYPT);

	dwKeyLen = RSAENH_MAX_KEY_SIZE;

	if (!unpad_data(pbDecrypted, pPubKey->dwBlockLen, pbDecrypted, &dwKeyLen, dwFlags)) 
	{
	    free(pbDecrypted);
	    return false;
	}
	new_key(pBlobHeader->aiKeyAlg, dwKeyLen<<19, &pCryptKey);
	if (!pCryptKey)
	{
	    free(pbDecrypted);
	    return false;
	}
	memcpy(pCryptKey->abKeyValue, pbDecrypted, dwKeyLen);
	free(pbDecrypted);
	setup_key(pCryptKey);
	*ppKey = pCryptKey;
	return true;

    default:
	return false;
    }
}

bool CryptGenKey (ALG_ID Algid, DWORD dwFlags, CRYPTKEY **ppKey)
{
    CRYPTKEY *pCryptKey;

    *ppKey = NULL;

    switch (Algid)
    {
    case AT_SIGNATURE:
    case CALG_RSA_SIGN:
	new_key(CALG_RSA_SIGN, dwFlags, &pCryptKey);
	if (pCryptKey) { 
	    new_key_impl(pCryptKey->aiAlgid, &pCryptKey->context, pCryptKey->dwKeyLen);
	    setup_key(pCryptKey);
	}
	break;

    case AT_KEYEXCHANGE:
    case CALG_RSA_KEYX:
	new_key(CALG_RSA_KEYX, dwFlags, &pCryptKey);
	if (pCryptKey) { 
	    new_key_impl(pCryptKey->aiAlgid, &pCryptKey->context, pCryptKey->dwKeyLen);
	    setup_key(pCryptKey);
	}
	break;
            
    case CALG_RC4:
    case CALG_SSL2_MASTER:
	new_key(Algid, dwFlags, &pCryptKey);
	if (pCryptKey) {
	    gen_rand(pCryptKey->abKeyValue, pCryptKey->dwKeyLen);
//	    gen_rand(pCryptKey->abKeyValue, RSAENH_MAX_KEY_SIZE);
	    setup_key(pCryptKey);
	}
	break;

    default:
	/* MSDN: Algorithm not supported specified by Algid */
	return false;
    }

    *ppKey = pCryptKey;

    return true;
}

bool CryptCreateHash(ALG_ID Algid, DWORD dwFlags, CRYPTHASH **ppHash)
{
    CRYPTHASH *pCryptHash;
    const PROV_ENUMALGS_EX *peaAlgidInfo;
    
    *ppHash = NULL;
    
    peaAlgidInfo = get_algid_info(Algid);
    if (!peaAlgidInfo) 
	return false;
    
    if (dwFlags)
        return false;

    pCryptHash = (CRYPTHASH*)malloc(sizeof(CRYPTHASH));

    if (pCryptHash) 
    {
	pCryptHash->aiAlgid = Algid;
	pCryptHash->dwState = RSAENH_HASHSTATE_IDLE;
	pCryptHash->dwHashSize = peaAlgidInfo->dwDefaultLen >> 3;

	if (!init_hash(pCryptHash)) 
	    return false;
	
	    *ppHash = pCryptHash;
	
	return true;
    }
    return false;
}

void CryptDestroyHash(CRYPTHASH *pCryptHash)
{
    free(pCryptHash);    
}

bool CryptHashData(CRYPTHASH *pCryptHash, const unsigned char *pbData, DWORD dwDataLen, DWORD dwFlags)
{
    if (dwFlags)
    {
        return false;
    }

    if (pCryptHash->dwState == RSAENH_HASHSTATE_IDLE)
        pCryptHash->dwState = RSAENH_HASHSTATE_HASHING;
    
    if (pCryptHash->dwState != RSAENH_HASHSTATE_HASHING)
    {
        return false;
    }

    update_hash(pCryptHash, pbData, dwDataLen);
    return true;
}

bool CryptGetHashParam(CRYPTHASH *pCryptHash, DWORD dwParam, unsigned char *pbData, DWORD *pdwDataLen, DWORD dwFlags) 
{
    if (dwFlags)
    {
	return false;
    }
    
    if (!pdwDataLen)
    {
	return false;
    }
    
    switch (dwParam)
    {
    case HP_ALGID:
	return copy_param(pbData, pdwDataLen, (unsigned char*)&pCryptHash->aiAlgid, sizeof(ALG_ID));

    case HP_HASHSIZE:
	return copy_param(pbData, pdwDataLen, (unsigned char*)&pCryptHash->dwHashSize, sizeof(DWORD));

    case HP_HASHVAL:
	if (pCryptHash->dwState == RSAENH_HASHSTATE_IDLE) return false;

	if (pbData && (pCryptHash->dwState != RSAENH_HASHSTATE_FINISHED))
	{
	    finalize_hash(pCryptHash);
	    pCryptHash->dwState = RSAENH_HASHSTATE_FINISHED;
	}
	
	return copy_param(pbData, pdwDataLen, (unsigned char*)pCryptHash->abHashValue, pCryptHash->dwHashSize);
    default:
	return false;
    }
}

bool CryptSignHash(CRYPTKEY *pCryptKey, CRYPTHASH *pCryptHash, DWORD dwFlags, unsigned char *pbSignature, DWORD *pdwSigLen)
{
    DWORD dwHashLen;
    unsigned char abHashValue[RSAENH_MAX_HASH_SIZE];
    ALG_ID aiAlgid;

    if (!pCryptKey) return false;

    if (dwFlags & ~(CRYPT_NOHASHOID|CRYPT_X931_FORMAT)) return false;
    
    if (!pbSignature) 
    {
	*pdwSigLen = pCryptKey->dwKeyLen;
	return true;
    }
    if (pCryptKey->dwKeyLen > *pdwSigLen)
    {
	*pdwSigLen = pCryptKey->dwKeyLen;
	return false;
    }
    *pdwSigLen = pCryptKey->dwKeyLen;

    dwHashLen = sizeof(DWORD);
    if (!CryptGetHashParam(pCryptHash, HP_ALGID, (unsigned char*)&aiAlgid, &dwHashLen, 0)) return false;
    dwHashLen = RSAENH_MAX_HASH_SIZE;
    if (!CryptGetHashParam(pCryptHash, HP_HASHVAL, abHashValue, &dwHashLen, 0)) return false;
    
    if (!build_hash_signature(pbSignature, *pdwSigLen, aiAlgid, abHashValue, dwHashLen, dwFlags))
	return false;
    
    return encrypt_block_impl(pCryptKey->aiAlgid, PK_PRIVATE, &pCryptKey->context, pbSignature, pbSignature, RSAENH_ENCRYPT);
}

bool CryptVerifySignature(CRYPTKEY *pCryptKey, CRYPTHASH *pHash, const unsigned char *pbSignature, DWORD dwSigLen, CRYPTKEY *pPubKey, DWORD dwFlags)
{
    unsigned char *pbConstructed = NULL, *pbDecrypted = NULL;
    DWORD dwHashLen;
    ALG_ID aiAlgid;
    unsigned char abHashValue[RSAENH_MAX_HASH_SIZE];
    bool res = false;

    if (dwFlags & ~(CRYPT_NOHASHOID|CRYPT_X931_FORMAT))
	return false;

    if (!pCryptKey) return false;
    
    dwHashLen = sizeof(DWORD);
    if (!CryptGetHashParam(pHash, HP_ALGID, (unsigned char*)&aiAlgid, &dwHashLen, 0)) return false;
    
    dwHashLen = RSAENH_MAX_HASH_SIZE;
    if (!CryptGetHashParam(pHash, HP_HASHVAL, abHashValue, &dwHashLen, 0)) return false;

    pbConstructed = (unsigned char*)malloc(dwSigLen);
    if (!pbConstructed)
	goto cleanup;

    pbDecrypted = (unsigned char*)malloc(dwSigLen);
    if (!pbDecrypted)
	goto cleanup;

    if (!encrypt_block_impl(pCryptKey->aiAlgid, PK_PUBLIC, &pCryptKey->context, pbSignature, pbDecrypted, RSAENH_DECRYPT)) 
	goto cleanup;

    if (!build_hash_signature(pbConstructed, dwSigLen, aiAlgid, abHashValue, dwHashLen, dwFlags)) 
	goto cleanup;

    if (memcmp(pbDecrypted, pbConstructed, dwSigLen)) 
	goto cleanup;

    res = true;
cleanup:
    free(pbConstructed);
    free(pbDecrypted);
    return res;
}

bool CryptEncrypt(CRYPTKEY *pCryptKey, CRYPTHASH *pHash, bool Final, DWORD dwFlags, unsigned char *pbData, DWORD *pdwDataLen, DWORD dwBufLen)
{
    unsigned char *in, out[RSAENH_MAX_BLOCK_SIZE], o[RSAENH_MAX_BLOCK_SIZE];
    DWORD dwEncryptedLen, i, j, k;

    if (dwFlags)
    {
        return false;
    }

    if (!pCryptKey)
    {
        return false;
    }

    if (pCryptKey->dwState == RSAENH_KEYSTATE_IDLE) 
        pCryptKey->dwState = RSAENH_KEYSTATE_ENCRYPTING;

    if (pCryptKey->dwState != RSAENH_KEYSTATE_ENCRYPTING) 
    {
        return false;
    }

    if (pHash) {
        if (!CryptHashData(pHash, pbData, *pdwDataLen, 0)) return false;
    }
    
    if (GET_ALG_TYPE(pCryptKey->aiAlgid) == ALG_TYPE_BLOCK) {
        if (!Final && (*pdwDataLen % pCryptKey->dwBlockLen)) {
            return false;
        }

        dwEncryptedLen = (*pdwDataLen/pCryptKey->dwBlockLen+(Final?1:0))*pCryptKey->dwBlockLen;
        for (i=*pdwDataLen; i<dwEncryptedLen && i<dwBufLen; i++) pbData[i] = dwEncryptedLen - *pdwDataLen;
        *pdwDataLen = dwEncryptedLen; 

        if (*pdwDataLen > dwBufLen) 
        {
            return false;
        }
    
        for (i=0, in=pbData; i<*pdwDataLen; i+=pCryptKey->dwBlockLen, in+=pCryptKey->dwBlockLen) {
            switch (pCryptKey->dwMode) {
                case CRYPT_MODE_ECB:
                    encrypt_block_impl(pCryptKey->aiAlgid, 0, &pCryptKey->context, in, out, RSAENH_ENCRYPT);
                    break;
                
                case CRYPT_MODE_CBC:
                    for (j=0; j<pCryptKey->dwBlockLen; j++) in[j] ^= pCryptKey->abChainVector[j];
                    encrypt_block_impl(pCryptKey->aiAlgid, 0, &pCryptKey->context, in, out, RSAENH_ENCRYPT);
                    memcpy(pCryptKey->abChainVector, out, pCryptKey->dwBlockLen);
                    break;

                case CRYPT_MODE_CFB:
                    for (j=0; j<pCryptKey->dwBlockLen; j++) {
                        encrypt_block_impl(pCryptKey->aiAlgid, 0, &pCryptKey->context, pCryptKey->abChainVector, o, RSAENH_ENCRYPT);
                        out[j] = in[j] ^ o[0];
                        for (k=0; k<pCryptKey->dwBlockLen-1; k++) 
                            pCryptKey->abChainVector[k] = pCryptKey->abChainVector[k+1];
                        pCryptKey->abChainVector[k] = out[j];
                    }
                    break;
                    
                default:
                    return false;
            }
            memcpy(in, out, pCryptKey->dwBlockLen);
        }
    } else if (GET_ALG_TYPE(pCryptKey->aiAlgid) == ALG_TYPE_STREAM) {
        encrypt_stream_impl(pCryptKey->aiAlgid, &pCryptKey->context, pbData, *pdwDataLen);
    } else if (GET_ALG_TYPE(pCryptKey->aiAlgid) == ALG_TYPE_RSA) {
        if (pCryptKey->aiAlgid == CALG_RSA_SIGN) {
            return false;
        }
        if (dwBufLen < pCryptKey->dwBlockLen) {
            return false;
        }
        if (!pad_data(pbData, *pdwDataLen, pbData, pCryptKey->dwBlockLen, dwFlags)) return false;
        encrypt_block_impl(pCryptKey->aiAlgid, PK_PUBLIC, &pCryptKey->context, pbData, pbData, RSAENH_ENCRYPT);
        *pdwDataLen = pCryptKey->dwBlockLen;
        Final = true;
    } else {
        return false;
    }

    if (Final) setup_key(pCryptKey);

    return true;
}

bool CryptDecrypt(CRYPTKEY *pCryptKey, CRYPTHASH *pHash, bool Final, DWORD dwFlags, unsigned char *pbData, DWORD *pdwDataLen)
{
    unsigned char *in, out[RSAENH_MAX_BLOCK_SIZE], o[RSAENH_MAX_BLOCK_SIZE];
    DWORD i, j, k;
    DWORD dwMax;

    if (dwFlags)
    {
        return false;
    }

    if (!pCryptKey)
    {
        return false;
    }

    if (pCryptKey->dwState == RSAENH_KEYSTATE_IDLE) 
        pCryptKey->dwState = RSAENH_KEYSTATE_DECRYPTING;

    if (pCryptKey->dwState != RSAENH_KEYSTATE_DECRYPTING)
    {
        return false;
    }

    dwMax=*pdwDataLen;

    if (GET_ALG_TYPE(pCryptKey->aiAlgid) == ALG_TYPE_BLOCK) {
        for (i=0, in=pbData; i<*pdwDataLen; i+=pCryptKey->dwBlockLen, in+=pCryptKey->dwBlockLen) {
            switch (pCryptKey->dwMode) {
                case CRYPT_MODE_ECB:
                    encrypt_block_impl(pCryptKey->aiAlgid, 0, &pCryptKey->context, in, out, RSAENH_DECRYPT);
                    break;
                
                case CRYPT_MODE_CBC:
                    encrypt_block_impl(pCryptKey->aiAlgid, 0, &pCryptKey->context, in, out, RSAENH_DECRYPT);
                    for (j=0; j<pCryptKey->dwBlockLen; j++) out[j] ^= pCryptKey->abChainVector[j];
                    memcpy(pCryptKey->abChainVector, in, pCryptKey->dwBlockLen);
                    break;

                case CRYPT_MODE_CFB:
                    for (j=0; j<pCryptKey->dwBlockLen; j++) {
                        encrypt_block_impl(pCryptKey->aiAlgid, 0, &pCryptKey->context, pCryptKey->abChainVector, o, RSAENH_ENCRYPT);
                        out[j] = in[j] ^ o[0];
                        for (k=0; k<pCryptKey->dwBlockLen-1; k++) 
                            pCryptKey->abChainVector[k] = pCryptKey->abChainVector[k+1];
                        pCryptKey->abChainVector[k] = in[j];
                    }
                    break;
                    
                default:
                    return false;
            }
            memcpy(in, out, pCryptKey->dwBlockLen);
        }
        if (Final) *pdwDataLen -= pbData[*pdwDataLen-1]; 

    } else if (GET_ALG_TYPE(pCryptKey->aiAlgid) == ALG_TYPE_STREAM) {
        encrypt_stream_impl(pCryptKey->aiAlgid, &pCryptKey->context, pbData, *pdwDataLen);
    } else if (GET_ALG_TYPE(pCryptKey->aiAlgid) == ALG_TYPE_RSA) {
        if (pCryptKey->aiAlgid == CALG_RSA_SIGN) {
            return false;
        }
        encrypt_block_impl(pCryptKey->aiAlgid, PK_PRIVATE, &pCryptKey->context, pbData, pbData, RSAENH_DECRYPT);
        if (!unpad_data(pbData, pCryptKey->dwBlockLen, pbData, pdwDataLen, dwFlags)) return false;
        Final = true;
    } else {
        return false;
    } 
    
    if (Final) setup_key(pCryptKey);

    if (pHash) {
        if (*pdwDataLen>dwMax || !CryptHashData(pHash, pbData, *pdwDataLen, 0)) return false;
    }
    
    return true;
}

