#ifndef __CRYPTO_H
#define __CRYPTO_H

#include "crypt_md5.h"
#include "crypt_sha.h"
#include "tomcrypt.h"

typedef unsigned int 	ALG_ID;
typedef unsigned char 	BYTE;
typedef char 		CHAR;
typedef unsigned int 	DWORD;
typedef unsigned short 	WORD;
typedef long 		LONG;

#define STORE32H(x, y) { (y)[0] = (unsigned char)(((x)>>24)&255); (y)[1] = (unsigned char)(((x)>>16)&255); (y)[2] = (unsigned char)(((x)>>8)&255); (y)[3] = (unsigned char)((x)&255); } 

#define HIWORD(l) 		((WORD)((DWORD_PTR)(l) >> 16))
#define MAKELONG(low,high) 	((LONG)(((WORD)((DWORD_PTR)(low) & 0xFFFF)) | ((DWORD)((WORD)((DWORD_PTR)(high) & 0xFFFF))) << 16))

#define GET_ALG_CLASS(x)	(x & (7 << 13))
#define GET_ALG_TYPE(x)	(x & (15 << 9))
#define GET_ALG_SID(x)	(x & (511))

#define ALG_CLASS_HASH		(4 << 13)
#define ALG_CLASS_SIGNATURE	(1 << 13)
#define ALG_CLASS_MSG_ENCRYPT	(2 << 13)
#define ALG_CLASS_KEY_EXCHANGE	(5 << 13)
#define ALG_CLASS_DATA_ENCRYPT	(3 << 13)

#define ALG_TYPE_ANY		(0)
#define ALG_TYPE_RSA		(2 << 9)
#define ALG_TYPE_BLOCK		(3 << 9)
#define ALG_TYPE_STREAM		(4 << 9)
#define ALG_TYPE_SECURECHANNEL	(6 << 9)

#define ALG_SID_MD2 		1
#define ALG_SID_MD4 		2
#define ALG_SID_MD5 		3
#define ALG_SID_SHA 		4
#define ALG_SID_RSA_ANY 		0
#define ALG_SID_RC4 		1
#define ALG_SID_SSL2_MASTER 	5

#define CALG_MD2                  (ALG_CLASS_HASH         | ALG_TYPE_ANY           | ALG_SID_MD2)
#define CALG_MD4                  (ALG_CLASS_HASH         | ALG_TYPE_ANY           | ALG_SID_MD4)
#define CALG_MD5                  (ALG_CLASS_HASH         | ALG_TYPE_ANY           | ALG_SID_MD5)
#define CALG_SHA                  (ALG_CLASS_HASH         | ALG_TYPE_ANY           | ALG_SID_SHA)
#define CALG_RSA_SIGN             (ALG_CLASS_SIGNATURE    | ALG_TYPE_RSA           | ALG_SID_RSA_ANY)
#define CALG_RSA_KEYX             (ALG_CLASS_KEY_EXCHANGE | ALG_TYPE_RSA           | ALG_SID_RSA_ANY)
#define CALG_RC4	(ALG_CLASS_DATA_ENCRYPT | ALG_TYPE_STREAM | ALG_SID_RC4)
#define CALG_SSL2_MASTER	(ALG_CLASS_MSG_ENCRYPT  | ALG_TYPE_SECURECHANNEL | ALG_SID_SSL2_MASTER)

#define CRYPT_NOHASHOID         0x00000001
#define CRYPT_TYPE2_FORMAT      0x00000002
#define CRYPT_X931_FORMAT       0x00000004

#define CRYPT_ENCRYPT           0x0001 
#define CRYPT_DECRYPT           0x0002
#define CRYPT_EXPORT            0x0004
#define CRYPT_READ              0x0008
#define CRYPT_WRITE             0x0010
#define CRYPT_MAC               0x0020

#define CRYPT_MODE_CBC 		1
#define CRYPT_MODE_ECB 		2
#define CRYPT_MODE_OFB 		3
#define CRYPT_MODE_CFB 		4

#define CRYPT_EXPORTABLE        0x00000001
#define CRYPT_USER_PROTECTED    0x00000002
#define CRYPT_CREATE_SALT       0x00000004
#define CRYPT_UPDATE_KEY        0x00000008
#define CRYPT_NO_SALT           0x00000010
#define CRYPT_PREGEN            0x00000040
#define CRYPT_SERVER            0x00000400
#define CRYPT_ARCHIVABLE        0x00004000

#define CRYPT_SSL2_FALLBACK     0x00000002
#define CRYPT_DESTROYKEY        0x00000004
#define CRYPT_OAEP              0x00000040

#define CRYPT_FLAG_PCT1    0x0001
#define CRYPT_FLAG_SSL2    0x0002
#define CRYPT_FLAG_SSL3    0x0004
#define CRYPT_FLAG_TLS1    0x0008
#define CRYPT_FLAG_IPSEC   0x0010
#define CRYPT_FLAG_SIGNING 0x0020

#define RSAENH_ENCRYPT 			1
#define RSAENH_DECRYPT 			0
#define RSAENH_MAGIC_RSA2 		0x32415352
#define RSAENH_MAGIC_RSA1 		0x31415352
#define RSAENH_MAX_KEY_SIZE 		48
#define RSAENH_MAX_BLOCK_SIZE 		24
#define RSAENH_MAX_HASH_SIZE 		104
#define RSAENH_HASHSTATE_IDLE 		0
#define RSAENH_HASHSTATE_HASHING 	1
#define RSAENH_HASHSTATE_FINISHED 	2
#define RSAENH_KEYSTATE_IDLE 		0
#define RSAENH_KEYSTATE_ENCRYPTING 	1
#define RSAENH_KEYSTATE_DECRYPTING 	2
#define RSAENH_KEYSTATE_MASTERKEY 	3
#define RSAENH_PKC_BLOCKTYPE 		0x02

#define RSAENH_PERSONALITY_BASE        0u

#define RSAENH_MAX_ENUMALGS 20
#define RSAENH_PCT1_SSL2_SSL3_TLS1 (CRYPT_FLAG_PCT1|CRYPT_FLAG_SSL2|CRYPT_FLAG_SSL3|CRYPT_FLAG_TLS1)

#define HP_ALGID                0x0001
#define HP_HASHVAL              0x0002
#define HP_HASHSIZE             0x0004

#define SIMPLEBLOB              0x1
#define PUBLICKEYBLOB           0x6
#define PRIVATEKEYBLOB          0x7
#define PLAINTEXTKEYBLOB        0x8
#define OPAQUEKEYBLOB           0x9
#define PUBLICKEYBLOBEX         0xA
#define SYMMETRICWRAPKEYBLOB    0xB

#define CUR_BLOB_VERSION 		2

#define AT_KEYEXCHANGE 		1
#define AT_SIGNATURE 		2

typedef union tagKEY_CONTEXT {
    rc2_key rc2;
    des_key des;
    des3_key des3;
    prng_state rc4;
    rsa_key rsa;
} KEY_CONTEXT;

typedef union tagHASH_CONTEXT {
    MD5_CTX md5;
    SHA_CTX sha;
} HASH_CONTEXT;

typedef struct _PUBLICKEYSTRUC {
    BYTE   bType;
    BYTE   bVersion;
    WORD   reserved;
    ALG_ID aiKeyAlg;
} BLOBHEADER, PUBLICKEYSTRUC;

typedef struct _RSAPUBKEY {
    DWORD   magic;
    DWORD   bitlen;
    DWORD   pubexp;
} RSAPUBKEY;

typedef struct tagCRYPTKEY
{
    ALG_ID			aiAlgid;
    DWORD		dwMode;
    DWORD		dwModeBits;
    DWORD		dwPermissions;
    DWORD		dwKeyLen;
    DWORD		dwSaltLen;
    DWORD		dwBlockLen;
    DWORD		dwState;
    KEY_CONTEXT		context;
    unsigned char	abKeyValue[RSAENH_MAX_KEY_SIZE];
    unsigned char	abInitVector[RSAENH_MAX_BLOCK_SIZE];
    unsigned char	abChainVector[RSAENH_MAX_BLOCK_SIZE];
} CRYPTKEY;

typedef struct tagCRYPTHASH
{
    ALG_ID		aiAlgid;
    DWORD		dwHashSize;
    DWORD		dwState;
    HASH_CONTEXT	context;
    unsigned char	abHashValue[RSAENH_MAX_HASH_SIZE];
} CRYPTHASH;

typedef struct _PROV_ENUMALGS_EX {
  ALG_ID aiAlgid;
  DWORD  dwDefaultLen;
  DWORD  dwMinLen;
  DWORD  dwMaxLen;
  DWORD  dwProtocols;
  DWORD  dwNameLen;
  CHAR   szName[20];
  DWORD  dwLongNameLen;
  CHAR   szLongName[40];
} PROV_ENUMALGS_EX;

bool init_hash_impl(ALG_ID aiAlgid, HASH_CONTEXT *pHashContext);
bool update_hash_impl(ALG_ID aiAlgid, HASH_CONTEXT *pHashContext, const unsigned char *pbData, DWORD dwDataLen);
bool finalize_hash_impl(ALG_ID aiAlgid, HASH_CONTEXT *pHashContext, unsigned char *pbHashValue);

void CryptDestroyKey(CRYPTKEY *pCryptKey);
bool CryptGenKey (ALG_ID Algid, DWORD dwFlags, CRYPTKEY **ppKey);
bool CryptImportKey(const unsigned char *pbData, DWORD dwDataLen,  CRYPTKEY *pPubKey, DWORD dwFlags, CRYPTKEY **ppKey);
bool CryptExportKey(CRYPTKEY *pCryptKey, CRYPTKEY *PubKey, DWORD dwBlobType, DWORD dwFlags, unsigned char *pbData, DWORD *pdwDataLen);
bool CryptGetHashParam(CRYPTHASH *pCryptHash, DWORD dwParam, unsigned char *pbData, DWORD *pdwDataLen, DWORD dwFlags);
bool CryptCreateHash(ALG_ID Algid, DWORD dwFlags, CRYPTHASH **ppHash);
void CryptDestroyHash(CRYPTHASH *pCryptHash);
bool CryptHashData(CRYPTHASH *pCryptHash, const unsigned char *pbData, DWORD dwDataLen, DWORD dwFlags);
bool CryptSignHash(CRYPTKEY *pCryptKey, CRYPTHASH *pCryptHash, DWORD dwFlags, unsigned char *pbSignature, DWORD *pdwSigLen);
bool CryptVerifySignature(CRYPTHASH *pHash, const unsigned char *pbSignature, DWORD dwSigLen, CRYPTKEY *pPubKey, DWORD dwFlags);
bool CryptDecrypt(CRYPTKEY *pCryptKey, CRYPTHASH *pHash, bool Final, DWORD dwFlags, unsigned char *pbData, DWORD *pdwDataLen);
bool CryptEncrypt(CRYPTKEY *pCryptKey, CRYPTHASH *pHash, bool Final, DWORD dwFlags, unsigned char *pbData, DWORD *pdwDataLen, DWORD dwBufLen);

#endif
