#ifndef __GLOBAL_H
#define __GLOBAL_H

#include "config.h"

//
// Some protocol defaults
//
#define XXX_PORT 8167
#define XXX_MAXPACKETSIZE 1400
#define XXX_PROTOMAJOR 2
#define XXX_PROTOMINOR 0
#define XXX_SOFTWARE "TriX"

//
// Path to file that contains NetBIOS name of the machine
//
#define SMB_CONF_PATH "/etc/samba/smb.conf"

//
// Activity control parameters
//
#define ACT_TIMER_INTERVAL 5000
#define FIRST_PING_LIMIT  36
#define SECOND_PING_LIMIT  39
#define WHOHERE_LIMIT  41
#define TIMEOUT_LIMIT  43

//
// Major channel and board names
//
#define MAIN_CHANNEL "#Main"
#define MAIN_BOARD "Board"
//
// Default data files location
//
#ifndef DATADIR
#ifdef WIN32
#define DATADIR "./data/"
#else
#define DATADIR "./"
#endif
#endif

//
// Status, activity and gender constants
//
enum {
 STATUS_NORMAL = '0',
 STATUS_DND,
 STATUS_AWAY,
 STATUS_OFFLINE,
 STATUS_UNKNOWN, 
 
 RA_ACTIVE = '0',
 RA_NOTACTIVE,
 RA_SPEAKING,
 RA_CONNECTED,
 
 WIN_NOTACTIVE = '0',
 WIN_ACTIVE,
 
 GENDER_MALE = '0',
 GENDER_FEMALE
};

//
//  Smiles enumeration
//
enum {
	SM_TEETH = 1,
	SM_SMILE,
	SM_WINK,
	SM_NORMAL,
	SM_SAD,
	SM_CRY,
	SM_SHADE,
	SM_CONFUSED,
	SM_EMBARASSED,
	SM_OMG,
	SM_TONGUE,
	SM_ANGRY,
	SM_ANGEL,
	SM_DEVIL,
	SM_KISS,
	SM_LOVE,
	SM_UNLOVE,
	SM_BAT,
	SM_THUMBSUP,
	SM_THUMBSDOWN,
	SM_LIGHT,
	SM_PHOTO,
	SM_FILM,
	SM_CAKE,
	SM_CUP,
	SM_PHONE,
	SM_PRESENT,
	SM_ROSE,
	SM_WILTED,
	SM_COCKTAIL,
	SM_STAR,
	SM_NOTE,
	SM_MAIL,
	SM_CLOCK,
	SM_BEER,
	SM_CAT,
	SM_DOG,
	SM_DUDEHUG,
	SM_GIRLHUG,
	SM_SLEEP
};

enum {
    CMD_NORMAL,
    CMD_AWAY,
    CMD_OFFLINE,
    CMD_DND,
    CMD_BEEP,
    CMD_HERE,
    CMD_INFO,
    CMD_JOIN,
    CMD_LEAVE,
    CMD_MESSAGE,
    CMD_MASSMESSAGE,
    CMD_NICK,
    CMD_TOPIC,
    CMD_ADD,
    CMD_ME,
    CMD_CHAT,
    CMD_EXEC,
    CMD_QUIT,
    CMD_HELP,
    CMD_TEST
};

//
// Smiles list
//
const struct
    {
    const char *smile;
    char id;
    const char *file;
    } smilespack[] = {
	 {":))", SM_TEETH, "teeth.png"},	 
	 { ":)", SM_SMILE, "smile.png"},	 
	 {";)", SM_WINK,  "wink.png"},
	 {":-|", SM_NORMAL, "normal.png"},
	 {":(", SM_SAD, "sad.png"},
	 {":'(", SM_CRY, "cry.png"},
	 {"(H)", SM_SHADE, "shade.png"},
	 {":-S", SM_CONFUSED, "confused.png"},
	 {":-$", SM_EMBARASSED, "embarassed.png"},
	 {":-O", SM_OMG, "omg.png"},
	 {":p", SM_TONGUE, "tongue.png"},
	 {":@", SM_ANGRY, "angry.png"},
	 {"(A)", SM_ANGEL, "angel.png"},
	 {"(6)", SM_DEVIL, "devil.png"},
	 {"(K)", SM_KISS, "kiss.png"},
	 {"(L)", SM_LOVE, "love.png"},
	 {"(U)", SM_UNLOVE, "unlove.png"},
	 {":-[",  SM_BAT, "bat.png"},
	 {"(Y)", SM_THUMBSUP, "thumbs_up.png"},
	 {"(N)", SM_THUMBSDOWN, "thumbs_down.png"},
	 {"(I)", SM_LIGHT, "lightbulb.png"},
	 {"(P)", SM_PHOTO, "photo.png"},
	 {"(~)", SM_FILM, "film.png"},
	 {"(^)", SM_CAKE, "cake.png"},
	 {"(C)", SM_CUP, "cup.png"},
	 {"(T)", SM_PHONE, "phone.png"},    
	 {"(G)", SM_PRESENT, "present.png"},
	 {"(F)", SM_ROSE, "rose.png"},
	 {"(W)", SM_WILTED, "wilted_rose.png"},
	 {"(D)", SM_COCKTAIL, "cocktail.png"},
	 {"(*)", SM_STAR, "star.png"},
	 {"(8)", SM_NOTE, "notify.png"},
	 {"(E)", SM_MAIL, "mail.png"},
	 {"(O)", SM_CLOCK, "clock.png"},
	 {"(B)", SM_BEER, "beer.png"},
	 {"(@)", SM_CAT, "cat.png"},
	 {"(&)", SM_DOG, "dog.png"},
	 {"({)", SM_DUDEHUG, "dude_hug.png"},
	 {"(})", SM_GIRLHUG, "girl_hug.png"},
	 {"(S)", SM_SLEEP, "sleep.png"},
	 
	 {"=)", SM_SMILE, "smile.png"}, 
	 {":-)", SM_SMILE,  "smile.png"},
	 {":[", SM_BAT,  "bat.png"},
	 {":$", SM_EMBARASSED, "embarassed.png"},
	 {":s", SM_CONFUSED, "confused.png"},
	 {":|", SM_NORMAL, "normal.png"},
	 {":o", SM_OMG, "omg.png"},
 	 {":-o", SM_OMG, "omg.png"},
	 {"B-)", SM_SHADE, "shade.png"},
	 {"8)", SM_SHADE, "shade.png"},	 
                 {":-P", SM_TONGUE, "tongue.png"},
	 {":,(", SM_CRY, "cry.png"},	 
	 {"=(", SM_SAD, "sad.png"},	 
 	 {":-(", SM_SAD, "sad.png"},
	 {":-@", SM_ANGRY, "angry.png"},
	 {":E", SM_ANGRY, "angry.png"},
	 {";-)", SM_WINK, "wink.png"},
	 {":D", SM_TEETH, "teeth.png"},
	 {":-D", SM_TEETH, "teeth.png"},
	 {":d", SM_TEETH, "teeth.png"},
	 {":-j", SM_BEER, "beer.png"},
	 {":-*", SM_KISS, "kiss.png"},
	 {"*)", SM_LOVE, "love.png"},
	 {"O:-)", SM_ANGEL, "angel.png"},
	 {'\0',    0, '\0'}
};


#endif
