/****************************************************************************
                          logs.cpp  -  description
                             -------------------
    begin                : Thu Sep 13 19:20:17 EST 2005
    copyright            : (C) 2004-2007 by Sergey Litvinov
    email                : linch@adsl.by
 ****************************************************************************/

/****************************************************************************
 *                                                                         							*
 *   This program is free software; you can redistribute it and/or modify  		   	*
 *   it under the terms of the GNU General Public License as published by  		   	*
 *   the Free Software Foundation; either version 2 of the License, or     		   	*
 *   (at your option) any later version.                                   				   	*
 *                                                                         					   		*
 ****************************************************************************/

#include <qdir.h>
#include <qfile.h>
#include <qfileinfo.h>
#include <qtextstream.h>
#include "logs.h"
#include "boardmsg.h"

//#include <iostream>

Logs::Logs(const QString& p, int ms, bool b):path(p), maxsize(ms), board(b)
{
    tpack.insert(CHN_TYPE, "<CHN>");
    tpack.insert(MSG_TYPE, "<MSG>");
    tpack.insert(PRV_TYPE, "<PRV>");
    
    specmap["<SEP>"] = "&lt;SEP&gt;";
    specmap["<SIZE>"] = "&lt;SIZE&gt;";
    specmap["<AUTHOR>"] = "&lt;AUTHOR&gt;";
    specmap["<TOPIC>"] = "&lt;TOPIC&gt;";
    specmap["<TXT>"] = "&lt;TXT&gt;";
    specmap["<UUID>"] = "&lt;UUID&gt;";
    specmap["<PDATE>"] = "&lt;PDATE&gt;";
    specmap["<PTIME>"] = "&lt;PTIME&gt;";
    specmap["<MAGIC1>"] = "&lt;MAGIC1&gt;";
    specmap["<EDATE>"] = "&lt;EDATE&gt;";
    specmap["<ETIME>"] = "&lt;ETIME&gt;";
    specmap["<MAGIC2>"] = "&lt;MAGIC2&gt;";
    specmap["<PORT>"] = "&lt;PORT&gt;";
    specmap["<IP>"] = "&lt;IP&gt;";
    specmap["<SIGN>"] = "&lt;SIGN&gt;";
    specmap["<ID>"] = "&lt;ID&gt;";
  
    locked = true;
    
    setup();
}

Logs::~Logs()
{
    if (board) writeID();
    f.close();
}

bool Logs::setPath(const QString& p)
{
    locked = true;
    f.close();
    path = p;
    setup();
    return true;
}

void Logs::setup()
{
    QFileInfo fi(path);
    QDir d(fi.dir(true));
	
    if (!d.exists()) 
	d.mkdir(d.absPath());
    
    f.setName(path);
    
    if (board)
    {
	bool ok;
//	char *buf;
	Q_UINT32 bmid;
	QString str;
	QStringList tmplist;
	
	fid.setName(path + "_id");
	if (fid.open(IO_ReadOnly)) 
	{
//	    buf = (char *)malloc(fid.size());
//	    fid.readBlock(buf, fid.size());
//	    QString bigstr = QString::fromUtf8(buf, fid.size());
	    QString bigstr = QString::fromUtf8(fid.readAll().data(), fid.size());
	    fid.close();
//	    free(buf);
	    QStringList idlist = QStringList::split(",", bigstr);
	    if (!idlist.isEmpty())
	    {
		for(uint i = 0; i < idlist.count(); i++)
		{
		    tmplist = QStringList::split(":", idlist[i]);
		    if (tmplist.count() == 2);
		    {
			bmid = tmplist[0].toUInt(&ok,36);
			str = tmplist[1];
			idmap[bmid] = str;
		    }
		}
	    }
	}
    }
    
    f.open(IO_ReadWrite | IO_Append);
    locked = false;
}

QStringList Logs::missed()
{
    QStringList bids;
    
    IDMap::Iterator it;
    for (it = idmap.begin(); it != idmap.end(); ++it)
	if (it.data().at(0) == MSG_ABSENT)
	    bids.append(QString::number(it.key(), 36));
    
    return bids;
}

QStringList Logs::missed(const QString &uid)
{
    QStringList bids;
    
    IDMap::Iterator it;
    for (it = idmap.begin(); it != idmap.end(); ++it)
	if (it.data().at(0) == MSG_ABSENT)
	    if (uid == it.data().right(it.data().length() - 1))
		bids.append(QString::number(it.key(), 36));
    
    return bids;
}

QStringList Logs::actMessages()
{
    QStringList bids;
    
    IDMap::Iterator it;
    for (it = idmap.begin(); it != idmap.end(); ++it)
	if (it.data().at(0) == MSG_PRESENT)
	    bids.append(QString::number(it.key(), 36));
    
    return bids;  
}

bool Logs::setMaxSize(int s)
{
    maxsize = s;
    return true;
}

QString Logs::addBoardMsgID(const QString &uid, const QString &bids, char state)
{
    bool ok;
    Q_UINT32 bmid;
    QStringList reqlist;
    QStringList tmplist = QStringList::split(",", bids);
    
    for (uint i = 0; i < tmplist.count(); i++)
    {
	bmid = tmplist[i].toUInt(&ok, 36);
	if ((!idmap.contains(bmid) || (idmap[bmid].at(0) == MSG_DELETED && idmap[bmid].right(idmap[bmid].length() - 1) != uid)) && bmid)
	{
	    idmap[bmid] = QString(state + uid);
	    reqlist.append(tmplist[i]);
	}
    }
    return reqlist.join(",");
}

bool Logs::isMissed(Q_UINT32 id)
{
    if (idmap.contains(id))
	if (idmap[id].at(0) == MSG_ABSENT)
	    return true;
    return false;
}

bool Logs::delBoardMsg(Q_UINT32 id)
{
    if (idmap.contains(id))
    {
	QStringList biglist;
	QString uid, i, tmp, remitem;
	int s, e;
	
	uid = idmap[id].right(idmap[id].length() - 1);
	
	idmap[id] = QString(MSG_DELETED + uid);
	
	biglist = QStringList::split(SEPARATOR, getAll());
	
	QStringList::iterator remit;
	
	for (QStringList::iterator it = biglist.begin(); it != biglist.end(); it++)
	{
	    tmp = QString(*it);
	    s = tmp.findRev("<SIGN>") + 6;
	    e = tmp.findRev("<ID>");
	    i = tmp.mid(s, e - s);
	    if (i == QString::number(id, 36)) remit = it;
	}
		
	biglist.remove(remit);
	
	writeID();
	writeAll(biglist.join(SEPARATOR));
	
	return true;
    }
    return false;
}

QString Logs::encode(const QString &str)
{
    QString enc = str;
    
    SpecMap::Iterator it;
    for (it = specmap.begin(); it != specmap.end(); ++it)
	enc.replace(it.key(), it.data());
    
    return enc;
}

QString Logs::decode(const QString &str)
{
    QString dec = str;
    
    SpecMap::Iterator it;
    for (it = specmap.begin(); it != specmap.end(); ++it)
	dec.replace(it.data(), it.key());
    
    return dec;
}

bool Logs::addBoardMsg(BoardMsg *bmsg)
{
    QString tmp;
    
    if (locked) return false;
    
    if (idmap.contains(bmsg->id()))
	if (idmap[bmsg->id()].at(0) != MSG_ABSENT)
	    return false; //TODO check this
    tmp = QString("<SEP>%1<SIZE>%2<AUTHOR>%3<TOPIC>%4<TXT>")
	  .arg(QString::number(bmsg->size()), bmsg->author(), bmsg->topic(), bmsg->text());
    tmp.append(QString("%1<UUID>%2<PDATE>%3<PTIME>%4<MAGIC1>").arg(bmsg->uuid(), bmsg->pdate(), bmsg->ptime(), QString::number(bmsg->magic1())));
    tmp.append(QString("%1<EDATE>%2<ETIME>%3<MAGIC2>%4<PORT>").arg(bmsg->edate(), bmsg->etime(), QString::number(bmsg->magic2()), QString::number(bmsg->port(), 16)));
    tmp.append(QString("%1<IP>%2<SIGN>%3<ID>").arg(bmsg->ip(), bmsg->sign(), QString::number(bmsg->id(), 36)));
    idmap[bmsg->id()] = QString(MSG_PRESENT + bmsg->uuid());
    if (writeID())
	return writeData(tmp);
    return false;
}

bool Logs::writeID()
{
    QString tmpstr;
    
    IDMap::Iterator it;
    for (it = idmap.begin(); it != idmap.end(); ++it)
	tmpstr.append(QString::number(it.key(), 36) + ":" + it.data() + ",");
    
    if (fid.open(IO_WriteOnly)) 
    {
	QTextStream stream(&fid);
	stream.setEncoding(QTextStream::UnicodeUTF8);
	stream << tmpstr;
	fid.close();
	return true;
    }
    return false;
}

bool Logs::write(const QString &date, const QString &time, const QString &src, char type, const QString &user, const QString &txt)
{
    QString tmp;
    tmp = QString("<SEP><DATE>%1<DATE><TIME>%2<TIME>%3%4")
	  .arg(date, time, tpack[type], src);
    tmp.append(QString("%1<USR>%2<USR>%3")
	       .arg(tpack[type], user, txt));
    return writeData(tmp);
}

bool Logs::writeData(const QString &data)
{
    if (locked) return false;
    locked = true;
    
    int addictsize, spos;
    QString tmp, buffer;
	
    if (maxsize > 0) 
    {
	if ((int)f.size() > (maxsize * 1024))
	{
	    f.reset();
	    buffer = QString(f.readAll());
	    f.close();	    
	    addictsize = maxsize / 10;
	    tmp = buffer.right((maxsize - addictsize) * 1024);
	    spos = tmp.find(SEPARATOR);
	    if (spos != -1) 
	    {
		buffer = tmp.mid(spos);
		f.open(IO_WriteOnly);
		f.writeBlock(buffer.data(), buffer.length());
		f.close();
	    }
	    f.open(IO_ReadWrite | IO_Append);
	}
    }
    
    QTextStream stream(&f);
    stream.setEncoding(QTextStream::UnicodeUTF8);
    stream << data;
    f.flush();
    locked = false;
    return true;  
}

QString Logs::getAll()
{
    QString str(NULL);
    if (!locked)
    {
	QFile file;
	file.setName(path);
	if (file.open(IO_ReadOnly))
	{
	    str = QString::fromUtf8(file.readAll().data(), file.size());
	    file.close();
	}
    }
    return str;
}

bool Logs::writeAll(const QString &bigstr)
{
    locked = true;
    f.close();
    f.open(IO_WriteOnly);
    QTextStream stream(&f);
    stream.setEncoding(QTextStream::UnicodeUTF8);
    stream << bigstr;
    f.close();
    f.open(IO_ReadWrite | IO_Append);
    locked = false;
    return true;
}
