#ifndef __PACKET_H
#define __PACKET_H

#include "global.h"
#include <stdlib.h>
#include <qcstring.h>
#include <quuid.h>

#include <iostream>

class Packet
{
	private:
	    bool over;	//packet overflow flag
	    uint psize;	//packet size
	    
	    char packet[XXX_MAXPACKETSIZE];

	public:
	    Packet(): over(false), psize(0)
	    {
		memset(&packet, 0, sizeof(packet));
	    }

	    ~Packet()
	    {
	    }

	    inline bool isOk() const { return !over; }
		
	    inline const void *data() const { return packet; }

	    inline const unsigned int size() const { return psize; }
	    
	    inline void clear()
	    {
		memset(&packet, 0, sizeof(packet));
		over = false;
		psize = 0;
	    }
	    
	    inline void init()
	    {
		if (psize + 10 < XXX_MAXPACKETSIZE)
		{
		    memmove(packet + 10, packet, psize);
		    packet[0] = 'X';
		    for (uint i = 1; i < 10; i++)
			packet[i] = (0x30 + random()%9);
		    psize += 10;
		}
	    }
		
	    inline void add(const Q_INT8 c)
	    {
		if (psize < XXX_MAXPACKETSIZE)
		    packet[psize++] = c;
		else
		    over = true;
	    }
	    
	    inline void add16(const Q_UINT16 w)
	    {
		Q_UINT8 byte;
		if (psize + sizeof(Q_UINT16) < XXX_MAXPACKETSIZE) 
		{
		    byte = w;
		    packet[psize++] = byte;
		    byte = w >>8;
		    packet[psize++] = byte;
		}
		else
		    over = true;
	    }

	    inline void add32(const Q_UINT32 dw)
	    {
		Q_UINT8 byte;
		if (psize + sizeof(Q_UINT32) < XXX_MAXPACKETSIZE) 
		{
		    byte = dw;
		    packet[psize++] = byte;
		    byte = dw >> 8;
		    packet[psize++] = byte;
		    byte = dw >> 16;
		    packet[psize++] = byte;
		    byte = dw >> 24;
		    packet[psize++] = byte;
		}
		else
		    over = true;
	    }
	    
	    inline void add(const QCString &str)
	    {
		const int n = str.length();
		if (n + psize < XXX_MAXPACKETSIZE)
		{
		    memcpy(packet + psize, str.data(), n);
		    psize = psize + n + 1;
		}
		else
		    over = true;
	    }

	    inline void add(const QUuid &u)
	    {
		bool o;
		QString uid = u.toString();
		uid = uid.remove('-');
		uid = uid.remove('{');
		uid = uid.remove('}');
    
		if (16 + psize < XXX_MAXPACKETSIZE)
		{
		    Q_UINT32 l = uid.left(8).toULong(&o, 16);
		    Q_UINT16 w1 = uid.mid(8, 4).toInt(&o, 16);
		    Q_UINT16 w2 = uid.mid(12, 4).toInt(&o, 16);
		    Q_UINT8 b1  = uid.mid(16, 2).toInt(&o, 16);
		    Q_UINT8 b2  = uid.mid(18, 2).toInt(&o, 16);
		    Q_UINT8 b3 = uid.mid(20, 2).toInt(&o, 16);
		    Q_UINT8 b4 = uid.mid(22, 2).toInt(&o, 16);
		    Q_UINT8 b5 = uid.mid(24, 2).toInt(&o, 16);
		    Q_UINT8 b6 = uid.mid(26, 2).toInt(&o, 16);
		    Q_UINT8 b7 = uid.mid(28, 2).toInt(&o, 16);
		    Q_UINT8 b8 = uid.mid(30, 2).toInt(&o, 16);

		    add32(l);
		    add16(w1);
		    add16(w2);
		    add(b1);
		    add(b2);
		    add(b3);
		    add(b4);
		    add(b5);
		    add(b6);
		    add(b7);
		    add(b8);
		}
		else 
		    over = true;
	    }

	    inline void addArray(const QByteArray &array)
	    {
		if (psize + array.size() < XXX_MAXPACKETSIZE)
		{
		    memcpy(packet + psize, array.data(), array.size());
		    psize = psize + array.size();
		}
		else 
		    over = true;
	    }
};

#endif
