#include "global.h"
#include "settings.h"
#include "conversions.h"
#include <qdir.h>
#include <qprocess.h>

#ifndef WIN32

#include <unistd.h>
#include <sys/ioctl.h>
#include <net/if.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <arpa/inet.h>

#ifdef FREEBSD
#include <net/route.h>
#include <netinet/in.h>
#endif

#endif

#include <stdlib.h>
#include <iostream>

Settings::Settings()
{
    cfg.beginGroup("/trix");
}

Settings::~Settings()
{
    CryptDestroyKey(pKey);
    cfg.endGroup();
}

const QString Settings::GetEnv(char *arg)
{
    QString tmp;
    QProcess *proc = new QProcess(NULL);
    proc->addArgument(arg);
 
    if ( !proc->start() ) 
    {
	tmp = "Unknown";
	delete proc;
	return tmp;
    }
    for (uint i = 0; i < 1000; i++) 
    {
	if (proc->canReadLineStdout()) 
	{
	    tmp = proc->readLineStdout();
	    break;
	}
	sleep(1);
    }
    delete proc;
    return tmp;
}

const QString Settings::GetSystemVersion()
{
    QString tmp, tmp2;
#if defined ( WIN32 )
    OSVERSIONINFOEX osvi;
    BOOL bOsVersionInfoEx;

    memset(&osvi, 0, sizeof(osvi));
    osvi.dwOSVersionInfoSize = sizeof(OSVERSIONINFOEX);

    if( !(bOsVersionInfoEx = GetVersionEx ((OSVERSIONINFO *) &osvi)) )
    {
 osvi.dwOSVersionInfoSize = sizeof (OSVERSIONINFO);
 if (! GetVersionEx ( (OSVERSIONINFO *) &osvi) )
     return QString("Unknown");
    }

    switch (osvi.dwPlatformId)
    {
    case VER_PLATFORM_WIN32_NT:


 if ( osvi.dwMajorVersion <= 4 )
     tmp.append("Microsoft Windows NT ");
   
   if ( osvi.dwMajorVersion == 5 && osvi.dwMinorVersion == 0 )
       tmp.append("Microsoft Windows 2000 ");
   
   if( bOsVersionInfoEx ) 
   { 
       if ( osvi.wProductType == VER_NT_WORKSTATION )
       {
    if ( osvi.dwMajorVersion == 5 && osvi.dwMinorVersion == 1 )
        tmp.append ("Microsoft Windows XP ");
    
    if( osvi.wSuiteMask & VER_SUITE_PERSONAL )
        tmp.append ( "Home Edition " );
    else
        tmp.append ( "Professional " );
       }
       else if ( osvi.wProductType == VER_NT_SERVER )
       {
    if ( osvi.dwMajorVersion == 5 && osvi.dwMinorVersion == 2 )
        tmp.append ("Microsoft Windows .NET ");

    if( osvi.wSuiteMask & VER_SUITE_DATACENTER )
        tmp.append ( "DataCenter Server " );
    else if( osvi.wSuiteMask & VER_SUITE_ENTERPRISE )
        if( osvi.dwMajorVersion == 4 )
     tmp.append ("Advanced Server " );
    else
        tmp.append ( "Enterprise Server " );
    else if ( osvi.wSuiteMask == VER_SUITE_BLADE )
        tmp.append ( "Web Server " );
    else
        tmp.append ( "Server " );
       }
   }
   else
   {
       HKEY hKey;
       char szProductType[80];
       DWORD dwBufLen=80;
       LONG lRet;

       lRet = RegOpenKeyEx( HKEY_LOCAL_MACHINE,
       (LPCWSTR)"SYSTEM\\CurrentControlSet\\Control\\ProductOptions",
       0, KEY_QUERY_VALUE, &hKey );
       if( lRet != ERROR_SUCCESS )
    return QString("Unknown");

       lRet = RegQueryValueEx( hKey, (LPCWSTR)"ProductType", NULL, NULL,
          (LPBYTE) szProductType, &dwBufLen);
       if( (lRet != ERROR_SUCCESS) || (dwBufLen > 80) )
    return QString("Unknown");

       RegCloseKey( hKey );

       if ( lstrcmpi( (LPCWSTR)"WINNT", (LPCWSTR)szProductType) == 0 )
    tmp.append( "Professional " );
       if ( lstrcmpi( (LPCWSTR)"LANMANNT", (LPCWSTR)szProductType) == 0 )
    tmp.append( "Server " );
       if ( lstrcmpi( (LPCWSTR)"SERVERNT", (LPCWSTR)szProductType) == 0 )
    tmp.append( "Advanced Server " );
   }

    if ( osvi.dwMajorVersion <= 4 )
   {
       tmp2.sprintf ("version %d.%d %s (Build %d)\n",
      osvi.dwMajorVersion,
      osvi.dwMinorVersion,
      osvi.szCSDVersion,
      osvi.dwBuildNumber & 0xFFFF);
       tmp.append(tmp2);
   }
   else
   { 
       tmp2.sprintf ("%s (Build %d)\n",
      osvi.szCSDVersion,
      osvi.dwBuildNumber & 0xFFFF);
       tmp.append(tmp2);
   }
   return tmp;

       case VER_PLATFORM_WIN32_WINDOWS:

   if (osvi.dwMajorVersion == 4 && osvi.dwMinorVersion == 0)
   {
       tmp.append ("Microsoft Windows 95 ");
       if ( osvi.szCSDVersion[1] == 'C' || osvi.szCSDVersion[1] == 'B' )
    tmp.append("OSR2 " );
   } 

   if (osvi.dwMajorVersion == 4 && osvi.dwMinorVersion == 10)
   {
       tmp.append ("Microsoft Windows 98 ");
       if ( osvi.szCSDVersion[1] == 'A' )
    tmp.append("SE " );
   } 
   
   if (osvi.dwMajorVersion == 4 && osvi.dwMinorVersion == 90)
   {
       tmp.append ("Microsoft Windows Millennium Edition ");
   } 
   return tmp;
      }
   return tmp; 
#else
 QProcess *proc = new QProcess(NULL);
 proc->addArgument( "uname" );
 proc->addArgument( "-sr" );
 
 if ( !proc->start() ) {
  tmp = "Unknown";
  delete proc;
  return tmp;
 }
 for (uint i = 0; i < 1000; i++) {
  if (proc->canReadLineStdout()) {
   tmp = proc->readLineStdout();
   break;
  }
  sleep(1);
 }
 delete proc;
 return tmp;
#endif
}

void Settings::getInterfaces()
{
    QString ifn;
    QString ifip;
    struct if_nameindex *ifaces = if_nameindex();
    struct ifreq ifr;
    
    int s = socket(AF_INET, SOCK_DGRAM, 0);
    
    if (ifaces != NULL) 
    {
	for (int i = 0; ifaces[i].if_index != 0; i++)
	{
	    ifn = ifaces[i].if_name;
	    
	    memset(&ifr, 0, sizeof(ifr));
	    strcpy(ifr.ifr_name, ifaces[i].if_name);
	    
	    if (ioctl(s, SIOCGIFADDR , &ifr) == 0)
	    {
		struct sockaddr_in *sa = (struct sockaddr_in*)&ifr.ifr_addr;
		ifip = QString(inet_ntoa(sa->sin_addr));
		
		if (( strcmp(ifaces[i].if_name, "eth0") == 0) || (i == 0))
		    def_ip = ifip;
		
		ifn.append(" (" + ifip + ")");
		ifname[ifn] = ifip;
	    }
	}
	if_freenameindex(ifaces);
    }
    close(s);
}

QStringList Settings::getIFNames()
{
    QStringList ifnames;
    
    for (IFMap::iterator it = ifname.begin(); it != ifname.end(); ++it)
	ifnames.append(it.key());
    
    return ifnames;
}

QString Settings::getIFbyIP(const QString& ifip)
{
    for (IFMap::iterator it = ifname.begin(); it != ifname.end(); ++it)
	if (it.data() == ifip) 
	    return it.key();
    return QString(NULL);
}

void Settings::getDefaults()
{
    QString datapath(DATADIR);

    def_encoding="CP 1251";
    def_topic = " "; 
    def_local = "trix_en.qm";
    def_sound = Sound::Command;
    def_maxhistsize=500; //Kb
    def_maxavatarsize=50;//Kb
    def_maxlinescount=5;
    def_histpath=QString(QDir::homeDirPath() + "/.trix/logs/logs");
    def_downloadpath=QDir::homeDirPath();
    def_awaylimit=1800;
    def_autoupdateinterval = 300;
#if defined ( WIN32 )    
    def_playcmd = "";
#else
    def_playcmd = "/usr/bin/aplay";
#endif
    def_snd[Sound::Chatline] = datapath + "snd/chatline.wav";
    def_snd[Sound::Beep] = datapath + "snd/beep.wav";
    def_snd[Sound::JoinChannel] = datapath + "snd/join.wav";
    def_snd[Sound::LeaveChannel] = datapath + "snd/leave.wav";
    def_snd[Sound::Message] = datapath + "snd/message.wav";

 //
 // get hostname
//
    char tmp[255];
    memset(&tmp, 0, sizeof(tmp));
    gethostname(tmp, 254);
    hostname = QString(tmp);
#ifndef WIN32
    domain = GetEnv("domainname");
    QFile file(SMB_CONF_PATH);
    if (file.open(IO_ReadOnly)) 
    {
	char buf[1024];
	while(file.readLine(buf, sizeof(buf)) != -1)
	{
	    QString smb_file_str = QString::fromLocal8Bit(buf);
	    smb_file_str = smb_file_str.simplifyWhiteSpace();
	    smb_file_str = smb_file_str.remove(' ');
	    if (QString::compare(smb_file_str.left(11).lower(), "netbiosname") == 0) 
		hostname = smb_file_str.mid(12, smb_file_str.length() - 12);
	    if (QString::compare(smb_file_str.left(9).lower(), "workgroup") == 0)
		domain = smb_file_str.mid(10, smb_file_str.length() - 10);
	}
	file.close();
    }
    char *login = getenv("LOGNAME");
    def_nick = QString(login);
    logonname = GetEnv("whoami");
#else
    def_nick = QString(tmp);
    
    DWORD HostIP = 0;
    LPHOSTENT lphost;
    char szIPAddr[17];
    
    if(lphost = gethostbyname(tmp))

    HostIP = ((LPIN_ADDR)lphost->h_addr)->s_addr;

    strcpy(szIPAddr,inet_ntoa(*((in_addr*)lphost->h_addr_list[0])));

    def_ip = QString(szIPAddr);
    domain = ""; //TODO Make workable
    logonname = QString(tmp);
#endif
    os = GetSystemVersion();
 //
 // get interface parameters
    def_ip = "127.0.0.1";
    getInterfaces();    
    def_bcastaddress = "255.255.255.255";
    def_mcastgrp = "227.0.0.2";
    def_port = XXX_PORT;
    def_ttl = 7;
 
 //
 // get colors
 def_back = "#000000";
 def_own = "#FFFF80";
 def_users = "#FFFFFF";
 def_prv = "#8080FF";
 def_system = "#55AAFF";
 def_notify = "#C0C0C0";
 def_joinleave = "#FF8080";
 def_userslistbackcolor = "#000000";
 def_userslisttextcolor = "#FFFFFF";
 def_nickeditbackcolor = "#000000";
 def_nickedittextcolor = "#FFFFFF";
 def_chatlineeditbackcolor = "#000000";
 def_chatlineedittextcolor = "#FFFFFF";
 
 //
 //get autoanswers
 def_dndanswer = "Don't disturb me now";
 def_awayanswer = "I'm away from computer";
 def_offlineanswer = "I'm offline now";
 
#ifdef WIN32
 def_webbrowser = "opera.exe";
 def_smbbrowser = "explorer.exe";
 def_emailclient = "thebat.exe";
 def_ftpclient = "opera.exe";
#else
 def_webbrowser = "konqueror";
 def_smbbrowser = "konqueror";
 def_emailclient = "thunderbird";
 def_ftpclient = "opera";
#endif
 
 def_npfile = QDir::homeDirPath() + "/.np.txt";
}

bool Settings::load()
{
 bool ok;
 DWORD bSize;
 QString pKeyBlob;
 QString c;
 unsigned char pbData[596];
 
 getDefaults();

 cfg.beginGroup("/Network");
  nick = cfg.readEntry("/Nick", def_nick);
  ip = cfg.readEntry("/IP", def_ip);
  port = cfg.readNumEntry("/Port", def_port);
  ttl = cfg.readNumEntry("/TTL", def_ttl);
  bcastaddress = cfg.readEntry("/BroadcastAddress", def_bcastaddress);
  mcastgrp = cfg.readEntry("/MulticastGroup", def_mcastgrp);
  usemcast = cfg.readBoolEntry("/UseMulticast", true);
 cfg.endGroup();
 cfg.beginGroup("/Misc");
  secondsintimestamp = cfg.readBoolEntry("/TimeStampWithSeconds", false);
  useuserscolors = cfg.readBoolEntry("/UseUsersColors", false);
  encoding = cfg.readEntry("/Encoding", def_encoding);
  useutf = cfg.readBoolEntry("/UTF", false);
  sendchlist = cfg.readBoolEntry("/SendChannelsList", true);
  np = cfg.readBoolEntry("/NowPlaying",false);
  npfile = cfg.readEntry("/NPFile", def_npfile);
  useamarok = cfg.readBoolEntry("/UseAmaroK", false);
  userpixmap = cfg.readEntry("/UserPixmap", "male.png");
  startchannels = cfg.readBoolEntry("/OpenChannelsOnStart", false);
  channelslist = cfg.readEntry("/ChannelsList", "");
  webbrowser = cfg.readEntry("/WebBrowser", def_webbrowser);
  smbbrowser = cfg.readEntry("/SMBBrowser", def_smbbrowser);
  emailclient = cfg.readEntry("/EmailClient", def_emailclient);
  ftpclient = cfg.readEntry("/FTPClient", def_ftpclient);
  boardenabled = cfg.readBoolEntry("/BoardEnabled", true);
  requestfromauthorsonly = cfg.readBoolEntry("/RequestFromAuthorsOnly", false);
  systraynotify = cfg.readBoolEntry("/SystrayNotifyEnabled", true);
  useros = cfg.readEntry("/UserOS", "");
  clientname = cfg.readEntry("/ClientName", "");
 cfg.endGroup();
 cfg.beginGroup("/General");
 version = cfg.readEntry("/Version", "0", &ok);
 if (!ok)
     cfg.writeEntry("/Version", VERSION);
 uuid = QUuid(cfg.readEntry("/UUID", QString::null));
 if (uuid.isNull())
 {
     uuid = QUuid::createUuid();
     cfg.writeEntry("/UUID", uuid.toString());
 }
  topic = cfg.readEntry("/Topic", def_topic);
  local = cfg.readEntry("/Local", def_local);
  keep_status = cfg.readBoolEntry("/KeepStatus", true);
  status = cfg.readNumEntry("/Status", STATUS_NORMAL);
  gender = cfg.readNumEntry("/Gender", GENDER_MALE);
  allowmass = cfg.readBoolEntry("/AllowMass", true);
  showmassonchannel = cfg.readBoolEntry("/ShowMassOnChannel", true);
  infoack = cfg.readBoolEntry("/InfoACK", true);
  startmin = cfg.readBoolEntry("/StartMinimized", true);
  showra = cfg.readBoolEntry("/ShowRemoteActive", true);
  histpath = cfg.readEntry("/HistoryFile", def_histpath);
  downloadpath = cfg.readEntry("/DownloadPath", def_downloadpath);
  maxhistsize = cfg.readNumEntry("/MaxHistorySize", def_maxhistsize);
  maxavatarsize = cfg.readNumEntry("/MaxAvatarSize", def_maxavatarsize);
  maxlinescount = cfg.readNumEntry("/MaxLinesCount", def_maxlinescount);
  acceptwoprompt = cfg.readBoolEntry("/AcceptWithoutPrompt", false);
  createsenderfolder = cfg.readBoolEntry("/CreateSenderFolder", true);
  legacystyle = cfg.readBoolEntry("/LegacyUpdateStyle", false);
  multiline = cfg.readBoolEntry("/Multiline", false);
  lastnicks = cfg.readEntry("/LastNicks", QString((nick + QChar(':')) + (QChar(gender) + QString("male.png"))));
//  lastnicks = cfg.readEntry("/LastNicks", nick);
  privmsgaction = cfg.readNumEntry("/PrivMsgAction", 1);
  privshowaction = cfg.readNumEntry("/PrivShowAction", 0);
  systemmsgshow = cfg.readNumEntry("/SystemMessages", 255);
  showpopups = cfg.readBoolEntry("/ShowPopups", true);
  awaylimit = cfg.readNumEntry("/AwayLimit", def_awaylimit);
  autoupdate = cfg.readBoolEntry("/AutoUpdate", false);
  autoupdateinterval = cfg.readNumEntry("/AutoUpdateInterval", def_autoupdateinterval);
 cfg.endGroup();
 cfg.beginGroup("/Sound");
 sound = static_cast<Sound::SoundType>(cfg.readNumEntry("/sound", def_sound));
 playcmd = cfg.readEntry("/PlayCommand", def_playcmd);
 cfg.endGroup();
 cfg.beginGroup("/Wavs");
 for (unsigned int i = 0; i < Sound::Dummy; i++)
     snd[i] = cfg.readEntry(getSampleKey(static_cast<Sound::Sfx>(i)), def_snd[static_cast<Sound::Sfx>(i)]);
 cfg.endGroup();
 cfg.beginGroup("/WavFlags");
 for (unsigned int i = 0; i < Sound::Dummy; i++)
     sndp[i] = cfg.readBoolEntry(getSampleKey(static_cast<Sound::Sfx>(i)), true);
 cfg.endGroup();

 cfg.beginGroup("/Window");
 x = cfg.readNumEntry("/x", -1);
 y = cfg.readNumEntry("/y", -1);
 width = cfg.readNumEntry("/Width", 480);
 height = cfg.readNumEntry("/Height", 360);
 framewidth = cfg.readNumEntry("/FrameWidth", 120);
 cfg.endGroup();
 cfg.beginGroup("/Autoanswers");
 dndanswer = cfg.readEntry("/DND", def_dndanswer);
 awayanswer = cfg.readEntry("/Away", def_awayanswer);
 offlineanswer = cfg.readEntry("/Offline", def_offlineanswer);
 cfg.endGroup(); 
 cfg.beginGroup("/Fonts");
 chatfont.fromString(cfg.readEntry("/Chat", "Sans,10,-1,5,50,0,0,0,0,0"));
 userslistfont.fromString(cfg.readEntry("/UsersList", "Sans,10,-1,5,50,0,0,0,0,0"));
 editlinefont.fromString(cfg.readEntry("/EditLine", "Sans,10,-1,5,50,0,0,0,0,0"));
 cfg.endGroup(); 
 cfg.beginGroup("/Colors");
 backgroundcolor.setNamedColor(cfg.readEntry("/Background", def_back));
 owncolor.setNamedColor(cfg.readEntry("/Own", def_own));
 userscolor.setNamedColor(cfg.readEntry("/Users", def_users));
 privatecolor.setNamedColor(cfg.readEntry("/Private", def_prv));
 systemcolor.setNamedColor(cfg.readEntry("/System", def_system));
 notifycolor.setNamedColor(cfg.readEntry("/Notify", def_notify));
 joinleavecolor.setNamedColor(cfg.readEntry("/Join-Leave", def_joinleave));
 userslistbackcolor.setNamedColor(cfg.readEntry("/UsersListBackColor", def_userslistbackcolor));
 userslisttextcolor.setNamedColor(cfg.readEntry("/UsersListTextColor", def_userslisttextcolor));
 nickeditbackcolor.setNamedColor(cfg.readEntry("/NickEditBackColor", def_nickeditbackcolor));
 nickedittextcolor.setNamedColor(cfg.readEntry("/NickEditTextColor", def_nickedittextcolor));
 chatlineeditbackcolor.setNamedColor(cfg.readEntry("/ChatLineEditBackColor", def_chatlineeditbackcolor));
 chatlineedittextcolor.setNamedColor(cfg.readEntry("/ChatLineEditTextColor", def_chatlineedittextcolor));
 cfg.endGroup();
 cfg.beginGroup("/Personal");
 fullname = cfg.readEntry("/FullName", " ");
 occupation = cfg.readEntry("/Occupation", " ");
 department = cfg.readEntry("/Department", " ");
 phone = cfg.readEntry("/Phone", " ");
 mobile = cfg.readEntry("/Mobile", " ");
 homepage = cfg.readEntry("/Homepage", " ");
 email = cfg.readEntry("/Email", " ");
 address = cfg.readEntry("/Address", " ");
 cfg.endGroup();
 cfg.beginGroup("/Crypto");
 pKeyBlob = cfg.readEntry("/PrivateKeyBlob", "0");
 cfg.endGroup();
 
 memset(&pbData, 0, sizeof(pbData));
 
 if  (pKeyBlob.length() == 1192)
 {
     for (uint i = 0, j = 0; i < pKeyBlob.length(); i += 2, j++)
     {
	 c = pKeyBlob.mid(i, 2);
	 pbData[j] = (unsigned char)c.toUShort(&ok, 16);
     }
 }
    
 if (!CryptImportKey(pbData, sizeof(pbData), NULL, 0, &pKey))
 {
     bSize = 0;
     CryptGenKey(AT_KEYEXCHANGE, 1024<<16 | CRYPT_EXPORTABLE , &pKey);
     
     memset(&pbData, 0, sizeof(pbData));
     
     CryptExportKey(pKey, NULL, PRIVATEKEYBLOB, 0, (unsigned char*)pbData, (DWORD *)&bSize);
     if (!CryptExportKey(pKey, NULL, PRIVATEKEYBLOB, 0, (unsigned char*)pbData, (DWORD *)&bSize))
	 std::cout << "Export pvtkey failed" << std::endl;
     
     pKeyBlob = Conversions::raw2str(pbData, sizeof(pbData));

     cfg.writeEntry("/Crypto/PrivateKeyBlob", pKeyBlob);
 }
	 
 return true;
}
