#include "sound.h"
#include "settings.h"
#include <qapplication.h>
#include <qmessagebox.h>
#include <qobject.h>
#include <qprocess.h>
#include <qsound.h>

Sound::SoundType Sound::sndtype;
SoundSystem *Sound::sndsys = NULL;

bool Sound::reload_samples(Settings &settings)
{
    return sndsys->reload_samples(settings);
}

bool Sound::init(Settings *settings)
{
    sndtype = settings->getSoundType();
   
    switch (sndtype)
    {
    case NoSound:
 sndsys = new SoundSystem();
 break;
    case PCSpeaker:
 sndsys = new SoundPCSpeaker();
 break;
    case Command:
 sndsys = new SoundCommand();
 break;
    case NAS:
 sndsys = new SoundNAS();
 break;
    default: 
 return false;
    }
    if (!sndsys)
 return false;
    bool status = sndsys->init();
    if (!status) {
             delete sndsys;
              sndsys = new SoundPCSpeaker();
              status = sndsys->init();
    }
    if (status)
              sndsys->reload_samples(*settings);
    return status;
} 

void Sound::close()
{
    if (sndsys)
   sndsys->close();
}

const QString& Sound::error()
{
    return sndsys->errstr;
}

void Sound::play(Sfx sfx)
{
    if (sndsys) sndsys->play(sfx);
}

SoundSystem::SoundSystem()
{
    for (int i = 0; i < Sound::Dummy; i++)
   sample << " ";
}

SoundSystem::~SoundSystem()
{
    
}

bool SoundSystem::reload_samples(Settings &settings)
{
    bool status = true;
    for (int i = 0; i < Sound::Dummy; i++)
   sample[i] = settings.getSampleFileName(static_cast<Sound::Sfx>(i)).latin1();
    return status;
}

bool SoundPCSpeaker::init()
{
    return true;
}

void SoundPCSpeaker::play(Sound::Sfx sfx)
{
    QApplication::beep();
}

void SoundPCSpeaker::close()
{
}

bool SoundNAS::init()
{
    if (!QSound::isAvailable()) 
    {
        errstr = QString("<p><b>Sorry, you are not running the Network Audio System.</b><p>If you have the `au' command, run it in the background before this program. The latest release of the Network Audio System can be obtained from:<pre>\n &nbsp;\n   ftp.ncd.com:/pub/ncd/technology/src/nas\n   ftp.x.org:/contrib/audio/nas\n</pre><p>Release 1.2 of NAS is also included with the X11R6 contrib distribution.<p>After installing NAS, you will then need to reconfigure Qt with NAS sound support");
        return false;
    }
    else 
	return true;
}

void SoundNAS::play(Sound::Sfx sfx)
{
    QSound::play(sample[sfx]);
}

void SoundNAS::close()
{
    
}

bool SoundCommand::reload_samples(Settings &settings)
{
    const QStringList cmd = QStringList::split(' ', settings.getPlayCommand());
    for (int i = 0; i < Sound::Dummy; i++)
    {
	arg[i].clear();
	arg[i] = cmd;
	arg[i].append(settings.getSampleFileName(static_cast<Sound::Sfx>(i)));
    }
    return true;
}

bool SoundCommand::init()
{
    return true;
}

void SoundCommand::play(Sound::Sfx sfx)
{
    QProcess *proc = new QProcess(arg[sfx], NULL);
    if (!proc->start())
	qWarning("Cannot run play command");
    delete proc;
}

