#ifndef __SOUND_H
#define __SOUND_H

#include <qmemarray.h>
#include <qstringlist.h>

class Settings;
class SoundSystem;

class Sound
{ 
 public:
  enum SoundType { NoSound = 0, PCSpeaker, NAS, Command };
  enum Sfx
   {  Chatline = 0, 
      Message, 
      JoinChannel,
      LeaveChannel, 
      Beep, 
      Dummy
  };

 private:
  static SoundSystem *sndsys;
  static SoundType sndtype;
  
 public:
  bool reload_samples(Settings &settings);
  static bool init(Settings *settings);
  static void close();
  static const QString& error();
  static void play(Sfx sfx);
};

class SoundSystem
{
 protected:
     QString errstr;
     QStringList sample;
  virtual bool reload_samples(Settings &settings);
  
 public:
  SoundSystem();
  virtual ~SoundSystem();
  virtual bool init() { return true; };
  virtual void close() {};
  virtual void play(Sound::Sfx sfx) {};
  friend class Sound;
};

class SoundPCSpeaker: public SoundSystem
{
 public:
  virtual bool init();
  virtual void close();
  virtual void play(Sound::Sfx sfx);
};

class SoundNAS: public SoundSystem
{
 public:
  virtual bool init();
  virtual void close();
  virtual void play(Sound::Sfx sfx);
};

class SoundCommand: public SoundSystem
{
 private:
  QStringList arg[Sound::Dummy];
  virtual bool reload_samples(Settings &settings);
  
 public:
  virtual bool init();
  virtual void play(Sound::Sfx sfx);
};

#endif

