#include "tcpsocket.h"
#include "tcpserver.h"
#include "filerecvdlg.h"
#include "xxxprotocol.h"
#include "packet.h"
#include "base64_codec.h"
#include "conversions.h"
#include <qfileinfo.h>
#include <qdir.h>
#include <qtimer.h>
#include <qurl.h>
#include <unistd.h>
#include <stdlib.h>
//#include <iostream>

TCPSocket::TCPSocket( QObject *p, CRYPTKEY *pCryptKey) : QSocket( p ), vis(false), chatsession(false), initialized(false), accepted(false), connectestablished(false), sessionestablished(false), httpsession(false), filesleft(0), receivedbytes(0), sendbytes(0), descfilessize(0), filenumber(0), totalfiles(0), currfilenum(0), currtransfer(0), rf(NULL), sf(NULL), pkey(pCryptKey), skey(NULL)
{
    connect( this, SIGNAL(readyRead()), SLOT(read()) );
//    std::cout << "TCPSocket created" << std::endl;
    dtimer = new QTimer(this);
    connect (dtimer, SIGNAL(timeout()), this, SLOT(closeConn()));
}

void TCPSocket::read()
{
//    std::cout << "TCPSocket read" << std::endl;
    Q_ULONG bytesAvail = bytesAvailable();
    if ( bytesAvail < 1 )
	return;
    QByteArray ba( bytesAvail );
    readBlock( ba.data(), bytesAvail );

    if (!initialized)
    {
	uint t;
	QCString cstr(ba.data());
	QStringList lines;
	uint datasize;
	int datapos = cstr.find("\n\n");
	if (datapos != -1) datapos += 2;
	int headerend = cstr.find("\r\n\r\n");
	if (headerend != -1) 
	{
	    datapos  = headerend + 4;
	    httpsession = true;
	}
	if (datapos == -1 )
	{
	    closeConn();
	    return;
	}
	datasize = ba.size() - datapos;
	QByteArray baData(datasize);
	cstr.resize(datapos + 1);
	QString sLines(cstr);
	
	if (datasize > 0) baData.duplicate(reinterpret_cast<char *>(ba.data() + datapos), datasize);

	if (!httpsession) lines = QStringList::split('\n', sLines);
	else lines = QStringList::split("\r\n", sLines);
	
	if (httpsession)
	{
	    QString headerstr = cstr.left(cstr.find("\r\n"));
	    QStringList param = QStringList::split(' ', headerstr);
	    
	    if (param.count() < 3)
	    {
		closeConn();
		return;
	    }
	    if (param[0] == "GET")
	    {
		QUrl urlstr(param[1]);
		
		filename = urlstr.fileName();
		prot = param[2];
		messagetype = HTTP_GET;
		
	    } else if (param[0] == "HTTP/1.1") {
		prot = param[0];
		text = param[1];
		messagetype = HTTP_POST;
		
		if (text != "200")
		{
		    closeConn();
		    return;
		}
	    }
	}
	
	for (uint i = (int)httpsession; i < lines.count(); i++)
	{
	    int splitpos = lines[i].find(':');
	    QString varName = lines[i].left(splitpos);
	    QString varValue = lines[i].right(lines[i].length() - splitpos - 2);
	    varName.stripWhiteSpace();
	    varValue.stripWhiteSpace();
	    
	    for (uint j = 0; varspack[j].varname; j++)
	    {
		if (QString::compare(varspack[j].varname, varName) == 0)
		{
			switch(varspack[j].id)
			{
			case PROTOCOL_ID:
			    prot = varValue;
			    break;
			case VERSION_ID:
			    vers = varValue.toInt();
			    break;
			case MESSAGE_TYPE_ID:
			    mtype = varValue;
			    break;
			case CONTENT_TYPE_ID:
			    contenttype = varValue;
			    break;
			case CONTENT_LENGTH_ID:
			    contentlength = varValue.toULong();
			    break;
			case FILE_NAME_ID:
			    filename = QString::fromUtf8(varValue.data());
			    break;
			case FILE_DESC_ID:
			    filedesc = QString::fromUtf8(varValue.data());
			    break;
			case FILES_LEFT_ID:
			    filesleft = varValue.toULong();
			    break;
			case DESC_FILES_NAME_ID:
			    descfilesname = QString::fromUtf8(varValue.data());
			    break;
			case DESC_FILES_SIZE_ID:
			    descfilessize = varValue.toULong();
			    break;
			case USER_NAME:
			    username = QString::fromUtf8(varValue.data());
			    break;
			case USER_UUID_ID:
			    useruuid = varValue;
			    break;
			case AUTO_ANSWER_ID:
			    autoanswer = QString::fromUtf8(varValue.data());
			    break;
			case USER_STATUS_ID:
			    userstatus = varValue.toUInt();
			    break;
			case PVT_MESSAGE_SIG_ID:
			    pvtmessagesig = varValue;
			    break;
			case PVT_MASS_ID:
			    pvtmass = varValue;
			    break;
			case DATA_SIGNATURE_ID:
			    datasignature = varValue;
			    break;
			case CKEY_ID:
			    ckey = varValue;
			    break;
			case MESSAGES_ID:
			    messages = varValue;
			    break;
			case TYPING_MODE_ID:
			    typingmode = varValue.toUInt();
			    break;
			case HTTP_CONTENT_LENGTH_ID:
			    contentlength = varValue.toULong();	
			    break;
			case HTTP_CONTENT_TYPE_ID:
			    contenttype = varValue;
			    break;
			default: break;
			}
		    }
	    }
	}
	// Message type checking
	if (!httpsession)
	{
	    for (t = 0; typespack[t].id; t++)
		if (QString::compare(typespack[t].tname, mtype) == 0) break;
	    messagetype = typespack[t].id;
	    if (!messagetype)
	    {
		closeConn();
		return;
	    }
	}
	if (messagetype == TCP_CHAT_REQUEST_OK)
	{
	    if (useruuid == sessuid)
	    {
		sessionestablished = true;
		emit chatSessionEstablished(useruuid);
		return;
	    }
	}
/*	if (messagetype == TCP_FILE_TRANSFER_DENY)
	{
	    closeConn();
	    return;
	}*/
	if (messagetype == TCP_BOARD_REPLY)
	{
	    buff.resize(contentlength);
	    
	    initialized = true;
	    
	    if (datasize  > 0)
		receiveBoardMessage(baData);

	    return;
	}
	if (messagetype == HTTP_POST)
	{
	    file.setName(fullpath);
	    QFileInfo fi(file);
	    QDir d(fi.dir(true));
	    if (!d.exists())
		d.mkdir(d.absPath());
	    if (file.open(IO_WriteOnly))
	    {
		messagetype = TCP_FILE_TRANSFER;
		receivedbytes = 0;
		accepted = true;
		initialized = true;
		
		if (datasize  > 0)
		    receiveFile(baData);
		
		return;
	    }
	}
	if (messagetype == HTTP_GET)
	{
	    TCPServer *par = (TCPServer *)parent();
	    if (filename == "photo")
		fullpath = par->getLocalAvatar();
	    else
		fullpath = par->getLocalFile();//TODO add filename
	    if (fullpath.isEmpty())
	    {
		httpError();
		return;
	    } else {
		file.setName(fullpath);
		QFileInfo fi(file);
		filename = fi.fileName();
		contentlength = fi.size();
		sendbytes = 0;
		currtransfer = 0;
		
		if (file.open(IO_ReadOnly))
		{
		    packet.clear();
		    packet.append("HTTP/1.1 200 OK");
		    packet.append("Content-Length: " + QString::number(contentlength));
		    packet.append("Content-Type: " + Conversions::ext2mime(fi.extension( false )));
		    packet.append("Cache-Control: no-cache");
		    packet.append("Connection: close");
		    initialized = true;
		    sendpacket();
		    connect( this, SIGNAL(bytesWritten(int)), SLOT(writtenBlock(int)) );
		    writeNextBlock();
		    return;
		} else {
		    httpError();
		    return;
		}
	    }
	}
	if (messagetype == TCP_REQUEST_MISSED)
	{
	    QString bids(baData);
	    emit boardRequestMissedReceived(useruuid, username, bids);
	    QTimer::singleShot( 5000, this, SLOT(closeConn()) );
	    return;
	}
	if (messagetype == TCP_PING)
	{
	    emit pingReceived(useruuid, username);
	    QTimer::singleShot( 5000, this, SLOT(closeConn()) );
	    return;
	}
	if (messagetype == TCP_CHAT_MSG)
	{
	    if (sessionestablished)
	    {
		QString msg;
		unsigned char *pbData = NULL;
		int dwDataLen = baData.size();
	    
		pbData = (unsigned char*)malloc(contentlength);
		
		if (pbData)
		{
		    memcpy(pbData, baData.data(), contentlength);
	    
		    if (CryptDecrypt(skey, 0, true, 0, pbData, (DWORD*)&dwDataLen))
			emit recvChatMessage(peerAddress(), QString::fromUtf8((const char*)pbData, contentlength));
		    free(pbData);
		}
		return;
	    }
	}
	if (messagetype == TCP_CHAT_MSG_TYPING)
	{
	    if (sessionestablished)
	    {
		emit recvTypingMode(sessuid, typingmode);
		return;
	    }
	}
	if (messagetype == TCP_CHAT_REQUEST)
	{
	    if (!sessionestablished)
	    {
		unsigned char *pbKeyData;
		uint ckeyb64len = ckey.length();
		int skeylen = (ckeyb64len / 4) * 3;
		
		TCPServer *par = (TCPServer *)parent();
		QString localnick = par->getLocalNick();
		QString uid = par->getLocalUid();

		pbKeyData = (unsigned char*)malloc(skeylen);
		
		if (pbKeyData) 
		{
		    if (decodeBase64((char*)ckey.data(), ckeyb64len, pbKeyData, ckeyb64len, (int *)&skeylen) == 0)
		    {
			if (CryptImportKey(pbKeyData, skeylen, pkey, CRYPT_OAEP, &skey))
			{
			    packet.clear();
			    packet.append(QString("protocol: %1").arg(TCP_PROTOCOL));
			    packet.append(QString("version: %1").arg(TCP_VERSION));
			    packet.append(QString("message-type: %1").arg(typespack[TCP_CHAT_REQUEST_OK - 1].tname));
			    packet.append(QString("user-uuid: %1").arg(uid));
			    packet.append(QString("user-name: %1").arg(localnick));
			    packet.append(QString("ckey: "));
			    packet.append(QString("content-length: 0"));
			    
			    sessuid = useruuid;
			    sessionestablished = true;
			    chatsession = true;
			    
			    sendpacket();
			    emit chatSessionEstablishedByRemote(useruuid);
		
			    return;
			}
		    }
		    free(pbKeyData);
		}
	    }
	}
	
	if (messagetype == TCP_PVTMSG_CONFIRM)
	{
	    emit messageACKReceived(peerAddress(), username, pvtmessagesig, autoanswer, '0' + userstatus);
	    QTimer::singleShot( 5000, this, SLOT(closeConn()) );
	    return;
	}
	
	//New in 2.1.4
	if (messagetype == TCP_MESSAGE_DELIVERY_CONFIRM)
	{
	    return;
	}
	if (messagetype == TCP_PVTMSG)
	{
	    buff.resize(contentlength);

	    initialized = true;

	    if (datasize  > 0)
		receiveMessage(baData);
	    return;
	}

	if (messagetype == TCP_FILE_TRANSFER)
	{
	    initialized = true;
	    if (!accepted) 
	    {
		QString tmppath;
		TCPServer *par = (TCPServer *)parent();
		remoteusername = par->getNick(peerAddress());
		tmppath = par->getDownloadPath();
		createsenderfolder = (tmppath[0] == '1') ? true : false;
		filespath = tmppath.right(tmppath.length() - 1);
		if (createsenderfolder) 
		{
		    if (!filespath.endsWith("/"))
			filespath.append("/");
		    filespath.append(remoteusername + "/");
		}
		totalfiles = filesleft + 1;
		if (!par->getAcceptWOPrompt())
		{
		    FileRecvDlg *fr = new FileRecvDlg(NULL, "File receiving", false);
		    fr->setup(remoteusername, (descfilesname.isEmpty()) ? filename : descfilesname, filespath, filedesc, (descfilessize > 0) ? descfilessize : contentlength, createsenderfolder);
		    connect( fr, SIGNAL(fileAccept(const QString &, const QString&, bool)), this, SLOT(acceptFile(const QString &, const QString &, bool)) );
		    connect( fr, SIGNAL(cancelled(int)), this, SLOT(fileTransferDeny(int)) );
		    fr->show();
		    return;
		}
	    }
	    acceptFile(remoteusername, filespath, false);
	    return;
	}
	if (messagetype == TCP_AVATAR_TRANSFER)
	{
	    initialized = true;
	    QString fpath = QDir::homeDirPath() + "/.trix/avatars/";
	    TCPServer *par = (TCPServer *)parent();
//	    QString sndr = par->getNick(peerAddress());
	    remoteusername = par->getNick(peerAddress());
	    if ((uint)(par->getMaxAvatarSize() << 10) < contentlength) 
	    {
		closeConn();
		return;
	    }
//	    acceptFile(sndr, fpath, true);
	    acceptFile(remoteusername, fpath, true);
	    return;
	}

	if (messagetype == TCP_FILE_TRANSFER_READY || messagetype == TCP_AVATAR_TRANSFER_READY)
	{
	    if (sf) sf->setAction(tr("File sending..."));	    
	    initialized = true;
	    connect( this, SIGNAL(bytesWritten(int)), SLOT(writtenBlock(int)) );
	    writeNextBlock();
	    return;
	}
	if (messagetype == TCP_FILE_TRANSFER_DONE)
	{
	    sendNextFile();
	    return;
	}
	closeConn();
    }
    else
    {
	if (messagetype == TCP_PVTMSG_CONFIRM)
	{
	    char c, stat;
	    int pos = 0;
	    QCString cusername;
	    QCString cautoanswer;
	    XXXProtocol::get(&c, ba.data(), pos, ba.size());
	    XXXProtocol::get(&stat, ba.data(), pos, ba.size());
	    XXXProtocol::get(&cusername, ba.data(), pos, ba.size());
	    XXXProtocol::get(&cautoanswer, ba.data(), pos, ba.size());
	    XXXProtocol::get(&pvtmessagesig, ba.data(), pos, ba.size());
	    emit messageACKReceived(peerAddress(), QString::fromUtf8(cusername.data()), pvtmessagesig, QString::fromUtf8(cautoanswer.data()), stat);
	    QTimer::singleShot( 5000, this, SLOT(closeConn()) );
	    return;
	}
	
	if (messagetype == TCP_PVTMSG)
	{
	    receiveMessage(ba);
	    return;
	}

	if (messagetype == TCP_FILE_TRANSFER || messagetype == TCP_AVATAR_TRANSFER)
	{
	    receiveFile(ba);
	}
	if (messagetype == TCP_BOARD_REPLY)
	{
	    receiveBoardMessage(ba);		    
	}
    }
}

void TCPSocket::httpError()
{
    QString err = "<body><h2>HTTP/1.1 404 Not Found</h2></body>";
    packet.clear();
    packet.append("HTTP/1.1 404 Not Found");
    packet.append("Connection: close");	
    packet.append("Content-Type: text/html");
    packet.append("Content-Length: " + QString::number(err.length() + 4));
    packet.append("");
    packet.append(err);
    sendpacket();
    QTimer::singleShot( 1000, this, SLOT(closeConn()) );
}

void TCPSocket::receiveBoardMessage(const QByteArray &ba)
{
    if (receivedbytes + ba.size() > contentlength)
    {
	closeConn();
	return;
    }
    memcpy(buff.data() + receivedbytes, ba.data(), ba.size());
    receivedbytes += ba.size();

    if (receivedbytes == contentlength)
	emit boardUpdated(buff);
}

void TCPSocket::receiveFile(const QByteArray &ba)
{
    receivedbytes += ba.size();
    if (receivedbytes > contentlength)
    {
	closeConn();
	return;
    }
    if (vis)
    {
	currtransfer += ba.size();
	if (currtransfer >= updlim)
	{
	    currtransfer = 0;
	    Q_UINT32 progress = receivedbytes / updlim;
	    rf->updateProgressBar( progress );
	}
    }
    file.writeBlock(ba.data(), ba.size());
    if (receivedbytes == contentlength) 
    {
	file.close();
	if (rf) 
	{
	    rf->updateProgressBar( 100 );
	    rf->setAction(tr("File received"));
	}
	emit fileRecvFinished(remoteusername, filename);
	if (filesleft == 0)
	{
	    QTimer::singleShot( 1000, this, SLOT(closeConn()) );
	    return;
	}
	if (rf) rf->setAction(tr("File received. Sending confirmation..."));
	packet.clear();
	packet.append(QString("protocol: %1").arg(TCP_PROTOCOL));
	packet.append(QString("version: %1").arg(TCP_VERSION));
	packet.append(QString("message-type: %1").arg(typespack[TCP_FILE_TRANSFER_DONE - 1].tname));
	initialized = false;
	sendpacket();
    }
}

void TCPSocket::receiveMessage(const QByteArray &ba)
{
    if (receivedbytes + ba.size() > contentlength)
    {
	closeConn();
	return;
    }
    
    memcpy(buff.data() + receivedbytes, ba.data(), ba.size());
    receivedbytes += ba.size();
  
    if (receivedbytes == contentlength)
    {
	QString text = QString::fromUtf8(buff.data(), contentlength);
	emit pvtMsgReceived(peerAddress(), useruuid, username, pvtmessagesig, datasignature, text);

	TCPServer *par = (TCPServer *)parent();
	QString localnick = par->getLocalNick();
	QString answer = par->getLocalAnswer();
	char status = par->getLocalStatus();
//    char gender = par->getLocalGender();
	    	    
	Packet p;
	p.add('1');
	p.add(status);
	p.add(localnick.utf8());
	p.add(answer.utf8());
	p.add(pvtmessagesig);
	p.add(0);
	writeBlock( (char*)p.data(),  p.size() );
	QTimer::singleShot( 5000, this, SLOT(closeConn()) );
    }
}

void TCPSocket::fileTransferDeny(int startfrom)
{
    packet.clear();
    packet.append(QString("protocol: %1").arg(TCP_PROTOCOL));
    packet.append(QString("version: %1").arg(TCP_VERSION));
    packet.append(QString("message-type: %1").arg(typespack[TCP_FILE_TRANSFER_DENY - 1].tname));
    packet.append(QString("start-from: %1").arg(QString::number(startfrom)));
    sendpacket();
    QTimer::singleShot( 5000, this, SLOT(closeConn()) );
}

void TCPSocket::sendNextFile()
{
    if (currfilenum > totalfiles)
    {
	closeConn();
	return;
    }
    file.close();
    file.setName(filesnames[currfilenum++]);
    QFileInfo fi(file);
    filename = fi.fileName();
    contentlength = fi.size();
    sendbytes = 0;
    currtransfer = 0;
    updlim = contentlength / 100;
    
    if (updlim == 0) updlim = 1;
	
    if (file.open(IO_ReadOnly))
    {
	packet.clear();
	packet.append(QString("protocol: %1").arg(TCP_PROTOCOL));
	packet.append(QString("version: %1").arg(TCP_VERSION));
	packet.append(QString("message-type: %1").arg(typespack[sendtype - 1].tname));
	packet.append(QString("content-type: %1").arg(TCP_BINARY_TYPE));
	packet.append(QString("content-length: %1").arg(QString::number(contentlength)));
	packet.append(QString("file-name: %1").arg(filename));
	packet.append(QString("file-desc: %1").arg(filedesc));
	packet.append(QString("files-left: %1").arg(--filesleft));
	packet.append(QString("desc-files-name: %1").arg(descfilesname));
	packet.append(QString("desc-files-size: %1").arg(QString::number(descfilessize)));
	packet.append(QString("user-uuid: %1").arg(useruuid));
	    
	descfilesname = "";
	descfilessize = 0;
	
	if (vis)
	{
	    sf->setFileName(filename);
	    sf->setFileSize(contentlength);
	    sf->setCurrentFileNumber(currfilenum, totalfiles);
	}
	    
	sendpacket();
    }
    else
	closeConn();
}

void TCPSocket::connectDone()
{
    connectestablished = true;

    if (messagetype == TCP_FILE_TRANSFER || messagetype == TCP_AVATAR_TRANSFER)
    {
	if (vis) sf->setAction(tr("Waiting for user accept (connected to ") + peerAddress().toString() + ")");
	if (!httpsession) sendNextFile();
	else sendpacket();
	return;
    }
    if (messagetype == TCP_BOARD_REQUEST)
    {
	sendpacket();
	QTimer::singleShot( 5000, this, SLOT(closeConn()) );
	return;
    }
    if (messagetype == TCP_REQUEST_MISSED)
    {
	initialized = true;
	sendpacket();
	flush();
	writeBlock(text.utf8(), text.utf8().length());
//	QTimer::singleShot( 5000, this, SLOT(closeConn()) );
	return;
    }
    if (messagetype == TCP_PVTMSG_CONFIRM)
    {
	sendpacket();
	return;
    }
    if (messagetype == TCP_PVTMSG)
    {
	messagetype = TCP_PVTMSG_CONFIRM;
	initialized = true;
	sendpacket();
	flush();
	writeBlock(text.utf8(), text.utf8().length());
	return;
    }
    if (messagetype == TCP_CHAT_REQUEST)
    {
	sendpacket();
	return;
    }
}

void TCPSocket::createChatSession(const QHostAddress &addr, Q_UINT16 port, const QString &uuid, CRYPTKEY *pubkey)
{
    int bSize, ckeylen;
    unsigned char *pbKeyData = NULL;
    
    chatsession = true;
    sessuid = uuid;
    
    TCPServer *par = (TCPServer *)parent();
    QString localnick = par->getLocalNick();
    QString luid = par->getLocalUid();
    
    messagetype = TCP_CHAT_REQUEST;

    if (CryptGenKey(CALG_RC4, CRYPT_EXPORTABLE , &skey))
    {
	bSize = 0;
	CryptExportKey(skey, pubkey, SIMPLEBLOB, CRYPT_OAEP, 0, (DWORD *)&bSize);
	if (bSize > 0) pbKeyData = (unsigned char*)malloc(bSize);
	
	if (pbKeyData) 
	{
	    if (CryptExportKey(skey, pubkey, SIMPLEBLOB, CRYPT_OAEP, (unsigned char*)pbKeyData, (DWORD *)&bSize))
	    {
		encodeBase64(pbKeyData, bSize, NULL, 0, &ckeylen);
		ckey.resize(ckeylen + 1);
		encodeBase64(pbKeyData, bSize, (char *)ckey.data(), ckey.size(), &ckeylen);
	    }
	    free(pbKeyData);
	}
    }
    
    packet.clear();
    packet.append(QString("protocol: %1").arg(TCP_PROTOCOL));
    packet.append(QString("version: %1").arg(TCP_VERSION));
    packet.append(QString("message-type: %1").arg(typespack[messagetype - 1].tname));
    packet.append(QString("user-uuid: %1").arg(luid));
    packet.append(QString("user-name: %1").arg(localnick));
    packet.append(QString("ckey: %1").arg(ckey));
    packet.append(QString("content-length: 0"));

    connect( this, SIGNAL(connected()), SLOT(connectDone()) );
    
    connectToHost(addr.toString(), port);
    QTimer::singleShot( 5000, this, SLOT(checkConnection()) );
}

void TCPSocket::sleepChatSession()
{
    dtimer->start(300000, true);  
}

void TCPSocket::wakeupChatSession()
{
    dtimer->stop();
}

void TCPSocket::sendTypingMode(uint tmode)
{
    if (sessionestablished)
    {
	TCPServer *par = (TCPServer *)parent();
	QString unick = par->getNick(peerAddress());
    
	packet.clear();
	packet.append(QString("protocol: %1").arg(TCP_PROTOCOL));
	packet.append(QString("version: %1").arg(TCP_VERSION));
	packet.append(QString("message-type: %1").arg(typespack[TCP_CHAT_MSG_TYPING - 1].tname));
	packet.append(QString("user-uuid: %1").arg(sessuid));
	packet.append(QString("user-name: %1").arg(unick));
	packet.append(QString("typing-mode: %1").arg(QString::number(tmode)));
	packet.append(QString("content-length: 0"));
    
	sendpacket();
    }
}

void TCPSocket::sendChatMessage(const QString &text)
{
    QCString ctext = text.utf8();
    uint ctextlen = ctext.length();
        
    if (sessionestablished)
    {
	TCPServer *par = (TCPServer *)parent();
	QString unick = par->getNick(peerAddress());
	
	CryptEncrypt(skey, 0, true, 0, (unsigned char*)ctext.data(), (DWORD*)&ctextlen, ctextlen);
	
	packet.clear();
	packet.append(QString("protocol: %1").arg(TCP_PROTOCOL));
	packet.append(QString("version: %1").arg(TCP_VERSION));
	packet.append(QString("message-type: %1").arg(typespack[TCP_CHAT_MSG - 1].tname));
	packet.append(QString("user-uuid: %1").arg(sessuid));
	packet.append(QString("user-name: %1").arg(unick));
	packet.append(QString("content-length: %1").arg(QString::number(ctextlen)));    
	sendpacket();
	
	writeBlock( (char*)ctext.data(), ctextlen );
    }
}

void TCPSocket::acceptFile(const QString& sender, const QString& fpath, bool avatar)
{
    vis = !avatar;
    filespath = fpath;
    fullpath = (fpath.endsWith("/")) ? fpath + filename : fpath + "/" + filename;
    QDir d(fpath);
    if (!d.exists())
	d.mkdir(d.absPath());
    	
    file.setName(fullpath);
    if (file.open(IO_WriteOnly))
    {
	filenumber++;
	receivedbytes = 0;
	if (vis)
	{
	    updlim = contentlength / 100;
	    if (updlim == 0) updlim = 1;
	    currtransfer = 0;
	    if (!accepted)
	    {
		rf = new RecvForm(NULL, "File transfer", false);
		rf->setup(remoteusername, fullpath, contentlength);
		connect( rf, SIGNAL(cancel()), SLOT(closeConn()) );
		rf->show();
	    }
	    rf->setCurrentFileNumber(filenumber, totalfiles);
	    rf->setFileName(fullpath);
	    rf->setFileSize(contentlength);
	    rf->setAction(tr("File receiving..."));
	}
	
	accepted = true;
	
	packet.clear();
	packet.append(QString("protocol: %1").arg(TCP_PROTOCOL));
	packet.append(QString("version: %1").arg(TCP_VERSION));
	packet.append(QString("message-type: %1").arg((avatar) ? typespack[TCP_AVATAR_TRANSFER_READY - 1].tname: typespack[TCP_FILE_TRANSFER_READY - 1].tname));
	sendpacket();
    }
}

void TCPSocket::boardMsgRequest(const QHostAddress& addr, const QString& nick, const QString& uid, const QString &bids, Q_UINT16 port)
{
    messagetype = TCP_BOARD_REQUEST;
    
    packet.clear();
    packet.append(QString("protocol: %1").arg(TCP_PROTOCOL));
    packet.append(QString("version: %1").arg(TCP_VERSION));
    packet.append(QString("message-type: %1").arg(typespack[messagetype - 1].tname));
    packet.append(QString("user-uuid: %1").arg(uid));
    packet.append(QString("user-name: %1").arg(nick));
    packet.append(QString("messages-id: %1").arg(bids));
    
    connect( this, SIGNAL(connected()), SLOT(connectDone()) );
    
    connectToHost(addr.toString(), port);
    QTimer::singleShot( 5000, this, SLOT(checkConnection()) );
}

void TCPSocket::requestMissed(const QHostAddress& addr, const QString& nick, const QString& uid, const QString &bids, Q_UINT16 port)
{
    messagetype = TCP_REQUEST_MISSED;

    packet.clear();
    packet.append(QString("protocol: %1").arg(TCP_PROTOCOL));
    packet.append(QString("version: %1").arg(TCP_VERSION));
    packet.append(QString("message-type: %1").arg(typespack[messagetype - 1].tname));
    packet.append(QString("user-uuid: %1").arg(uid));
    packet.append(QString("user-name: %1").arg(nick));
    packet.append(QString("content-length: %1").arg(bids.utf8().length()));
    this->text = bids;

    connect( this, SIGNAL(connected()), SLOT(connectDone()) );

    connectToHost(addr.toString(), port);
    QTimer::singleShot( 5000, this, SLOT(checkConnection()) );
}

void TCPSocket::sendMessageACK(const QHostAddress& addr, const QString& nick, const QString& uid, char status, const QString& answer, const QCString& sig, Q_UINT16 port)
{
    messagetype = TCP_PVTMSG_CONFIRM;
    
    packet.clear();
    packet.append(QString("protocol: %1").arg(TCP_PROTOCOL));
    packet.append(QString("version: %1").arg(TCP_VERSION));
    packet.append(QString("message-type: %1").arg(typespack[messagetype - 1].tname));
    packet.append(QString("user-uuid: %1").arg(uid));
    packet.append(QString("user-name: %1").arg(nick));
    packet.append(QString("auto-answer: %1").arg(answer));
    packet.append(QString("user-status: %1").arg(status));
    packet.append(QString("pvt-message-sig: %1").arg(sig));
    
    connect( this, SIGNAL(connected()), SLOT(connectDone()) );
    
    connectToHost(addr.toString(), port);
    QTimer::singleShot( 5000, this, SLOT(checkConnection()) );
}

void TCPSocket::sendMessage(const QHostAddress& addr, const QString& nick, const QString& uid, const QString &text, Q_UINT16 port)
{
    DWORD slen = 0;
    CRYPTHASH *pCryptHash;
    QByteArray sn;
    QCString snb64;
    QString pid("0000");
 
    CryptCreateHash(CALG_MD5, 0, &pCryptHash);
    if (pCryptHash)
    {
	if (CryptHashData(pCryptHash, (const unsigned char*)text.utf8().data(), text.utf8().length(), 0)) 
	{
	    CryptSignHash(pkey, pCryptHash, 0, 0, (DWORD *)&slen);
    
	    if (slen > 0)
	    {
		sn.resize(slen);

		CryptSignHash(pkey, pCryptHash, 0,(unsigned char*) sn.data(), (DWORD *)&slen);
	    }
	}
	CryptDestroyHash(pCryptHash);
    }
    
    slen = 0;
    encodeBase64((unsigned char *)sn.data(), sn.size(), (char*)snb64.data(), 0, (int*)&slen);
    if (slen > 0)
    {
	snb64.resize(slen + 1);
	encodeBase64((unsigned char *)sn.data(), sn.size(), (char*)snb64.data(), snb64.size(), (int*)&slen);
    }
    
    for (uint i = 1; i < 6; i++)
	pid.append('0' + random()%9);
    
    messagetype = TCP_PVTMSG;
    
    packet.clear();
    packet.append(QString("protocol: %1").arg(TCP_PROTOCOL));
    packet.append(QString("version: %1").arg(TCP_VERSION));
    packet.append(QString("message-type: %1").arg(typespack[messagetype - 1].tname));
    packet.append(QString("user-uuid: %1").arg(uid));
    packet.append(QString("user-name: %1").arg(nick));
    packet.append(QString("pvt-mass: %1").arg(0));
    packet.append(QString("pvt-message-sig: %1").arg(pid));
    packet.append(QString("data-signature: %1").arg(snb64));
    packet.append(QString("content-length: %1").arg(text.utf8().length()));
    this->text = text;
    
    connect( this, SIGNAL(connected()), SLOT(connectDone()) );
    
    connectToHost(addr.toString(), port);
    QTimer::singleShot( 5000, this, SLOT(checkConnection()) );
}

void TCPSocket::httpGet(const QString& addr, Q_UINT16 port, const QString& uid, const QString& frq, const QString &fsd)
{
//    QUrl requrl;
    
    messagetype = TCP_FILE_TRANSFER;
    
    filename = frq;
    fullpath = fsd;
    remoteusername = uid;
    
//    requrl.setProtocol("http");
//    requrl.setHost(addr.toString());
//    requrl.setPort(port);
//    requrl.setFileName(fn);
    
//    std::cout << requrl.toString() << std::endl;
    packet.clear();
//    packet.append(QString("GET %1 HTTP/1.1").arg(requrl.toString()));
    packet.append(QString("GET %1 HTTP/1.1").arg(frq));
    packet.append(QString("Accept: */*"));
    packet.append(QString("Accept-Language: en"));
    packet.append(QString("User-Agent: TriX (X11; Linux i686; U; en)"));
    packet.append(QString("Host: %1:%2").arg(addr, QString::number(port)));
    packet.append(QString("Connection: Keep-Alive"));
    
    httpsession = true;
    
    connect( this, SIGNAL(connected()), SLOT(connectDone()) );
    
    connectToHost(addr, port);
    QTimer::singleShot( 5000, this, SLOT(checkConnection()) );
}

void TCPSocket::sendFile(const QStringList& fn, const QString& descr, const QString& recv, const QString& uuid, const QHostAddress& addr, Q_UINT16 port, bool avatar)
{
    vis = !avatar;
    remoteusername = recv;
    sendtype = messagetype = (avatar) ? TCP_AVATAR_TRANSFER : TCP_FILE_TRANSFER;
    useruuid = uuid;
    filedesc = descr;
    totalfiles = filesleft = fn.count();
    filesnames = fn;

    if (totalfiles > 1)
    {
	for(uint i = 0; i < fn.count(); i++)
	{
	    QFileInfo fi( fn[i] );
	    descfilessize += fi.size();
	    descfilesname.append('\"' + fi.fileName() + "\" ");
	}
    }
        
    if (vis)
    {
	QFileInfo fi(fn[0]);
	sf = new SendForm(NULL, "File transfer", Qt::WDestructiveClose);
	sf->setup(remoteusername, fi.fileName(), fi.size());
	connect( sf, SIGNAL(cancel()), SLOT(closeConn()) );
	sf->show();
	sf->setAction(tr("Connecting to ") + remoteusername + "(" + addr.toString() + ")");
    }

    connect( this, SIGNAL(connected()), SLOT(connectDone()) );
    
    connectToHost(addr.toString(), port);
    QTimer::singleShot( 5000, this, SLOT(checkConnection()) );
}

void TCPSocket::checkConnection()
{
    if (!connectestablished)
	closeConn();
}

void TCPSocket::sendpacket()
{
    QString packetdata = (httpsession) ? packet.join("\r\n") : packet.join("\n");
    if (httpsession) packetdata.append("\r\n\r\n");
    else packetdata.append("\n\n");
    QTextStream os( this );
    os.setEncoding( QTextStream::UnicodeUTF8 );
    os << packetdata;
}

void TCPSocket::writtenBlock(int w)
{
    if (vis)
    {
	currtransfer += w;
	if (currtransfer >= updlim)
	{
	    currtransfer = 0;
	    Q_UINT32 progress = sendbytes / updlim;
	    sf->updateProgressBar( progress );
	}
    }
    if (sendbytes == contentlength)
    {
	disconnect(this, SIGNAL(bytesWritten(int)), this, SLOT(writtenBlock(int)));
	if (!httpsession) emit fileSendFinished(remoteusername, filename);
	if (sf) 
	{
	    sf->updateProgressBar( 100 );
	    sf->setAction(tr("File sended"));
	}
	if (filesleft == 0)
	    QTimer::singleShot( 1000, this, SLOT(closeConn()) );
	return;
    }
    writeNextBlock();
}

void TCPSocket::writeNextBlock()
{
    QByteArray ba(SENDING_BLOCK);
    Q_LONG readed = file.readBlock( ba.data(), ba.size() );
    
    if ( readed == -1) fileReadError();
   
    sendbytes += readed;
    
    if (sendbytes == contentlength)
	initialized = false;

     writeBlock( ba.data(), readed );
}

void TCPSocket::fileReadError()
{
    closeConn();
}

TCPSocket::~TCPSocket()
{
    if (messagetype == TCP_FILE_TRANSFER || messagetype == TCP_FILE_TRANSFER_READY || messagetype == TCP_FILE_TRANSFER_DENY)
    {
	file.close();
	
	if (vis)
	{
	    if (rf) delete rf;
	    if (sf) delete sf;
	}
    }
    if (skey) CryptDestroyKey(skey);
//    std::cout << "TCPSocket destroyed" << std::endl;
}

void TCPSocket::closeConn()
{
    emit closeConnection();
}
