#include "user.h"
#include "global.h"
#include <qpainter.h>
#include <qpixmapcache.h>
#include <qdir.h>

//#include <iostream>

User::User(QListView *list, const QString &n, const QHostAddress &addr, char gend, const QString &fid, const QString &c): QListViewItem(list), gender(gend), ip(addr), color(c)
{
    timer = 5 * random()%10;
    setNick(n);
    setStatus(STATUS_NORMAL);
    if (fid.isEmpty()) 
	setAvatar(fid, true);
    else
    	setAvatar(fid, false);
}

User::~User()
{
}

void User::setNick(const QString &n)
{
    nick = n;
    setText(0, n);
}

bool User::setUserColor(const QString &c)
{
    color = c;
    return true;
}

bool User::setAvatar(const QString& a, bool p)
{
    avatar = a;
    photo = p;
    if ( !QPixmapCache::find(avatar, userpixmap))
    {
	if(userpixmap.load(QString(DATADIR) + "avatars/" + avatar))
	{
	    QPixmapCache::insert(avatar, userpixmap);
	}
	else
	{
	    if(userpixmap.load(QDir::homeDirPath() + "/.trix/avatars/" + avatar))
	    {
		QPixmapCache::insert(avatar, userpixmap);
	    }
	    else
	    {
		if ((int)gender == GENDER_FEMALE) 
		{
		    avatar = "DEFAULT_FEMALE";
		    if(!QPixmapCache::find(avatar, userpixmap))
		    {
			userpixmap.load(QString(DATADIR) + "gfx/female.png");
			QPixmapCache::insert("DEFAULT_FEMALE", userpixmap);
		    }
		}
		else 
		{
		    avatar = "DEFAULT_MALE";
		    if(!QPixmapCache::find(avatar, userpixmap)) 
		    {
			userpixmap.load(QString(DATADIR) + "gfx/male.png");
			QPixmapCache::insert("DEFAULT_MALE", userpixmap);
		    }
		}
	    }
	}
    }
    generateTinyPixmap(listView()->fontMetrics().height() + 4);
    setPixmap(0, userpixmaptiny);
    return true;
}

void User::generateTinyPixmap(int size)
{
    QString a(avatar);
    a.append(":tiny");
    if(!QPixmapCache::find(a, userpixmaptiny))
    {
	QImage img = userpixmap.convertToImage();
	userpixmaptiny = img.smoothScale(size, size, QImage::ScaleMin);
	QPixmapCache::insert(a, userpixmaptiny);
    }
}

int User::height() const
{
    return QMAX(listView()->fontMetrics().height(), curpixmap.height()) + 4;
}

void User::updatePixmap()
{
    QPixmapCache::remove(avatar + ":tiny");
    generateTinyPixmap(listView()->fontMetrics().height());
    setPixmap(0, userpixmaptiny);
}

void User::paintCell(QPainter *p, const QColorGroup &cg, int column, int width, int alignment)
{
    if (!isSelected())
    {
	curpixmap = userpixmaptiny;
	setHeight(height());
	QListViewItem::paintCell(p, cg, column, width, alignment);
    }
    else
    {
	curpixmap = userpixmap;
	setHeight(height());
	p->fillRect(0, 0, listView()->columnWidth(0), height(), QColor(cg.highlight()));
	if ((int)ractive == WIN_ACTIVE) p->drawPixmap(2, 2, curpixmap);
	p->setPen(cg.highlightedText());
	p->drawText(curpixmap.width() + 4, (height() + (p->fontMetrics().ascent() + p->fontMetrics().descent())/2) / 2, text(0));
    }

    if ((int)ractive != WIN_ACTIVE)
    {
	QImage img = curpixmap.convertToImage();
	p->drawImage(2, 2, toGray(img));
    }
    
//    p->setPen("#e7ab17");
//    p->drawRect(2, 2, userpixmap.width(), userpixmap.height());

    switch (status)
    {
    case STATUS_DND: 
	p->drawPixmap(0, height() - 13, QPixmap::fromMimeSource("userdnd.png")); 
	break;
    case STATUS_AWAY: 
	p->drawPixmap(0, height() - 13, QPixmap::fromMimeSource("useraway.png")); 
	break;
    case STATUS_OFFLINE: 
	p->drawPixmap(0, height() - 13, QPixmap::fromMimeSource("useroffline.png"));
	break;
    default: break;
    }
}

// Code below is token from KDE

QImage& User::toGray(QImage &img)
{
    if (img.width() == 0 || img.height() == 0)
	return img;

    if (img.depth() == 32) 
    {
	register uchar * r(img.bits());
	register uchar * g(img.bits() + 1);
	register uchar * b(img.bits() + 2);
	    
	uchar * end(img.bits() + img.numBytes());
	    
	while (r != end) {
		
	    *r = *g = *b = (((*r + *g) >> 1) + *b) >> 1; // (r + b + g) / 3

	    r += 4;
	    g += 4;
	    b += 4;
	}
    }
    else
    {
	for (int i = 0; i < img.numColors(); i++)
	{
	    register uint r = qRed(img.color(i));
	    register uint g = qGreen(img.color(i));
	    register uint b = qBlue(img.color(i));
    
	    register uint gray = (((r + g) >> 1) + b) >> 1;
	    img.setColor(i, qRgba(gray, gray, gray, qAlpha(img.color(i))));
	}
    }
    return img;
}
