#ifndef __USERSLIST_H
#define __USERSLIST_H

#include "user.h"
#include <qlistview.h>
#include <qmap.h>
#include <qhostaddress.h>

class QPopupMenu;
class QStringList;

class UsersList: public QListView
{
    Q_OBJECT
  
    private:
	typedef QMap<QString, User *> UserMap;
	UserMap umap;
	QString nick;
	QPopupMenu *ownmenu;
	QPopupMenu *menu;
	QPopupMenu *usermenu;
	
	void updateTotal();
	void updateToolTip(User* user);
	void mouseMoveEvent ( QMouseEvent * e );

    public:
	enum 
	{
	    ID_MESSAGE = 200,
	    ID_REQINFO,
	    ID_PRVCHAT,
	    ID_BEEP,
	    ID_SAYNICK,
	    ID_FINDBYIP,
	    ID_FILESEND,
	    ID_MASSMESSAGE,
	    ID_REFRESH,
	    ID_EDITPERSONAL
	};
	
    private slots:
	void dummy();
	void popupContextMenu(QListViewItem *it, const QPoint &p, int col);
	void onDoubleClicked(QListViewItem *it, const QPoint &pos, int col);
	void menuAction(int i);
    public slots:
	virtual void clear();
    signals:
	void doubleClickedOnUser(const QString &nick);
	void sendMassMessage();
	void requestInfo(const QString &nick);
	void beep(const QString &nick);
	void privateChat(const QString &nick);
	void refresh();
	void sayNick(const QString &nick);
	void findByIP();
	void filesend(const QString &nick);
	void pingRequest(const QString &);
	void whohereRequest(const QString &);
	void deleteRequest(const QString &);
	void activityDetected();
	void editpersonal();
	void getLocalAutoanswer();

    public:
	UsersList(QWidget *, const char*);

	~UsersList();

	bool exists(const QString &nick) { return umap.contains(nick); }
	User *getUser(const QString &nick) { return exists(nick) ? umap[nick] : NULL; }
	bool newNick(const QString &onick, const QString &nnick);
	User *addUser(const QString &nick, char gender, char status, const QHostAddress &addr, const QString &fid, const QString &color);
	bool delUser(const QString &nick);
	QHostAddress getIP(const QString &nick) { return umap.contains(nick) ? umap[nick]->getIP() : QHostAddress(); }
	QString getNick(const QHostAddress &source);
	QString getToolTip(const QString &unick);
	bool setToolTip(const QString &nick, const QString &tooltip);
	bool setToolTip(User *user, const QString &tooltip) { return user->setToolTip(tooltip); }
	QStringList getNicks(const QString &source);  
	char getStatus(const QString &nick);
	char getStatus(User *user) { return user->getStatus(); }
	bool setStatus(const QString &nick, char s);
	bool setStatus(User *user, char s) { return user->setStatus(s); }
	char getRactive(User *user) { return user->getRactive(); }
	char getRactive(const QString &nick);
	bool setRactive(User *user, char r) { return user->setRactive(r); }
	bool setRactive(const QString &nick, char r);
	char getGender(User *user) { return user->getGender(); }
	char getGender(const QString &nick);
	bool setGender(User *user, char g) { return user->setGender(g); }
	bool setGender(const QString &nick, char g);
	bool setAvatar(User *user, const QString &avatar, bool p = false) { return user->setAvatar(avatar, p); }
	bool setAvatar(const QString &nick, const QString &avatar, bool p = false);
	bool setUserColor(User *user, const QString &color) { return user->setUserColor(color); }
	bool setUserColor(const QString &nick, const QString &color);
	bool setAutoanswer(const QString &nick, const QString &autoanswer);
	QString getAvatar(User *user) { return user->getAvatar(); }
	QString getAvatar(const QString &nick);
	QString getUserColor(const QString &nick);
	QString getAutoanswer(User *user) { return user->getAutoanswer(); }
	QString getAutoanswer(const QString &nick);
	bool AvatarIsAPhoto(const QString &nick);
	bool resetActivityTimer(const QHostAddress &addr);
	void incActivityTimer();
	void scrollToUser(const QString &name);
	void updatePixmaps();
	void setOwnNick(const QString &n) { nick = n; }
	QStringList getAllUsers();
};

#endif

