#include "xxxprotocol.h"
#include "packet.h"
#include "mainform.h"
#include "tcpserver.h"
#include "global.h"
#include "boardmsg.h"
#include "conversions.h"
#include <quuid.h>
#include <qsocketnotifier.h>
#include <qmessagebox.h>
#ifdef WIN32
#include <winsock.h>
#else
#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <net/if_arp.h>
#include <unistd.h>
#include <fcntl.h>
#endif

#ifdef FREEBSD
#include <err.h>
#include <errno.h>
#include <net/if.h>
#include <sys/sysctl.h>
#include <netinet/if_ether.h>
#include <net/route.h>
#include <net/if_dl.h>
#include <net/if_types.h>
#endif

#include <stdlib.h>
//#include <iostream>

#define XXX_ME  'A'
#define XXX_SAY  '2'
#define XXX_NICKCHANGE '3'
#define XXX_AVATARCHANGE 'V'
#define XXX_AVATARREQUEST 'R'
#define XXX_WHOHERE '0'
#define XXX_JOIN   '4'
#define XXX_LEAVE  '5'
#define XXX_MESSAGE  '6'
#define XXX_MASSMESSAGE 'E'
#define XXX_MESSAGEACK  '7'
#define XXX_REMOTEEXEC  '8'
#define XXX_EXECACK  '9'
#define XXX_HEREREQ  'L'
#define XXX_HEREACK  'K'
#define XXX_IMHERE  '1'
#define XXX_NEWTOPIC  'B'
#define XXX_CURRTOPIC  'C'
#define XXX_CHANGESTAT  'D'
#define XXX_INFOREQ  'F'
#define XXX_INFOREQACK  'G'
#define XXX_BEEP  'H'
#define XXX_BEEP_SIGNAL '0'
#define XXX_BEEP_ACK '1'
#define XXX_SOUNDREQ  'I'
#define XXX_PRIVATE  'J'
#define XXX_PRIVATE_JOIN  '0'
#define XXX_PRIVATE_LEAVE  '1'
#define XXX_PRIVATE_SAY  '2'
#define XXX_PRIVATE_ME  '3'
#define XXX_CHANGEACTIV  'M'
#define XXX_REQCHANNELS  'N'
#define XXX_REQCHANNELSACK  'O'
#define XXX_PING 'P'
#define XXX_PINGREQ '0'
#define XXX_PINGACK '1'
#define XXX_FLOODNOTIFY 'Z'
#define XXX_BOARDMSGUPD 'T'
#define XXX_BOARDMSGDEL 'X'
#define XXX_SENDTOPROCESS 'S'

XXXProtocol::XXXProtocol(QObject *parent, QTextCodec *c, const QUuid &u, CRYPTKEY *pCryptKey, bool useutf): QObject(parent), tcpserver(NULL), ansicodec(c), rnotifier(NULL)
{
    DWORD bSize;
    uuid = u;
    utfcodec = QTextCodec::codecForName("UTF-8");
    setUseUTF(useutf);
    
    pkey = pCryptKey;
    
    bSize = 0;

    CryptExportKey(pkey, NULL, PUBLICKEYBLOB, 0, (unsigned char*)pKeyBlob.data(), (DWORD *)&bSize);
    if (bSize > 0)
    {
	pKeyBlob.resize(bSize);
	CryptExportKey(pkey, NULL, PUBLICKEYBLOB, 0, (unsigned char*)pKeyBlob.data(), (DWORD *)&bSize);
    }
}

XXXProtocol::~XXXProtocol()
{
    closeSockets();
}

void XXXProtocol::setAnsiCodec(QTextCodec *c)
{
    ansicodec = c;
}

void XXXProtocol::setUseUTF(bool f)
{
    sndcodec = ((useutf = f)) ? utfcodec : ansicodec;
}

bool XXXProtocol::setupInterface(const QHostAddress &ip, const QString &bcastaddress, const QHostAddress &multicastgroup, uint ttl, uint port, bool usemcast)
{
    struct sockaddr_in saddr;
    struct ip_mreq mcast;
    int flags;
    int opt = 1;
    
    ifok = false;
    
#if defined ( WIN32 )
    WSADATA wsdata;
    
    if (WSAStartup(MAKEWORD(2, 0), &wsdata) != 0)
    {
	return false;
    }
#endif

    
    if ((recvsd = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP)) < 0)
    {
	QMessageBox::warning(0, "TriX", "Socket not initialized!", "OK");
	return false;
    }

    if (!usemcast)
    {
	if ((setsockopt(recvsd, SOL_SOCKET, SO_BROADCAST, &opt, sizeof(int)) < 0) || (setsockopt(recvsd, SOL_SOCKET, SO_REUSEADDR, &opt, sizeof(int)) < 0))
	{
	    QMessageBox::warning(0, "TriX", "Setup broadcast socket failed!", "OK");
	    return false;
	}
	this->multicast= bcastaddress;
    }
    else
    {
	if ((setsockopt(recvsd, SOL_SOCKET, SO_REUSEADDR, &opt, sizeof(int)) < 0) || 
	    (setsockopt(recvsd, IPPROTO_IP, IP_MULTICAST_LOOP, &opt, sizeof(int)) < 0) ||
	    (setsockopt(recvsd, IPPROTO_IP, IP_MULTICAST_TTL, &ttl, sizeof(ttl)) < 0))
	{
	    QMessageBox::warning(0, "TriX", "Setup multicast socket failed!", "OK");
	    return false;
	}
	mcast.imr_multiaddr.s_addr = htonl(multicastgroup.toIPv4Address());
	mcast.imr_interface.s_addr = htonl(ip.toIPv4Address());

	if (setsockopt(recvsd, IPPROTO_IP, IP_MULTICAST_IF, &mcast.imr_interface.s_addr, sizeof(struct in_addr)) < 0)
	{
	    QMessageBox::warning(0, "TriX", "Setup multicast socket failed!", "OK");
	    return false;
	}
	if (setsockopt(recvsd, IPPROTO_IP, IP_ADD_MEMBERSHIP, (char *)&mcast, sizeof(mcast)) < 0)
	{
	    QMessageBox::warning(0, "TriX", "Can't set multicast group!", "OK");
	    return false;
	}
	this->multicast= multicastgroup.toString();
    }
    
#if defined ( WIN32 )
   if (ioctlsocket(recvsd, FIONBIO, (ulong *)&opt) < 0)
       QMessageBox::warning(0, "TriX", "Can't set nonblocking socket!", "OK");
#else
    if ((flags = fcntl(recvsd, F_GETFL, 0)) < 0) 
    {
	printf("[x] Can't get socket flags!\n");
	return false;
    }
    flags |= O_NONBLOCK;
    if (fcntl(recvsd, F_SETFL, flags) < 0) 
    {
	printf("[x] Can't set socket flags!\n");     
	return false;
    }
#endif
 
    memset(&saddr, 0, sizeof(saddr));
    saddr.sin_family = AF_INET;
    saddr.sin_addr.s_addr = INADDR_ANY;
    saddr.sin_port = htons(port);
 
    if (bind(recvsd, (struct sockaddr *)&saddr, sizeof(saddr)) < 0)
    {
	QMessageBox::warning(0, "TriX", "Bind error!", "OK");
	return false;
    }
 
    this->IP = ip;
    this->port = port;

    rnotifier = new QSocketNotifier(recvsd, QSocketNotifier::Read, this);
    connect(rnotifier, SIGNAL(activated(int)), this, SLOT(dataReceived()));
    tcpserver = new TCPServer(ip, port, this, pkey);
    connect(tcpserver, SIGNAL(httpFileRecvFinished(const QString&, const QString&)), SIGNAL(sig_httpFileRecvFinished(const QString&, const QString&)));
    connect(tcpserver, SIGNAL(tcpFileRecvFinished(const QString&, const QString&)), SIGNAL(sig_tcpFileRecvFinished(const QString&, const QString&)));
    connect(tcpserver, SIGNAL(tcpAvatarRecvFinished(const QString&, const QString&)), SIGNAL(sig_tcpAvatarRecvFinished(const QString&, const QString&)));
    connect(tcpserver, SIGNAL(tcpFileSendFinished(const QString&, const QString&)), SIGNAL(sig_tcpFileSendFinished(const QString&, const QString&)));
    connect(tcpserver, SIGNAL(tcpMessageACKReceived(const QHostAddress&, const QString&, const QString&, const QString&, char)), this, SLOT(tcpMessageACKReceivedCheck(const QHostAddress&, const QString&, const QString&, const QString&, char)));
    connect(tcpserver, SIGNAL(tcpPvtMsgReceived(const QHostAddress&, const QString&, const QString&, const QString&, const QString&, const QString&)), this, SLOT(tcpPvtMsgReceivedCheck(const QHostAddress&, const QString&, const QString&, const QString&, const QString&, const QString&)));
    connect(tcpserver, SIGNAL(tcpRecvChatMessage(const QHostAddress&, const QString&)), SIGNAL(sig_tcpRecvChatMessage(const QHostAddress&, const QString&)));
    connect(tcpserver, SIGNAL(tcpChatSessionEstablished(const QString&)), SIGNAL(sig_tcpChatSessionEstablished(const QString&)));
    connect(tcpserver, SIGNAL(tcpCloseChatSession(const QString&)), SIGNAL(sig_tcpCloseChatSession(const QString&)));
    connect(tcpserver, SIGNAL(tcpRecvTypingMode(const QString&, uint)), SIGNAL(sig_tcpRecvTypingMode(const QString&, uint)));
    connect(tcpserver, SIGNAL(tcpBoardUpdated(const QByteArray &)), this, SLOT(tcpBoardUpdated(const QByteArray &)));
 
    ifok = true;
    
    return true;
}

QString XXXProtocol::getHWAddr(const QString& addr, const QString& iface)
{ 
    QString macaddr(tr("Unknown"));
#if defined( FREEBSD )
    struct sockaddr_inarp addr_inarp;

    bzero(&addr_inarp, sizeof(addr_inarp));
    addr_inarp.sin_len = sizeof(addr_inarp);
    addr_inarp.sin_family = AF_INET;
    addr_inarp.sin_addr.s_addr = inet_addr(addr);
    if (addr_inarp.sin_addr.s_addr == INADDR_NONE) 
    {
	sockaddr_dl *sdl = search(addr_inarp.sin_addr.s_addr);
	if (sdl != 0 && sdl->sdl_alen) 
	{
	    if ((sdl->sdl_type == IFT_ETHER || sdl->sdl_type == IFT_L2VLAN) && sdl->sdl_alen == ETHER_ADDR_LEN)
		macaddr.sprintf("%s\n", ether_ntoa((struct ether_addr *)LLADDR(sdl)));
	}
    }
#else
    struct arpreq areq;
    struct sockaddr_in *sin;
    unsigned char *hwaddr;
    
    if (!iface.isEmpty())
    {
	memset(&areq, '\0', sizeof(areq));
	strncpy(areq.arp_dev, iface, 16);

	sin = (struct sockaddr_in *) & (areq.arp_pa);
	sin->sin_family = AF_INET;
    
	if ( inet_aton(addr, &(sin->sin_addr )) != 0 )
	{
	    if ( ioctl( recvsd, SIOCGARP, &areq ) >= 0 )
	    {
		hwaddr = (unsigned char *) areq.arp_ha.sa_data;
		macaddr.sprintf("%02hhX:%02hhX:%02hhX:%02hhX:%02hhX:%02hhX", hwaddr[0], hwaddr[1], hwaddr[2], hwaddr[3], hwaddr[4], hwaddr[5]);
	    }
	}
    }
#endif
    return macaddr;
}

#if defined( FREEBSD )
struct sockaddr_dl * XXXProtocol::search(u_long addr)
{
    int mib[6];
    size_t needed;
    char *lim, *buf, *newbuf, *next;
    struct rt_msghdr *rtm;
    struct sockaddr_inarp *sin2;
    struct sockaddr_dl *sdl = 0;
    char ifname[IF_NAMESIZE];
    int st, found_entry = 0;

    mib[0] = CTL_NET;
    mib[1] = PF_ROUTE;
    mib[2] = 0;
    mib[3] = AF_INET;
    mib[4] = NET_RT_FLAGS;
    mib[5] = RTF_LLINFO;
    if (sysctl(mib, 6, NULL, &needed, NULL, 0) < 0)
	err(1, "route-sysctl-estimate");
    if (needed == 0)	/* empty table */
	return 0;
    buf = NULL;
    for (;;) {
	newbuf = (char*)realloc(buf, needed);
	if (newbuf == NULL) {
	    if (buf != NULL)
		free(buf);
	    errx(1, "could not reallocate memory");
	}
	buf = newbuf;
	st = sysctl(mib, 6, buf, &needed, NULL, 0);
	if (st == 0 || errno != ENOMEM)
	    break;
	needed += needed / 8;
    }
    if (st == -1)
	err(1, "actual retrieval of routing table");
    lim = buf + needed;
    for (next = buf; next < lim; next += rtm->rtm_msglen) {
	rtm = (struct rt_msghdr *)next;
	sin2 = (struct sockaddr_inarp *)(rtm + 1);
	sdl = (struct sockaddr_dl *)((char *)sin2 + SA_SIZE(sin2));
	/*if (rifname && if_indextoname(sdl->sdl_index, ifname) &&  strcmp(ifname, rifname)) continue;*/
	if (addr) {
	    if (addr != sin2->sin_addr.s_addr)
		continue;
	    found_entry = 1;
	}
	//print_entry(sdl, sin2, rtm);
    }
    free(buf);
    return sdl;//(found_entry);
}
#endif

void XXXProtocol::closeSockets()
{
#if defined ( WIN32 )
    closesocket(recvsd);
    WSACleanup();
#else
    close(recvsd);
#endif
    if (rnotifier) delete rnotifier;
    if (tcpserver) delete tcpserver;
    user.clear();
}

void XXXProtocol::dataReceived()
{
    QHostAddress source;
    struct sockaddr_in saddr;
#if defined ( WIN32 ) 
    int slen = sizeof(saddr);
#else
    socklen_t slen = sizeof(struct sockaddr_in);
#endif
 
    char buffer[XXX_MAXPACKETSIZE];
    int len = recvfrom(recvsd, buffer, XXX_MAXPACKETSIZE, 0, (struct sockaddr *)&saddr, &slen);
    source.setAddress(ntohl(saddr.sin_addr.s_addr));
    decode(buffer, len, source);
}

QTextCodec *XXXProtocol::setUserCodecUTF(const QHostAddress &addr, char c)
{
    const Q_INT32 a = addr.toIPv4Address();
    if (!user.contains(a))
	initializeUserSpec(a);
    //TODO set only once per session
    if (!user[a].uid.isEmpty() && user[a].codec == '2') user[a].codec = c;
    return (c == '0') ? ansicodec : utfcodec;
}

void XXXProtocol::setUserVersion(const QHostAddress& addr, Q_UINT32 v)
{
    const Q_INT32 a = addr.toIPv4Address();
    if (!user.contains(a))
	initializeUserSpec(a);
    
    if (user[a].version == 0) user[a].version = v;
}

void XXXProtocol::setUserUuid(const QHostAddress& addr, const QString& u)
{
    const Q_INT32 a = addr.toIPv4Address();
    if (!user.contains(a))
	initializeUserSpec(a);
    if (getUuid(addr).isEmpty()) user[a].uid = u;
}

QString XXXProtocol::getUuid(const QHostAddress &addr)
{
    const Q_INT32 a = addr.toIPv4Address();
    if (!user.contains(a))
	return QString("");
    return user[a].uid;
}

QTextCodec *XXXProtocol::codecForUser(const QHostAddress &addr, bool *ok)
{
    *ok = false;
    const Q_INT32 a = addr.toIPv4Address();
    if (!user.contains(a))	
	return utfcodec;
    if (user[a].uid.isEmpty())
	return utfcodec;

    *ok = true;
    return (user[a].codec == '0') ? ansicodec : utfcodec;
}

Q_UINT16 XXXProtocol::getProtoVersionMajor(const QHostAddress &addr)
{
    const Q_INT32 a = addr.toIPv4Address();
    if (user.contains(a))
	return (user[a].version >> 16) & 0xffff;
    return 0;
}

Q_UINT16 XXXProtocol::getProtoVersionMinor(const QHostAddress &addr)
{
    const Q_INT32 a = addr.toIPv4Address();
    if (user.contains(a))
	return user[a].version & 0xffff;
    return 0;
}

QString XXXProtocol::getProtoVersion(const QHostAddress &addr)
{
    return QString(QString::number(getProtoVersionMajor(addr)) + '.' + QString::number(getProtoVersionMinor(addr)));
}

void XXXProtocol::initializeUserSpec(Q_INT32 a)
{
    user[a].codec = '2'; //undefined
    user[a].version = 0;
    user[a].pPubKey = NULL;
}

bool XXXProtocol::delUser(const QHostAddress &addr)
{
    const Q_INT32 a = addr.toIPv4Address();
    if (!user.contains(a))
	return false;
    CryptDestroyKey(user[a].pPubKey);
    user.remove(a);
    return true;
}

QString XXXProtocol::Uuid(const QUuid& uid)
{
    QString u = uid.toString();
    u.remove('{');
    u.remove('}');
    return u;
}

bool XXXProtocol::get(QCString *str, char *buffer, int &pos, const int p_len)
{
    const int spos = pos;
    for (;pos<p_len; pos++)
    {
	if (buffer[pos] == 0)
	{
	    *str = QCString(buffer + spos);
	    pos++;
	    return true;
	}
    }
    return false;
}

bool XXXProtocol::get(QString *str, char *buffer, int &pos, const int p_len, QTextCodec *tc)
{
    const int spos = pos;
    for (;pos<p_len; pos++)
    {
	if (buffer[pos] == 0)
	{
	    *str = tc->toUnicode(reinterpret_cast<char *>(buffer + spos));
	    pos++;
	    return true;
	}
    }
    return false;
}

bool XXXProtocol::get(char *c, char *buffer, int &pos, const int p_len)
{
    if (pos < p_len)
    {
	*c = buffer[pos++];
	return true;
    }
    return false;
}

bool XXXProtocol::get(Q_UINT16 *s, char *buffer, int &pos, const int p_len)
{
    Q_UINT16 w = 0;
    unsigned char c;
    if (((uint)pos + sizeof(Q_UINT16)) <= (uint)p_len)
    {
	c = buffer[pos++];
	w = c;
	c = buffer[pos++];
	w |= c<<8;
	*s = w;
	return true;
    }
    return false;
}

bool XXXProtocol::get(Q_UINT32 *s, char *buffer, int &pos, const int p_len)
{
    Q_UINT32 dw = 0;
    unsigned char c;
    if (((uint)pos + sizeof(Q_UINT32)) <= (uint)p_len)
    {
	c = buffer[pos++];
	dw = c;
	c = buffer[pos++];
	dw |= c<<8;
	c = buffer[pos++];
	dw |= c<<16;
	c = buffer[pos++];
	dw |= c<<24;
	*s = dw;
	return true;
    }
    return false;
}

bool XXXProtocol::get(QUuid *u, char *buffer, int &pos, const int p_len)
{
    Q_UINT32 l;
    Q_UINT16 w1, w2;
    char b1,b2, b3, b4, b5, b6, b7, b8;
    
    if (pos + 16 < p_len)
    {
	get(&l, buffer, pos, p_len);
	get(&w1, buffer, pos, p_len);
	get(&w2, buffer, pos, p_len);
	get(&b1, buffer, pos, p_len);	
	get(&b2, buffer, pos, p_len); 
	get(&b3, buffer, pos, p_len);
	get(&b4, buffer, pos, p_len);
	get(&b5, buffer, pos, p_len);
	get(&b6, buffer, pos, p_len);
	get(&b7, buffer, pos, p_len);
	get(&b8, buffer, pos, p_len);

	*u = QUuid((uint)l, (ushort)w1, (ushort)w2, (uchar)b1, (uchar)b2, (uchar)b3, (uchar)b4, (uchar)b5, (uchar)b6, (uchar)b7, (uchar)b8);
	return true;
    }
    return false;
}

bool XXXProtocol::get(QByteArray *array, char *buffer, int &pos, const int p_len, Q_INT32 psize)
{
    const int spos = pos;
    if (spos + psize > p_len) return false;
    *array->duplicate(reinterpret_cast<char *>(buffer + spos), psize);
    pos = spos + psize;
    return true;
}

bool XXXProtocol::verifyPacket(const QHostAddress& source, const QByteArray &hash, char *p_data, int p_len)
{
    //TODO Check sign
    return true;
}

QByteArray XXXProtocol::signPacket(Packet *packet)
{
    DWORD slen = 0;
    CRYPTHASH *pCryptHash;
    QByteArray sn;
 
    CryptCreateHash(CALG_MD5, 0, &pCryptHash);
    if (pCryptHash)
    {
	if (CryptHashData(pCryptHash, (const unsigned char*)packet->data(), packet->size(), 0)) 
	{
	    CryptSignHash(pkey, pCryptHash, 0, 0, (DWORD *)&slen);
    
	    if (slen > 0)
	    {
		sn.resize(slen);

		CryptSignHash(pkey, pCryptHash, 0,(unsigned char*) sn.data(), (DWORD *)&slen);
	    }
	}
	CryptDestroyHash(pCryptHash);
    }
    return sn;
}

bool XXXProtocol::decode(char *p_data, int p_len, const QHostAddress &source)
{
    bool ok;
    bool v = false;
    char status, gender, ractive, codepage, c;
    char r, g, b;
    int pos = 0;
    int psize = 0;
    Q_UINT32 version;
    QString from, to, channel, text, fname, host, ip, name, channels, domain, software, time, uid, os, fuid;
    QCString a_to, a_from, a_channel, a_fname;
    QUuid luuid;
    QByteArray hash(128);
       
    if (*p_data++ != 'X') return false;
    
    QCString sig(p_data, 11);	// Signature(9 bytes) + PacketType(1 byte) + NullTerminator(1 byte)
    p_data += 9;		//skip packet signature
    char p = *(p_data++);	//PacketType
    
    if (prevsig[p] == sig)  return false;	//Drop duplicate packets
    prevsig[p] = sig;
     
    p_len -= 11;
	
    QTextCodec *codec = codecForUser(source, &ok);
    if (!ok && p != XXX_WHOHERE && p != XXX_IMHERE && p != XXX_JOIN)
	emit sig_xxx_whohereneeded(source);
//TODO if codec undefined - send codec request and repeat decoding (use QTimer)
//                                                                      ^^^^ done         ^^^^ TODO
    
    emit sig_activityDetected(source);
    
    switch (p)
    {
	case XXX_WHOHERE:
	    get(&a_from, p_data, pos, p_len);
	    get(&codepage, p_data, pos, p_len);
	    codec = setUserCodecUTF(source, codepage);
	    emit sig_xxx_whohere(source, codec->toUnicode(a_from));
	    break;
	case XXX_IMHERE:
	    {
		char icon = 0;
		QCString bids;
		Q_UINT32 tmp32;
		Q_UINT16 tmp16, blen;
		Q_UINT16 year, month, weekday, day, hour, min, sec, msgcnt;
		QColor clr;

		get(&a_to, p_data, pos, p_len);			
		get(&a_from, p_data, pos, p_len);
		get(&status, p_data, pos, p_len);
		if (!get(&ractive, p_data, pos, p_len))
		    return false;
		get(&c, p_data, pos, p_len);
		if (get(&version, p_data, pos, p_len))
		{
		    setUserVersion(source, version);
		    get(&gender, p_data, pos, p_len);
		    get(&luuid, p_data, pos, p_len);
		    setUserUuid(source, Uuid(luuid));
		    get(&c, p_data, pos, p_len);
		    get(&tmp32, p_data, pos, p_len);
		    get(&codepage, p_data, pos, p_len);
		    codec = setUserCodecUTF(source, codepage);
		    get(&r, p_data, pos, p_len);
		    get(&g, p_data, pos, p_len);
		    get(&b, p_data, pos, p_len);
		    get(&c, p_data, pos, p_len);
		    get(&c, p_data, pos, p_len);
		    clr.setRgb((unsigned char)r, (unsigned char)g, (unsigned char)b);
		    if (getProtoVersionMajor(source) == 2)
		    {
//			std::cout << a_from << ":" << p_len - pos << std::endl;
			if ((XXX_PROTOMAJOR == 2) || (p_len - pos > 130)) 
			{
			    get(&year, p_data, pos, p_len);
			    get(&month, p_data, pos, p_len);
			    get(&weekday, p_data, pos, p_len);
			    get(&day, p_data, pos, p_len);
			    get(&hour, p_data, pos, p_len);
			    get(&min, p_data, pos, p_len);
			    get(&sec, p_data, pos, p_len);
			    get(&tmp16, p_data, pos, p_len); //unknown
			    get(&msgcnt, p_data, pos, p_len); // Messages count
			    get(&bids, p_data, pos, p_len); // Messages IDs
			    get(&blen, p_data, pos, p_len); //Blob size - 148 bytes (94)
			    
			    if (user[source.toIPv4Address()].pPubKey == NULL) CryptImportKey((unsigned char*)(p_data + pos), blen, NULL, 0, &user[source.toIPv4Address()].pPubKey);
			    pos += blen;
			    get(&icon, p_data, pos, p_len); // Male : 00...04, Female : 00...02
			}
			psize = pos; 		//Packet size without hash
			if (getProtoVersionMinor(source) == 1)
			{
			    get(&hash, p_data, pos, p_len, 128);
			    get(&c, p_data, pos, p_len);
			    get(&c, p_data, pos, p_len);
			    v = true;
			}
		    }
		    get(&a_fname, p_data, pos, p_len);
		    if (!v || (v && verifyPacket(source, hash, p_data - 1, psize))) emit sig_xxx_imhere(source, codec->toUnicode(a_to), codec->toUnicode(a_from), status, ractive, gender, clr.name(), codec->toUnicode(a_fname), bids);
		}
		else
		    emit sig_xxx_imhere(source, codec->toUnicode(a_to), codec->toUnicode(a_from), status, ractive);
		break;
	    }
	case XXX_SAY:
	    get(&channel, p_data, pos, p_len, codec);
	    get(&from, p_data, pos, p_len, codec);
	    get(&text, p_data, pos, p_len, codec);
	    if (text.isEmpty()) break;
	    psize = pos;
	    if (getProtoVersionMajor(source) == 2)
	    {
		get(&hash, p_data, pos, p_len, 128);
		v = true;
	    }
	    if (!v || (v && verifyPacket(source, hash, p_data - 1, psize))) emit sig_xxx_say(channel, from, text);
	    break;
	case XXX_ME:
	    get(&channel, p_data, pos, p_len, codec);
	    get(&from, p_data, pos, p_len, codec);
	    get(&text, p_data, pos, p_len, codec);
	    if (text.isEmpty()) break;
	    psize = pos;
	    if (getProtoVersionMajor(source) == 2)
	    {
		get(&hash, p_data, pos, p_len, 128);
		v = true;
	    }
	    if (!v || (v && verifyPacket(source, hash, p_data - 1, psize))) emit sig_xxx_me(channel, from, text);
	    break;
	case XXX_CHANGEACTIV:
	    get(&from, p_data, pos, p_len, codec);
	    get(&ractive, p_data, pos, p_len);
	    emit sig_xxx_changeactive(from, ractive);
	    break;
	case XXX_REQCHANNELS:
	    get(&from, p_data, pos, p_len, codec);
	    emit sig_xxx_reqchannels(source, from);
	    break;
	case XXX_REQCHANNELSACK:
	    get(&from, p_data, pos, p_len, codec);
	    get(&channels, p_data, pos, p_len, codec);
	    emit sig_xxx_reqchannelsack(source, from, channels);
	    break;
	case XXX_NICKCHANGE: 
	    get(&from, p_data, pos, p_len, codec);
	    get(&to, p_data, pos, p_len, codec);
	    get(&gender, p_data, pos, p_len);
	    psize = pos;
	    if (getProtoVersionMajor(source) == 2)
	    {
		get(&hash, p_data, pos, p_len, 128);
		v = true;
	    }
	    if (!v || (v && verifyPacket(source, hash, p_data - 1, psize))) emit sig_xxx_nickchange(from, to, gender);
	    break;
	case XXX_AVATARCHANGE: 
	    get(&from, p_data, pos, p_len, codec);
	    get(&fname, p_data, pos, p_len, codec);
	    get(&gender, p_data, pos, p_len);
	    emit sig_xxx_avatarchange(from, fname, gender);
	    break;
	case XXX_AVATARREQUEST:
	    get(&fname, p_data, pos, p_len, codec);
	    emit sig_xxx_avatarrequest(source, fname);
	    break;
	case XXX_BEEP:
	    get(&p, p_data, pos, p_len);
	    switch (p)
	    {
	    case XXX_BEEP_SIGNAL: 
		get(&to, p_data, pos, p_len, codec);
		if (!get(&from, p_data, pos, p_len, codec))
		    return false;
		emit sig_xxx_beep(source, to, from);
		break;
	    case XXX_BEEP_ACK:
		get(&to, p_data, pos, p_len, codec);
		get(&from, p_data, pos, p_len, codec);
		if (!get(&gender, p_data, pos, p_len))
		    return false;
		emit sig_xxx_beepack(to, from, gender);
		break;
	    default: break;
	    }
	    break;
	case XXX_JOIN: 
	{
		char icon = 0;
		QCString bids;
		Q_UINT16 tmp16, blen;
		Q_UINT16 year, month, weekday, day, hour, min, sec, msgcnt;
		QColor clr;
		
		get(&a_from, p_data, pos, p_len);
		get(&a_channel, p_data, pos, p_len);
		get(&status, p_data, pos, p_len);
		if (!get(&gender, p_data, pos, p_len))
		    return false;
		get(&c, p_data, pos, p_len);
		if (a_channel.lower() == QCString(MAIN_CHANNEL).lower())
		{
		    get(&version, p_data, pos, p_len);
		    setUserVersion(source, version);
		    get(&luuid, p_data, pos, p_len);
		    setUserUuid(source, Uuid(luuid));
		    get(&c, p_data, pos, p_len);
		    get(&codepage, p_data, pos, p_len);
		    codec = setUserCodecUTF(source, codepage);
		    get(&r, p_data, pos, p_len);
		    get(&g, p_data, pos, p_len);
		    get(&b, p_data, pos, p_len);
		    clr.setRgb((unsigned char)r, (unsigned char)g, (unsigned char)b);
		    get(&c, p_data, pos, p_len); // 00
		    get(&c, p_data, pos, p_len); // 00
		    if (getProtoVersionMajor(source) == 2)
		    {
			get(&year, p_data, pos, p_len);
			get(&month, p_data, pos, p_len);
			get(&weekday, p_data, pos, p_len);
			get(&day, p_data, pos, p_len);
			get(&hour, p_data, pos, p_len);
			get(&min, p_data, pos, p_len);
			get(&sec, p_data, pos, p_len);
			get(&tmp16, p_data, pos, p_len); //unknown
			get(&msgcnt, p_data, pos, p_len);
			get(&bids, p_data, pos, p_len);
			get(&blen, p_data, pos, p_len); //Blob size - 148 bytes (94)
			if (user[source.toIPv4Address()].pPubKey == NULL) CryptImportKey((unsigned char*)(p_data + pos), blen, NULL, 0, &user[source.toIPv4Address()].pPubKey);
			pos += blen;
			get(&icon, p_data, pos, p_len);
			psize = pos; 		//Packet size without hash
			if (getProtoVersionMinor(source) == 1)
			{
			    get(&hash, p_data, pos, p_len, 128);
			    get(&c, p_data, pos, p_len);
			    get(&c, p_data, pos, p_len);
			    v = true;
			}
		    }
		    get(&a_fname, p_data, pos, p_len);
		    if (!v || (v && verifyPacket(source, hash, p_data - 1, psize)))
			emit sig_xxx_join(source, codec->toUnicode(a_from), codec->toUnicode(a_channel), status, gender, clr.name(), icon, codec->toUnicode(a_fname), bids);
		}
		else
		    emit sig_xxx_join(source, codec->toUnicode(a_from), codec->toUnicode(a_channel), status, gender);
		break;
	    }
	case XXX_LEAVE:
	    get(&from, p_data, pos, p_len, codec);
	    get(&channel, p_data, pos, p_len, codec);
	    get(&gender, p_data, pos, p_len);
	    psize = pos;
	    if (getProtoVersionMajor(source) == 2)
	    {
		get(&hash, p_data, pos, p_len, 128);
		v = true;
	    }
	    if (!v || (v && verifyPacket(source, hash, p_data - 1, psize))) 
		emit sig_xxx_leave(from, channel, gender);
	    break;
	case XXX_PRIVATE:
	    get(&p, p_data, pos, p_len);
	    get(&from, p_data, pos, p_len, codec);
	    get(&to, p_data, pos, p_len, codec);
	    switch (p)
	    {
		case XXX_PRIVATE_JOIN: 
		    get(&gender, p_data, pos, p_len);
		    emit sig_xxx_private_join(from, to, gender);
		    break;
		case XXX_PRIVATE_LEAVE:
		    get(&gender, p_data, pos, p_len);
		    emit sig_xxx_private_leave(from, to, gender);
		    break;
		case XXX_PRIVATE_SAY:
		    if (!get(&text, p_data, pos, p_len, codec))
			return false;
		    emit sig_xxx_private_say(from, to, text);
		    break;
		case XXX_PRIVATE_ME:
		    if (!get(&text, p_data, pos, p_len, codec))
			return false;
		    emit sig_xxx_private_me(from, to, text);
		    break;
		default: break;
	    }
	    break;       
	case XXX_MESSAGE: 
	{
	    get(&from, p_data, pos, p_len, codec);
	    get(&to, p_data, pos, p_len, codec);
	    if (!get(&text, p_data, pos, p_len, codec))
		return false;
	    if ((XXX_PROTOMAJOR == 1) || (getProtoVersionMajor(source) == 1)) emit sig_xxx_message(source, sig, from, to, text);
	    break;
	}
	case XXX_MESSAGEACK: 
	{
		QCString pktsig;
		get(&status, p_data, pos, p_len);
		get(&from, p_data, pos, p_len, codec);
		get(&to, p_data, pos, p_len, codec);
		get(&gender, p_data, pos, p_len);
		get(&text, p_data, pos, p_len, codec);
		if (!get(&pktsig, p_data, pos, p_len))
		    emit sig_xxx_messageack(status, from, to, gender, text);
		else
		    emit sig_xxx_messageack(pktsig, status, from, to, gender, text);
		break;
	}
	case XXX_MASSMESSAGE: 
	{
	    get(&from, p_data, pos, p_len, codec);
	    get(&to, p_data, pos, p_len, codec);
	    if (!get(&text, p_data, pos, p_len, codec))
		return false;
	    emit sig_xxx_massmessage(source, sig, from, to, text); 
	    break;
	}       
	case XXX_HEREREQ:
	    get(&from, p_data, pos, p_len, codec);
	    if (!get(&channel, p_data, pos, p_len, codec))
		return false;
	    emit sig_xxx_herereq(source, from, channel);
	    break;       
	case XXX_HEREACK:
	    get(&to, p_data, pos, p_len, codec);
	    get(&channel, p_data, pos, p_len, codec);
	    get(&from, p_data, pos, p_len, codec);
	    if (!get(&ractive, p_data, pos, p_len))
		return false;
	    emit sig_xxx_hereack(to, channel, from, ractive);
	    break;
	case XXX_PING:
	    {
		int cpos;
		Q_UINT16 tmp16, blen;
		Q_UINT16 year, month, weekday, day, hour, min, sec;
		QCString bids;
		QByteArray pubkey(128);
		
		get(&p, p_data, pos, p_len);
		
		get(&to, p_data, pos, p_len, codec);
		get(&from, p_data, pos, p_len, codec);
		
		cpos = pos;
		get(&time, p_data, pos, p_len, codec);
		pos = cpos + 9;
		
		if (getProtoVersionMajor(source) == 2)
		{
		    get(&year, p_data, pos, p_len);
		    get(&month, p_data, pos, p_len);
		    get(&weekday, p_data, pos, p_len);
		    get(&day, p_data, pos, p_len);
		    get(&hour, p_data, pos, p_len);
		    get(&min, p_data, pos, p_len);
		    get(&sec, p_data, pos, p_len);
		    get(&tmp16, p_data, pos, p_len); //unknown
		    get(&tmp16, p_data, pos, p_len); //unknown
		    get(&bids, p_data, pos, p_len);
		    get(&blen, p_data, pos, p_len); //Blob size
		    if (user[source.toIPv4Address()].pPubKey == NULL) CryptImportKey((unsigned char*)(p_data + pos), blen, NULL, 0, &user[source.toIPv4Address()].pPubKey);
		}

		switch(p)
		{
		case XXX_PINGREQ:
		    emit sig_xxx_ping(source, to, from, time, bids);
		    break;
		case XXX_PINGACK:
		    emit sig_xxx_pong(source, to, from, time, bids);
		    break;
		default: break;
		}
		break;
	    }
	case XXX_INFOREQ: 
	    get(&to, p_data, pos, p_len, codec);
	    if (!get(&from, p_data, pos, p_len, codec))
		return false;
	    emit sig_xxx_inforeq(source, to, from);
	    break;
	case XXX_INFOREQACK:
	    {
		QString fullname, occupation, department, phone, mobile, homepage, email, address;
		get(&to, p_data, pos, p_len, codec);
		get(&from, p_data, pos, p_len, codec);
		get(&host, p_data, pos, p_len, codec);
		get(&name, p_data, pos, p_len, codec);
		get(&ip, p_data, pos, p_len, codec);
		get(&channels, p_data, pos, p_len, codec);
		get(&text, p_data, pos, p_len, codec);
		get(&domain, p_data, pos, p_len, codec);
		get(&os, p_data, pos, p_len, codec);
		get(&software, p_data, pos, p_len, codec);
		if (getProtoVersionMajor(source) == 2)
		{
		    get(&fullname, p_data, pos, p_len, codec);
		    get(&occupation, p_data, pos, p_len, codec);
		    get(&department, p_data, pos, p_len, codec);
		    get(&phone, p_data, pos, p_len, codec);
		    get(&mobile, p_data, pos, p_len, codec);
		    get(&homepage, p_data, pos, p_len, codec);
		    get(&email, p_data, pos, p_len, codec);
		    get(&address, p_data, pos, p_len, codec);
		}
		emit sig_xxx_inforeqack(source, to, from, host, name, ip, channels, text, os, domain, software, codecForUser(source, &ok) == utfcodec, fullname, occupation, department, phone, mobile, homepage, email, address);
		break;
	    }
	case XXX_CURRTOPIC: 
	    if (user.contains(source.toIPv4Address()))
	    {
		get(&to, p_data, pos, p_len, codec);
		get(&channel, p_data, pos, p_len, codec);
		if (!get(&text, p_data, pos, p_len, codec))
		    return false;
		emit sig_xxx_currenttopic(source, to, channel, text);
	    }
	    break;
	case XXX_NEWTOPIC:
	    get(&channel, p_data, pos, p_len, codec);
	    get(&text, p_data, pos, p_len, codec);
	    psize = pos;
	    if (getProtoVersionMajor(source) == 2)
	    {
		get(&hash, p_data, pos, p_len, 128);
		v = true;
	    }
	    if (!v || (v && verifyPacket(source, hash, p_data - 1, psize))) emit sig_xxx_newtopic(source, channel, text);
	    break;
	case XXX_CHANGESTAT:
	    get(&from, p_data, pos, p_len, codec);
	    get(&status, p_data, pos, p_len);
	    get(&gender, p_data, pos, p_len);
	    get(&text, p_data, pos, p_len, codec);
	    emit sig_xxx_changestatus(getUuid(source), from, status, gender, text);
	    break;
	case XXX_FLOODNOTIFY:
	    get(&to, p_data, pos, p_len, codec);
	    get(&from, p_data, pos, p_len, codec);
	    if (!get(&text, p_data, pos, p_len, codec))
		return false;
	    emit sig_xxx_floodnotify(to, from, text);
	    break;
	case XXX_BOARDMSGUPD:
	    {
		Q_UINT32 bid;
		
		get(&uid, p_data, pos, p_len, codec);
		get(&bid, p_data, pos, p_len);
		emit sig_xxx_boardmsgupd(source, uid, bid);
		break;
	    }
	case XXX_BOARDMSGDEL:
	    {
		Q_UINT32 bid;
		
		get(&uid, p_data, pos, p_len, codec);
		get(&bid, p_data, pos, p_len);
		get(&c, p_data, pos, p_len);
		get(&c, p_data, pos, p_len);
		psize = pos;
		if (getProtoVersionMajor(source) == 2)
		{
		    get(&hash, p_data, pos, p_len, 128);
		    v = true;
		}
		if (!v || (v && verifyPacket(source, hash, p_data - 1, psize))) emit sig_xxx_boardmsgdel(source, uid, bid);
		break;
	    }
	case XXX_SENDTOPROCESS:
	    {
		get(&uid, p_data, pos, p_len, utfcodec);
		get(&fname, p_data, pos, p_len, utfcodec);
		emit sig_xxx_sendtoprocess(source, uid, fname);
		break;
	    }
	default: 
	    break;
    }
    return true;
}

void XXXProtocol::sendpacket(const QString &addrlist, Packet *packet)
{
    QHostAddress addr;
    struct sockaddr_in saddr;
    
    if (packet->isOk() && ifok)
    {
	QStringList bcastlist = QStringList::split(",", addrlist);
	for(uint i = 0; i < bcastlist.count(); i++)
	{
	    addr.setAddress(bcastlist[i]);
	    memset(&saddr, 0, sizeof(saddr));
	    saddr.sin_family = AF_INET;
	    saddr.sin_addr.s_addr = htonl(addr.toIPv4Address());
	    saddr.sin_port = htons(port);
	    sendto(recvsd, packet->data(), packet->size(), 0, (struct sockaddr *)&saddr, sizeof(saddr));
	}
    }
}
	 
void XXXProtocol::xxx_say(const QString &channel, const QString &nick, const QString &text)
{
    Packet p;
    int textsize, basesize;
    textsize = text.utf8().length();
    
    // X + signature + PacketID + NickSize + ChnSize  + SignSize= baseSize
    basesize = 15 + nick.utf8().size() + channel.utf8().size();
    if (XXX_PROTOMAJOR == 2) basesize += 128;
    
    if (textsize + basesize > XXX_MAXPACKETSIZE)
    {
	QString textblock;
	int psize, curpos;
	
	curpos = 0;
	psize = XXX_MAXPACKETSIZE - basesize;
	if (psize & 0x01) psize--; 	//Checking for odd to prevent unicode character broke
	
	while(true)
	{
	    if (psize > textsize - curpos) psize = textsize - curpos;
	    textblock = QString::fromUtf8(text.utf8().mid(curpos, psize));
	    
	    p.clear();
	    p.add(XXX_SAY);
	    p.add(sndcodec->fromUnicode(channel));
	    p.add(sndcodec->fromUnicode(nick));
	    p.add(sndcodec->fromUnicode(textblock));
	    if (XXX_PROTOMAJOR == 2) 
		p.addArray(signPacket(&p));
	    p.init();
	    sendpacket(multicast, &p);
	    
	    curpos += psize;
	    if (curpos >= textsize) break;
	}
    }
    else
    {
	p.add(XXX_SAY);
	p.add(sndcodec->fromUnicode(channel));
	p.add(sndcodec->fromUnicode(nick));
	p.add(sndcodec->fromUnicode(text));
	if (XXX_PROTOMAJOR == 2) 
	    p.addArray(signPacket(&p));
	p.init();
	sendpacket(multicast, &p);
    }
}

void XXXProtocol::xxx_me(const QString &channel, const QString &nick, const QString &text)
{
    Packet p;
    p.add(XXX_ME);
    p.add(sndcodec->fromUnicode(channel));
    p.add(sndcodec->fromUnicode(nick));
    p.add(sndcodec->fromUnicode(text));
    if (XXX_PROTOMAJOR == 2) 
	p.addArray(signPacket(&p));
    p.init();
    sendpacket(multicast, &p);
}

void XXXProtocol::xxx_whohere(const QString &updater)
{
    Packet p;
    p.init();
    p.add(XXX_WHOHERE);
    p.add(sndcodec->fromUnicode(updater));
    p.add((useutf ? '1' : '0'));
    sendpacket(multicast, &p);
}

void XXXProtocol::xxx_whohere(const QHostAddress &addr, const QString &updater)
{
    Packet p;
    p.init();
    p.add(XXX_WHOHERE);
    p.add(sndcodec->fromUnicode(updater));
    p.add((useutf ? '1' : '0'));
    sendpacket(addr.toString(), &p);
}

void XXXProtocol::xxx_imhere(const QHostAddress &addr, const QString &updater, const QString &responder, char status, char gender, char winstate,  const QString &fname, QColor color)
{
    Packet p;
    p.init();
    p.add(XXX_IMHERE);
    p.add(sndcodec->fromUnicode(updater));
    p.add(sndcodec->fromUnicode(responder));
    p.add(status);
    p.add(winstate);
    p.add(0);
    p.add(XXX_PROTOMINOR);
    p.add(0);
    p.add(XXX_PROTOMAJOR);
    p.add(0);
    p.add(gender);
    p.add(uuid);
    p.add(0);
    p.add(0);
    p.add(0);
    p.add(0);
    p.add(0);
    p.add((useutf ? '1' : '0'));
    p.add(Q_INT8(color.red())); //R
    p.add(Q_INT8(color.green())); //G
    p.add(Q_INT8(color.blue())); //B
    p.add(0);
    p.add(0);
    if (XXX_PROTOMAJOR == 2)
    {
	p.add16(0); //year
	p.add16(0); //month
	p.add16(0); //weekday
	p.add16(0); //day
	p.add16(0); //hour
	p.add16(0); //minute
	p.add16(0); //seconds
	p.add16(0);
	p.add16(0); //number of messages
	p.add(0); //messages IDs, separated by ','
	p.add16(pKeyBlob.size());
	p.addArray(pKeyBlob);
	p.add(1); //icon
    }
    p.add(sndcodec->fromUnicode(fname));
    sendpacket(addr.toString(), &p);
}

void XXXProtocol::xxx_beep(const QHostAddress &addr, const QString& to, const QString& from)
{
    Packet p;
    p.init();
    p.add(XXX_BEEP);
    p.add(XXX_BEEP_SIGNAL);
    p.add(sndcodec->fromUnicode(to));
    p.add(sndcodec->fromUnicode(from));
    sendpacket(addr.toString(), &p);
}

void XXXProtocol::xxx_beepack(const QHostAddress &addr, const QString& to, const QString& from, char gender)
{
    Packet p;
    p.init();
    p.add(XXX_BEEP);
    p.add(XXX_BEEP_ACK);
    p.add(sndcodec->fromUnicode(to));
    p.add(sndcodec->fromUnicode(from));
    p.add(gender);
    sendpacket(addr.toString(), &p);
}

void XXXProtocol::xxx_join(const QString &who, const QString &channel, char status, char gender, const QString &fname, QColor color)
{ 
    Packet p;
    p.init();
    p.add(XXX_JOIN);
    p.add(sndcodec->fromUnicode(who));
    p.add(sndcodec->fromUnicode(channel));
    p.add(status);
    p.add(gender);
    p.add(0);
    if (channel.lower() == QString(MAIN_CHANNEL).lower()) {
	p.add(XXX_PROTOMINOR);
	p.add(0);
	p.add(XXX_PROTOMAJOR);
	p.add(0);
	p.add(uuid);
	p.add(0);
	p.add((useutf ? '1' : '0'));
	p.add(Q_INT8(color.red())); //R
	p.add(Q_INT8(color.green())); //G
	p.add(Q_INT8(color.blue())); //B
	p.add(0);
	p.add(0);
	if (XXX_PROTOMAJOR == 2)
	{
	    p.add16(0); //year
	    p.add16(0); //month
	    p.add16(0); //weekday
	    p.add16(0); //day
	    p.add16(0); //hour
	    p.add16(0); //minute
	    p.add16(0); //seconds
	    p.add16(0);
	    p.add16(0); //number of messages
	    p.add(0); //messages IDs, separated by ','
	    p.add16(pKeyBlob.size());
	    p.addArray(pKeyBlob);
	    p.add(1); //icon
	}
	p.add(sndcodec->fromUnicode(fname));
    }
    sendpacket(multicast, &p);
}

void XXXProtocol::xxx_leave(const QString &who, const QString &channel, char gender)
{
    Packet p;
    p.add(XXX_LEAVE);
    p.add(sndcodec->fromUnicode(who));
    p.add(sndcodec->fromUnicode(channel));
    p.add(gender);
    if (XXX_PROTOMAJOR == 2) 
	p.addArray(signPacket(&p));
    p.init();
    sendpacket(multicast, &p);
}

void XXXProtocol::xxx_private_join(const QHostAddress &addr, const QString &from, const QString &to, char gender)
{
    Packet p;
    p.init();
    p.add(XXX_PRIVATE);
    p.add(XXX_PRIVATE_JOIN);
    p.add(sndcodec->fromUnicode(from));
    p.add(sndcodec->fromUnicode(to));
    p.add(gender);
    sendpacket(addr.toString(), &p);
}

void XXXProtocol::xxx_private_leave(const QHostAddress &addr, const QString &from, const QString &to, char gender)
{
    Packet p;
    p.init();
    p.add(XXX_PRIVATE);
    p.add(XXX_PRIVATE_LEAVE);
    p.add(sndcodec->fromUnicode(from));
    p.add(sndcodec->fromUnicode(to));
    p.add(gender);
    sendpacket(addr.toString(), &p);
}

void XXXProtocol::xxx_private_say(const QHostAddress &addr, const QString &from, const QString &to, const QString &text)
{
    Packet p;
    int textsize, basesize;
    textsize = text.utf8().size();
    basesize = 15 + from.utf8().size() + to.utf8().size();
    
    if (textsize + basesize > XXX_MAXPACKETSIZE)
    {
	QString textblock;
	int psize, curpos;
	
	curpos = 0;
	psize = XXX_MAXPACKETSIZE - basesize;
	if (psize & 0x01) psize--; 	//Checking for odd to prevent unicode character broke
	
	while(true)
	{
	    if (psize > textsize - curpos) psize = textsize - curpos;
	    textblock = QString::fromUtf8(text.utf8().mid(curpos, psize));
	    
	    p.clear();
	    p.add(XXX_PRIVATE);
	    p.add(XXX_PRIVATE_SAY);
	    p.add(sndcodec->fromUnicode(from));
	    p.add(sndcodec->fromUnicode(to));
	    p.add(sndcodec->fromUnicode(textblock));

	    sendpacket(addr.toString(), &p);
	    
	    curpos += psize;
	    if (curpos >= textsize) break;
	}
    }
    else
    {
	p.init();
	p.add(XXX_PRIVATE);
	p.add(XXX_PRIVATE_SAY);
	p.add(sndcodec->fromUnicode(from));
	p.add(sndcodec->fromUnicode(to));
	p.add(sndcodec->fromUnicode(text));
	sendpacket(addr.toString(), &p);
    }
}

void XXXProtocol::xxx_private_me(const QHostAddress &addr, const QString &from, const QString &to, const QString &text)
{
    Packet p;
    p.init();
    p.add(XXX_PRIVATE);
    p.add(XXX_PRIVATE_ME);
    p.add(sndcodec->fromUnicode(from));
    p.add(sndcodec->fromUnicode(to));
    p.add(sndcodec->fromUnicode(text));
    sendpacket(addr.toString(), &p);
}

void XXXProtocol::xxx_message(const QHostAddress &addr, const QString &from, const QString &to, const QString &text)
{
    QCString sig;
    if ((getProtoVersionMajor(addr) == 2) && (XXX_PROTOMAJOR == 2)) 
    {
	tcpserver->sendMessage(addr, from, Uuid(uuid), text, port);
    }
    else
    {
	Packet p;
	p.init();
	p.add(XXX_MESSAGE);
	p.add(sndcodec->fromUnicode(from));
	p.add(sndcodec->fromUnicode(to));
	p.add(sndcodec->fromUnicode(text));
	sendpacket(addr.toString(), &p);
    }
}

void XXXProtocol::xxx_massmessage(const QHostAddress &addr, const QString &from, const QString &to, const QString &text)
{
    Packet p;
    p.init();
    p.add(XXX_MASSMESSAGE);
    p.add(sndcodec->fromUnicode(from));
    p.add(sndcodec->fromUnicode(to));
    p.add(sndcodec->fromUnicode(text));
    sendpacket(addr.toString(), &p);
}

void XXXProtocol::xxx_messageack(const QHostAddress &addr, const QCString &sig, char status, const QString &from, const QString &to, char gender, const QString &answer)
{
    if ((getProtoVersionMajor(addr) == 2) && (XXX_PROTOMAJOR == 2)) 
    {
	tcpserver->sendMessageACK(addr, to, Uuid(uuid), status, answer, sig, port);
    }
    else
    {
	Packet p;
	p.init();
	p.add(XXX_MESSAGEACK);
	p.add(status);
	p.add(sndcodec->fromUnicode(from));
	p.add(sndcodec->fromUnicode(to));
	p.add(gender);
	p.add(sndcodec->fromUnicode(answer));
	p.addArray(sig);
	p.add(0);
	sendpacket(addr.toString(), &p);
    }
}

void XXXProtocol::xxx_herereq(const QString &from, const QString &channel)
{
    Packet p;
    p.init();
    p.add(XXX_HEREREQ);
    p.add(sndcodec->fromUnicode(from));
    p.add(sndcodec->fromUnicode(channel));
    sendpacket(multicast, &p);
}

void XXXProtocol::xxx_hereack(const QHostAddress &addr, const QString &to, const QString &channel, const QString &from, char ra)
{
    Packet p;
    p.init();
    p.add(XXX_HEREACK);
    p.add(sndcodec->fromUnicode(to));
    p.add(sndcodec->fromUnicode(channel));
    p.add(sndcodec->fromUnicode(from));
    p.add(ra);
    sendpacket(addr.toString(), &p);
}

void XXXProtocol::xxx_inforeq(const QHostAddress &addr, const QString &to, const QString &from)
{
    Packet p;
    p.init();
    p.add(XXX_INFOREQ);
    p.add(sndcodec->fromUnicode(to));
    p.add(sndcodec->fromUnicode(from));
    sendpacket(addr.toString(), &p);
}

void XXXProtocol::xxx_inforeqack(const QHostAddress &addr, const QString &to, const QString &from, const QString &hostname, const QString &name, const QString &ip, const QString &channels, const QString &answer, const QString &os, const QString &software, const QString &domain, const QString &fullname, const QString &occupation, const QString &department, const QString &phone, const QString &mobile, const QString &homepage, const QString& email, const QString &address)
{
    Packet p;
    p.init();
    p.add(XXX_INFOREQACK);
    p.add(sndcodec->fromUnicode(to));
    p.add(sndcodec->fromUnicode(from));
    p.add(sndcodec->fromUnicode(hostname));
    p.add(sndcodec->fromUnicode(name));
    p.add(sndcodec->fromUnicode(ip));
    p.add(sndcodec->fromUnicode(channels));
    p.add(sndcodec->fromUnicode(answer));
    p.add(sndcodec->fromUnicode(domain));
    p.add(sndcodec->fromUnicode(os));
    p.add(sndcodec->fromUnicode(software));
    p.add(sndcodec->fromUnicode(fullname));
    p.add(sndcodec->fromUnicode(occupation));
    p.add(sndcodec->fromUnicode(department));
    p.add(sndcodec->fromUnicode(phone));
    p.add(sndcodec->fromUnicode(mobile));
    p.add(sndcodec->fromUnicode(homepage));
    p.add(sndcodec->fromUnicode(email));
    p.add(sndcodec->fromUnicode(address));
    sendpacket(addr.toString(), &p);
}

void XXXProtocol::xxx_nickchange(const QString &from, const QString &to, char gender)
{
    Packet p;
    p.add(XXX_NICKCHANGE);
    p.add(sndcodec->fromUnicode(from));
    p.add(sndcodec->fromUnicode(to));
    p.add(gender);
    if (XXX_PROTOMAJOR == 2) 
	p.addArray(signPacket(&p));
    p.init();
    sendpacket(multicast, &p);
}

void XXXProtocol::xxx_avatarchange(const QString &from, const QString &fname, char gender)
{
    Packet p;
    p.init();
    p.add(XXX_AVATARCHANGE);
    p.add(sndcodec->fromUnicode(from));
    p.add(sndcodec->fromUnicode(fname));
    p.add(gender);
    sendpacket(multicast, &p);
}

void XXXProtocol::xxx_changeactive(const QString &from, char ractive)
{
    Packet p;
    p.init();
    p.add(XXX_CHANGEACTIV);
    p.add(sndcodec->fromUnicode(from));
    p.add(ractive);
    sendpacket(multicast, &p);
}

void XXXProtocol::xxx_reqchannels(const QString &from)
{
    Packet p;
    p.init();
    p.add(XXX_REQCHANNELS);
    p.add(sndcodec->fromUnicode(from));
    sendpacket(multicast, &p);
}

void XXXProtocol::xxx_reqchannelsack(const QHostAddress& addr, const QString &from, const QString& chlist)
{
    Packet p;
    p.init();
    p.add(XXX_REQCHANNELSACK);
    p.add(sndcodec->fromUnicode(from));
    p.add(sndcodec->fromUnicode(chlist));
    sendpacket(addr.toString(), &p);
}

void XXXProtocol::xxx_changestatus(const QString &who, char status, char gender, const QString &answer)
{
    Packet p;
    p.init();
    p.add(XXX_CHANGESTAT);
    p.add(sndcodec->fromUnicode(who));
    p.add(status);
    p.add(gender);
    p.add(sndcodec->fromUnicode(answer));
    sendpacket(multicast, &p);
}

void XXXProtocol::xxx_currenttopic(const QHostAddress &addr, const QString &to, const QString &channel, const QString &topic)
{
    if (!topic.isEmpty())
    {
	Packet p;
	p.init();
	p.add(XXX_CURRTOPIC);
	p.add(sndcodec->fromUnicode(to));
	p.add(sndcodec->fromUnicode(channel));
	p.add(sndcodec->fromUnicode(topic));
	sendpacket(addr.toString(), &p);
    }
}

void XXXProtocol::xxx_newtopic(const QString &channel, const QString &topic)
{
    Packet p;
    p.add(XXX_NEWTOPIC);
    p.add(sndcodec->fromUnicode(channel));
    p.add(sndcodec->fromUnicode(topic));
    if (topic.isEmpty()) p.add(0);
    if (XXX_PROTOMAJOR == 2) 
	p.addArray(signPacket(&p));
    p.init();
    sendpacket(multicast, &p);
}

void XXXProtocol::xxx_ping(const QHostAddress &addr, const QString &to, const QString &from, const QString &time)
{
    Packet p;
    p.init();
    p.add(XXX_PING);
    p.add(XXX_PINGREQ);
    p.add(sndcodec->fromUnicode(to));
    p.add(sndcodec->fromUnicode(from));
    p.add(sndcodec->fromUnicode(time));
    if (XXX_PROTOMAJOR == 2)
    {
	p.add16(0); //year
	p.add16(0); //month
	p.add16(0); //weekday
	p.add16(0); //day
	p.add16(0); //hour
	p.add16(0); //minute
	p.add16(0); //seconds
	p.add16(0); //unknown
	p.add16(0); //unknown
	p.add(0); //board messages
	p.add16(pKeyBlob.size());
	p.addArray(pKeyBlob);
    }
    sendpacket(addr.toString(), &p);
}

void XXXProtocol::xxx_pong(const QHostAddress &addr, const QString &to, const QString &from, const QString &time)
{
    Packet p;
    p.init();
    p.add(XXX_PING);
    p.add(XXX_PINGACK);
    p.add(sndcodec->fromUnicode(to));
    p.add(sndcodec->fromUnicode(from));
    p.add32(0);
    p.add32(0);
    p.add(0);

    if (XXX_PROTOMAJOR == 2)
    {
	p.add16(0); //year
	p.add16(0); //month
	p.add16(0); //weekday
	p.add16(0); //day
	p.add16(0); //hour
	p.add16(0); //minute
	p.add16(0); //seconds
	p.add16(0); //unknown
	p.add16(0); //unknown
	p.add(0); //board messages
	p.add16(pKeyBlob.size());
	p.addArray(pKeyBlob);
    }
    sendpacket(addr.toString(), &p);
}

void XXXProtocol::xxx_floodnotify(const QHostAddress &addr, const QString &to, const QString &from, int seconds)
{
    Packet p;
    p.init();
    p.add(XXX_FLOODNOTIFY);
    p.add(sndcodec->fromUnicode(to));
    p.add(sndcodec->fromUnicode(from));
    p.add(sndcodec->fromUnicode(QString::number(seconds)));
    sendpacket(addr.toString(), &p);
}

void XXXProtocol::xxx_httpGet(const QString& addr, const QString& frq, const QString& fsd)
{
    QHostAddress ipaddr;
    ipaddr.setAddress(addr);
    tcpserver->httpGet(addr, port, getUuid(ipaddr), frq, fsd);
}

void XXXProtocol::xxx_httpGet(const QString& addr, int p, const QString &comment, const QString& frq, const QString& fsd)
{
    tcpserver->httpGet(addr, p, comment, frq, fsd);
}

void XXXProtocol::xxx_tcpfilesend(const QStringList& filesnames, const QString& descr, const QString& receiver, const QHostAddress& addr, bool avatar)
{
    tcpserver->sendFile(filesnames, descr, receiver, Uuid(uuid), addr, port, avatar);
}

void XXXProtocol::xxx_avatarrequest(const QHostAddress& addr, const QString& avatar)
{
    Packet p;
    p.init();
    p.add(XXX_AVATARREQUEST);
    p.add(sndcodec->fromUnicode(avatar));
    sendpacket(addr.toString(), &p);
}

void XXXProtocol::xxx_boardmsgupdate(Q_UINT32 id)
{
    Packet p;
    p.init();
    p.add(XXX_BOARDMSGUPD);
    p.add(sndcodec->fromUnicode(Uuid(uuid)));
    p.add32(id);
    p.add(0);
    sendpacket(multicast, &p);
}

void XXXProtocol::xxx_boardmsgrequest(const QHostAddress& addr, const QString& nick, const QString &bids)
{
     tcpserver->boardMsgRequest(addr, nick, Uuid(uuid), bids, port);
}

void XXXProtocol::xxx_requestmissed(const QHostAddress& addr, const QString& nick, const QString &bids)
{
     tcpserver->requestMissed(addr, nick, Uuid(uuid), bids, port);
}

bool XXXProtocol::xxx_createChatSession(const QHostAddress& addr, const QString &uid)
{
    if (tcpserver->createChatSession(addr, port, uid, user[addr.toIPv4Address()].pPubKey) == NULL)
	return false;
    return true;
}

bool XXXProtocol::xxx_isSessionEstablished(const QString &uid)
{
    return tcpserver->findChatSession(uid) ? true: false;
}

void XXXProtocol::xxx_sendChatMessage(const QString &uid, const QString &text)
{
    tcpserver->sendChatMessage(uid, text);
}

void XXXProtocol::xxx_sendTypingMode(const QString &uid, uint tmode)
{
    tcpserver->sendTypingMode(uid, tmode);
}

void XXXProtocol::xxx_sleepChatSession(const QString &uid)
{
    tcpserver->sleepChatSession(uid);
}

void XXXProtocol::xxx_wakeupChatSession(const QString &uid)
{
    tcpserver->wakeupChatSession(uid);
}

void XXXProtocol::tcpMessageACKReceivedCheck(const QHostAddress& addr, const QString& from, const QString& usersig, const QString& answer, char status)
{
    emit sig_tcpMessageACKReceived(from, answer, status);
}

void XXXProtocol::tcpPvtMsgReceivedCheck(const QHostAddress& addr, const QString& uid, const QString& from, const QString& msig, const QString& dsig, const QString& text)
{
    if (user[addr.toIPv4Address()].uid == uid) emit sig_tcpPvtMsgReceived(from, text);
}

void XXXProtocol::tcpBoardUpdated(const QByteArray &ba)
{
    int offset = 0;

    while (readBMessage(ba, offset));
}

bool XXXProtocol::readBMessage(const QByteArray &ba, int &offset)
{
    bool ok;
    char a, b, c, d;
    Q_UINT32 bmsize, magic2, id, spos;
    Q_UINT16 year, mnth, wd, day, hour, min, sec, magic1, p;
    QDate pdate, edate;
    QTime ptime, etime;
    QString author, topic, text, ip;
    QUuid uid;
    QByteArray sign(128);
    
    spos = offset;
    
    get(&bmsize, ba.data(), offset, ba.size());
    get(&author, ba.data(), offset, ba.size(), utfcodec);
    get(&topic, ba.data(), offset, ba.size(), utfcodec);
    get(&text, ba.data(), offset, ba.size(), utfcodec);
    get(&uid, ba.data(), offset, ba.size());
    get(&year, ba.data(), offset, ba.size());
    get(&mnth, ba.data(), offset, ba.size());
    get(&wd, ba.data(), offset, ba.size());
    get(&day, ba.data(), offset, ba.size());
    get(&hour, ba.data(), offset, ba.size());
    get(&min, ba.data(), offset, ba.size());
    get(&sec, ba.data(), offset, ba.size());
    pdate.setYMD(year, mnth, day);
    ptime.setHMS(hour, min, sec);
    get(&magic1, ba.data(), offset, ba.size());
    get(&year, ba.data(), offset, ba.size());
    get(&mnth, ba.data(), offset, ba.size());
    get(&wd, ba.data(), offset, ba.size());
    get(&day, ba.data(), offset, ba.size());
    get(&hour, ba.data(), offset, ba.size());
    get(&min, ba.data(), offset, ba.size());
    get(&sec, ba.data(), offset, ba.size());
    edate.setYMD(year, mnth, day);
    etime.setHMS(hour, min, sec);
    get(&magic2, ba.data(), offset, ba.size());
    get(&p, ba.data(), offset, ba.size());
    get(&a, ba.data(), offset, ba.size());
    get(&b, ba.data(), offset, ba.size());
    get(&c, ba.data(), offset, ba.size());
    get(&d, ba.data(), offset, ba.size());
    ip = QString("%1.%2.%3.%4").arg(QString::number((unsigned char)a, 10), QString::number((unsigned char)b, 10), QString::number((unsigned char)c, 10), QString::number((unsigned char)d, 10));
    offset += 120;
    get(&sign, ba.data(), offset, ba.size(), 128);
    ok = get(&id, ba.data(), offset, ba.size());
    
    if (offset - spos != bmsize) return false;
    if (ok) 
    {
//TODO Add signature verification	
	BoardMsg bmsg(bmsize, author, topic, text, Uuid(uid), pdate.toString(), ptime.toString(), magic1, edate.toString(), etime.toString(), magic2, p, ip, Conversions::raw2str((const unsigned char*)sign.data(), sign.size()), id);
	emit sig_addToBoard(&bmsg);
	if (offset < ba.size()) return true;
    }
    return false;
}

QString XXXProtocol::getNick(const QHostAddress& addr)
{
    MainForm *par = (MainForm*)parent();
    return par->getNick(addr);
}

QHostAddress XXXProtocol::getIP(const QString &uid)
{
    QHostAddress addr;
    for (QMap<Q_INT32, UserSpec>::iterator it = user.begin(); it!=user.end(); ++it)
    {
	if (QString::compare(it.data().uid, uid) == 0) 
	{
	    addr.setAddress(it.key());
	    break;
	}
    }
    return addr;
}

QString XXXProtocol::getDownloadPath()
{
    MainForm *par = (MainForm*)parent();
    return par->getDownloadPath();
}

bool XXXProtocol::getAcceptWOPrompt()
{
    MainForm *par = (MainForm*)parent();
    return par->getAcceptWOPrompt();
}

int XXXProtocol::getMaxAvatarSize()
{
    MainForm *par = (MainForm*)parent();
    return par->getMaxAvatarSize();
}

char XXXProtocol::getLocalStatus()
{
    MainForm *par = (MainForm*)parent();
    return par->getLocalStatus();
}

char XXXProtocol::getLocalGender()
{
    MainForm *par = (MainForm*)parent();
    return par->getLocalGender();
}

QString XXXProtocol::getLocalNick()
{
    MainForm *par = (MainForm*)parent();
    return par->getLocalNick();
}

QString XXXProtocol::getLocalUid()
{
      return Uuid(uuid);
}
	
QString XXXProtocol::getLocalAnswer()
{
    MainForm *par = (MainForm*)parent();
    return par->getLocalAnswer();
}

QString XXXProtocol::getLocalAvatar()
{
    MainForm *par = (MainForm*)parent();
    return par->getLocalAvatar();
}

QString XXXProtocol::getLocalFile() //TODO add filename
{
    MainForm *par = (MainForm*)parent();
    return par->getLocalFile();
}
