#ifndef __XXXPROTOCOL_H
#define __XXXPROTOCOL_H

#include "global.h"
#include "tcpsocket.h"
#include "crypto.h"

#ifdef WIN32
#include <winsock.h>
#endif
#include <qobject.h>
#include <qhostaddress.h>
#include <qtextcodec.h>
#include <qmemarray.h>
#include <qmap.h>
#include <quuid.h>

class QSocketNotifier;
class BoardMsg;
class TCPServer;
class Packet;

class XXXProtocol: public QObject
{
    Q_OBJECT
  
private:
    bool useutf;
    bool ifok;		//interface state flag

    uint port;
    int recvsd;		//UDP socket
    
#if defined( FREEBSD )
    struct sockaddr_dl *search(u_long addr);
#endif
    
    struct UserSpec {
	char codec;
	Q_UINT32 version;
	QString uid;
	CRYPTKEY *pPubKey;
    };
    
    CRYPTKEY *pkey;
    int siglen;
    
    TCPServer *tcpserver;
    QString multicast;
    QHostAddress IP;
    QTextCodec *ansicodec, *utfcodec, *sndcodec;
    QSocketNotifier *rnotifier;
    QMap<Q_INT32, UserSpec> user;
    QMap<char, QCString> prevsig;
    QUuid uuid;
    QByteArray pKeyBlob;
  
    QTextCodec *setUserCodecUTF(const QHostAddress &addr, char c);
    QTextCodec *codecForUser(const QHostAddress &addr, bool *ok);
    
    void initializeUserSpec(Q_INT32 a);
    bool verifyPacket(const QHostAddress& source, const QByteArray& hash, char *p_data, int p_len);
    QByteArray signPacket(Packet *packet);
  
    void sendpacket(const QString &addrlist, Packet *packet);

    bool decode(char *p_data, int p_len, const QHostAddress &source);
  
  private slots:
      void dataReceived();
      void tcpMessageACKReceivedCheck(const QHostAddress& addr, const QString& from, const QString& usersig, const QString& answer, char status);
      void tcpPvtMsgReceivedCheck(const QHostAddress& addr, const QString& uid, const QString& from, const QString& msig, const QString& dsig, const QString& text);
      void tcpBoardUpdated(const QByteArray &ba);

  signals:
      void sig_xxx_say(const QString &channel, const QString &from, const QString &text);
      void sig_xxx_me(const QString &channel, const QString &from, const QString &text);     
      void sig_xxx_whohere(const QHostAddress &source, const QString &updater);
      void sig_xxx_imhere(const QHostAddress &source, const QString &updater, const QString &responder, char status, char winstate);
      void sig_xxx_imhere(const QHostAddress &source, const QString &updater, const QString &responder, char status, char winstate, char gender, const QString &cname, const QString &fname, const QString &bids);
      void sig_xxx_beep(const QHostAddress &source, const QString &to, const QString &from);
      void sig_xxx_beepack(const QString &to, const QString &from, char gender);
      void sig_xxx_join(const QHostAddress &source, const QString&, const QString&, char, char, const QString &cname, char, const QString &fname, const QString &bids);
      void sig_xxx_join(const QHostAddress &source, const QString&, const QString&, char, char);
      void sig_xxx_leave(const QString &who, const QString &channel, char gender);
      void sig_xxx_private_join(const QString &from, const QString &to, char gender);
      void sig_xxx_private_leave(const QString &from, const QString &to, char gender);
      void sig_xxx_private_say(const QString &from, const QString &to, const QString &text);
      void sig_xxx_private_me(const QString &from, const QString &to, const QString &text);
      void sig_xxx_message(const QHostAddress &source, const QCString &sig, const QString &from, const QString &to, const QString &text);
      void sig_xxx_massmessage(const QHostAddress &source, const QCString &sig, const QString &from, const QString &to, const QString &text);
      void sig_xxx_messageack(char status, const QString &from, const QString &to, char gender, const QString &answer);
      void sig_xxx_messageack(const QCString &sig, char status, const QString &from, const QString &to, char gender, const QString &answer);
      void sig_xxx_herereq(const QHostAddress &source, const QString &from, const QString &channel);
      void sig_xxx_hereack(const QString &to, const QString &channel, const QString &from, char ra);
      void sig_xxx_inforeq(const QHostAddress &source, const QString &to, const QString &from);     
      void sig_xxx_inforeqack(const QHostAddress &source, const QString &to, const QString &from, const QString &hostname, const QString &name, const QString &ip, const QString &channels, const QString &answer, const QString &os, const QString &domain, const QString &software, bool utf, const QString &fullname, const QString &occupation, const QString &department, const QString &phone, const QString &mobile, const QString &homepage, const QString &email, const QString &address);
      void sig_xxx_nickchange(const QString &from, const QString &to, char gender);
      void sig_xxx_avatarchange(const QString &from, const QString &fname, char gender);
      void sig_xxx_changeactive(const QString &from, char ractive);
      void sig_xxx_reqchannels(const QHostAddress& addr, const QString &from);
      void sig_xxx_reqchannelsack(const QHostAddress& addr, const QString &from, const QString& channels);
      void sig_xxx_changestatus(const QString &uid, const QString &who, char status, char gender, const QString &answer);
      void sig_xxx_currenttopic(const QHostAddress &source, const QString &to, const QString &channel, const QString &topic);
      void sig_xxx_newtopic(const QHostAddress &source, const QString &channel, const QString &topic);
      void sig_xxx_ping(const QHostAddress &source, const QString &to, const QString &from, const QString &time, const QCString &bids);
      void sig_xxx_pong(const QHostAddress &source, const QString &to, const QString &from, const QString &time, const QCString &bids);
      void sig_xxx_floodnotify(const QString &to, const QString &from, const QString &seconds);
      void sig_xxx_avatarrequest(const QHostAddress &addr, const QString & fname);
      void sig_xxx_whohereneeded(const QHostAddress &addr);
      void sig_xxx_boardmsgupd(const QHostAddress &source, const QString &uid, Q_UINT32 id);
      void sig_xxx_boardmsgdel(const QHostAddress &source, const QString &uid, Q_UINT32 id);
      void sig_xxx_sendtoprocess(const QHostAddress &source, const QString &uid, const QString &fname);
      void sig_activityDetected(const QHostAddress &addr);
      void sig_httpFileRecvFinished(const QString&, const QString&);
      void sig_tcpFileRecvFinished(const QString&, const QString&);
      void sig_tcpAvatarRecvFinished(const QString&, const QString&);
      void sig_tcpFileSendFinished(const QString&, const QString&);
      void sig_tcpMessageACKReceived(const QString&, const QString&, char);
      void sig_tcpPvtMsgReceived(const QString&, const QString&);
      void sig_tcpRecvChatMessage(const QHostAddress&, const QString&);
      void sig_tcpChatSessionEstablished(const QString&);
      void sig_tcpCloseChatSession(const QString&);
      void sig_tcpRecvTypingMode(const QString&, uint);
      void sig_addToBoard(BoardMsg *);
     
  public:
	  XXXProtocol(QObject *parent, QTextCodec *c, const QUuid &u, CRYPTKEY *pCryptKey, bool useutf);
	  ~XXXProtocol();
		
	  QString getHWAddr(const QString& addr, const QString& iface);
	  QString getProtoVersion(const QHostAddress& addr);
	  Q_UINT16 getProtoVersionMajor(const QHostAddress& addr);
	  Q_UINT16 getProtoVersionMinor(const QHostAddress& addr);
	  QString getUuid(const QHostAddress& addr);
	  QString getLocalUid();
	  bool delUser(const QHostAddress &addr);
	  void setAnsiCodec(QTextCodec *c);
	  void setUseUTF(bool f);
	  void setUserVersion(const QHostAddress& addr, Q_UINT32 v);
	  void setUserUuid(const QHostAddress& addr, const QString& u);
	  QString getNick(const QHostAddress& addr);
	  QHostAddress getIP(const QString &uid);
	  static QString Uuid(const QUuid& uid);
	  QString getDownloadPath();
	  QString getLocalNick();
	  QString getLocalAnswer();
	  QString getLocalAvatar();
	  QString getLocalFile();
	  bool getAcceptWOPrompt();
	  int getMaxAvatarSize();
	  char getLocalStatus();
	  char getLocalGender();
	  
	  static bool get(QCString *str, char *buffer, int &pos, const int p_len);
	  static bool get(QString *str, char *buffer, int &pos, const int p_len, QTextCodec *tc);  
	  static bool get(char *c, char *buffer, int &pos, const int p_len);
	  static bool get(Q_UINT16 *s, char *buffer, int &pos, const int p_len);
	  static bool get(Q_UINT32 *s, char *buffer, int &pos, const int p_len);
	  static bool get(QUuid *u, char *buffer, int &pos, const int p_len);  
	  static bool get(QByteArray *arr, char *buffer, int &pos, const int p_len, Q_INT32 psize);
	  
	  bool readBMessage(const QByteArray &ba, int &offset);
	  
	  bool setupInterface(const QHostAddress &ip, const QString &bcastaddress, const QHostAddress &multicastgroup, uint ttl, uint port, bool usemcast);
	  void closeSockets();
	  
	  bool xxx_createChatSession(const QHostAddress &addr, const QString &uid);
	  bool xxx_isSessionEstablished(const QString &uid);
      
  public slots:
      void xxx_say(const QString &channel, const QString &nick, const QString &text);
      void xxx_me(const QString &channel, const QString &nick, const QString &text);
      void xxx_whohere(const QString &updater);
      void xxx_whohere(const QHostAddress &addr, const QString &updater);
      void xxx_imhere(const QHostAddress &addr, const QString &updater, const QString &responder, char status, char gender, char winstate, const QString &fname, QColor color);
      void xxx_beep(const QHostAddress &addr, const QString& to, const QString& from);
      void xxx_beepack(const QHostAddress &addr, const QString& to, const QString& from, char gender);     
      void xxx_join(const QString&, const QString&, char, char, const QString &fname, QColor color);
      void xxx_leave(const QString &who, const QString &channel, char gender);
      void xxx_private_join(const QHostAddress &addr, const QString &from, const QString &to, char gender); 
      void xxx_private_leave(const QHostAddress &addr, const QString &from, const QString &to, char gender);
      void xxx_private_say(const QHostAddress &addr, const QString &from, const QString &to, const QString &text);
      void xxx_private_me(const QHostAddress &addr, const QString &from, const QString &to, const QString &text);
      void xxx_message(const QHostAddress &addr, const QString &from, const QString &to, const QString &text);
      void xxx_massmessage(const QHostAddress &addr, const QString &from, const QString &to, const QString &text);
       void xxx_messageack(const QHostAddress &addr, const QCString &sig, char status, const QString &from, const QString &to, char gender, const QString &answer);
      void xxx_herereq(const QString &from, const QString &channel);
      void xxx_hereack(const QHostAddress &addr, const QString &to, const QString &channel, const QString &from, char ra);
      void xxx_inforeq(const QHostAddress &addr, const QString &to, const QString &from);
      void xxx_inforeqack(const QHostAddress &addr, const QString &to, const QString &from, const QString &hostname, const QString &name, const QString &ip, const QString &channels, const QString &answer, const QString &os, const QString &software, const QString &domain, const QString &fullname, const QString &occupation, const QString &department, const QString &phone, const QString &mobile, const QString &homepage, const QString &email, const QString &address);
      void xxx_nickchange(const QString &from, const QString &to, char gender);
      void xxx_changeactive(const QString &from, char ractive);
      void xxx_reqchannels(const QString& from);
      void xxx_reqchannelsack(const QHostAddress &addr, const QString& from, const QString& chlist);
      void xxx_avatarchange(const QString &from, const QString &fname, char gender);
      void xxx_changestatus(const QString &who, char status, char gender, const QString &answer);
      void xxx_currenttopic(const QHostAddress &addr, const QString &to, const QString &channel, const QString &topic);
      void xxx_newtopic(const QString &channel, const QString &topic);
      void xxx_ping(const QHostAddress &addr, const QString &to, const QString &from, const QString &time);
      void xxx_pong(const QHostAddress &addr, const QString &to, const QString &from, const QString &time);
      void xxx_floodnotify(const QHostAddress &addr, const QString &to, const QString &from, int seconds);
      void xxx_avatarrequest(const QHostAddress& addr, const QString& avatar);
      void xxx_boardmsgupdate(Q_UINT32 id);
      void xxx_boardmsgrequest(const QHostAddress& addr, const QString& nick, const QString &bids);
      void xxx_requestmissed(const QHostAddress& addr, const QString& nick, const QString &bids);
      void xxx_tcpfilesend(const QStringList& filesnames, const QString& descr, const QString& receiver, const QHostAddress& addr, bool avatar);
      void xxx_httpGet(const QString &addr, const QString &frq, const QString &fsd);
      void xxx_httpGet(const QString &addr, int p, const QString &comment, const QString &frq, const QString &fsd);
      void xxx_sendChatMessage(const QString &uid, const QString &text);
      void xxx_sendTypingMode(const QString &uid, uint tmode);
      void xxx_sleepChatSession(const QString &uid);
      void xxx_wakeupChatSession(const QString &uid);
};

#endif
